# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# useful macros/fcts


MACRO(MACRO_AddTest srcDir)
    #message(STATUS "Adding test directory ${srcDir}")

    #Extract the number of MPI processes used to 
    #launch one test.
    #If no extra arguments are given, the number
    #is set to 1.
    set (extra_macro_args ${ARGN})
    list(LENGTH extra_macro_args num_extra_args)
    if (${num_extra_args} GREATER 0)
        list(GET extra_macro_args 0 nProcs)
    else()
        set(nProcs 1)
    endif ()

    file(GLOB tfiles RELATIVE ${srcDir} ${srcDir}/*)
    #message(STATUS "tfiles=${tfiles}")
    foreach(tfile ${tfiles})
        set(spath ${srcDir}/${tfile})
        if(NOT IS_DIRECTORY ${spath} AND 
           ${spath} MATCHES ".+\\.py$" AND 
           NOT ${tfile} STREQUAL "__init__.py")
            string(REPLACE "${PROJECT_SOURCE_DIR}/" "" strip ${spath}) 
            message(STATUS "Adding test ${strip}")
            if (${nProcs} GREATER 1)
                add_test(NAME "${strip}-MPI${nProcs}"
                        WORKING_DIRECTORY ${PROJECT_SOURCE_DIR} 
                        COMMAND mpirun -np ${nProcs} ${PYTHON_EXECUTABLE} run.py --nogui --clean ${strip})
            else()
                add_test(NAME ${strip} 
                        WORKING_DIRECTORY ${PROJECT_SOURCE_DIR} 
                        COMMAND ${PYTHON_EXECUTABLE} run.py --nogui --clean ${strip})            
            endif ()
        else()
            MACRO_AddTest(${srcDir}/${tfile} ${nProcs})
        endif()
    endforeach()
ENDMACRO()


MACRO(MACRO_DebugPostfix libname)
    IF(MSVC)
        SET_TARGET_PROPERTIES(${libname} PROPERTIES DEBUG_POSTFIX "_d")
    ENDIF(MSVC)
ENDMACRO()
