#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import print_function
import katropton.CMAME_tests.model as model

import katropton as m
import tbox
import tbox.gmsh as gmsh
import shlex, subprocess
import os
import numpy as np

from fwk.wutils import parseargs

from PyTrilinos import Teuchos

import katropton.utilities as u

def main():

    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print("info: MPI found")
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print("info: MPI not found => MPI disabled")


    args=parseargs()

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    # Units in the meshes are in mm
    msh = u.fuse_meshes('plate_1.msh','plate_1_tmp','plate_2.msh','plate_2_tmp','plate_fused',file_dir,work_dir)

    np.random.seed(42)
    N = 1
    num_random_variables = 20
    #randomVariable = np.random.normal(0, 1, num_random_variables*N)
    #randomVariable = np.resize(randomVariable,(N,num_random_variables))
    randomVariable = np.zeros((N,num_random_variables))
    
    p = model.defaut_params
    p['MueLu xml from current_dir'] = False
    p['MueLu xml use template'] = False
    p['MueLu xml file name'] = '/../tests/direct_solver.xml'
    p['Write txt files'] = True
    p['Write matrix and vectors'] = True

    ensemble_size = 1
    
    pm = model.plate_model(ensemble_size, randomVariable, msh, comm, file_dir, work_dir, p)
    pm.run(args.k)

if __name__ == "__main__":
    main()
