#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import print_function
from builtins import str
from builtins import range
import katropton.CMAME_tests.model as model

import katropton as m
import tbox
import tbox.gmsh as gmsh
import shlex, subprocess
import os
import numpy as np

from fwk.wutils import parseargs

from PyTrilinos import Teuchos

import katropton.utilities as u

def evaluate_gamma(gamma, msh, comm, file_dir):

    work_dir = os.getcwd()
    
    p = model.defaut_params
    p['MueLu xml from current_dir'] = False
    p['MueLu xml use template'] = False
    p['MueLu xml file name'] = '/../preconditioners/SIMPLE_direct.xml'
    p['gamma'] = gamma
    
    N = 1
    num_random_variables = 20
    randomVariable = np.zeros((N,num_random_variables))
    
    pm = model.plate_model(ensemble_size, randomVariable, msh, comm, file_dir, work_dir, p)
    pm.run(args.k)


def evaluate_all_gamma(gammas, msh, comm, file_dir):

    N = gammas.shape[0]
    for i in range(0,N):
        directory = str(i)
        if not os.path.exists(directory):
            os.makedirs(directory)
        os.chdir(directory)

        evaluate_gamma(gammas[i], msh, comm, file_dir)

        os.chdir('..')

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print("info: MPI found")
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print("info: MPI not found => MPI disabled")

    args=parseargs()

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    # Units in the meshes are in mm
    msh = u.fuse_meshes('plate_1.msh','plate_1_tmp','plate_2.msh','plate_2_tmp','plate_fused',file_dir,work_dir)

    N = 200
    gamma_min_exp = -2
    gamma_max_exp = 10

    gammas = np.logspace(gamma_min_exp, gamma_max_exp, N, endpoint=True)
    
    np.savetxt('gammas.txt', gammas)

    evaluate_all_gamma( gammas, msh, comm, file_dir)

if __name__ == "__main__":
    main()
