from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div
# require a git clone https://github.com/kliegeois/Same-Size-K-Means.git

from clustering.equal_groups import EqualGroupsKMeans
import numpy as np
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from mpl_toolkits.mplot3d import Axes3D

def compute_the_groups(N,ensemble_size,num_random_variables):

    np.random.seed(42)

    n_clusters = int(old_div(N,ensemble_size))

    randomVariable = np.random.normal(0, 1, num_random_variables*N)
    X = np.resize(randomVariable,(N,num_random_variables))

    clf = EqualGroupsKMeans(n_clusters=n_clusters)

    clf.fit(X)

    Z = clf.predict(X)
    Z = Z.astype(int)

    for i in range(0,n_clusters):
      l_s = len(np.argwhere(Z==i))
      if l_s != ensemble_size:
        print('Ensemble ' + str(i) + ' has ' + str(l_s) + ' samples instead of ' + str(ensemble_size))

    np.savetxt('Groupe_id_s_'+str(ensemble_size)+'_'+str(N)+'.txt',Z, fmt='%i')

    indices = np.argsort(Z)

    np.savetxt('Reordering_s_'+str(ensemble_size)+'_'+str(N)+'.txt',indices, fmt='%i')

    '''
    reduced_data = PCA(n_components=3).fit_transform(X)

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    x_min, x_max = reduced_data[:, 0].min() - 1, reduced_data[:, 0].max() + 1
    y_min, y_max = reduced_data[:, 1].min() - 1, reduced_data[:, 1].max() + 1
    z_min, z_max = reduced_data[:, 2].min() - 1, reduced_data[:, 2].max() + 1
    ax.scatter(reduced_data[:, 0], reduced_data[:, 1],reduced_data[:, 2], s = 20, c=Z)
    ax.set_xlim(x_min, x_max)
    ax.set_ylim(y_min, y_max)
    ax.set_zlim(z_min, z_max)
    plt.show()
    '''

def main():

    N = 960
    num_random_variables = 20
    
    compute_the_groups(N,8,num_random_variables)
    compute_the_groups(N,16,num_random_variables)
    compute_the_groups(N,32,num_random_variables)

if __name__ == "__main__":
    main()

