#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div
from katropton.eigenvalues import *
import os
import fwk

def plot_as_bar_plot(x,y,semilog=False):
    import matplotlib.pyplot as plt
    plt.figure()

    for i in range(0,len(x)):
      if i == 0:
        x_bar = np.array([x[0], 0.5*x[0]+0.5*x[1]])
        y_bar = np.array([y[0], y[0]])
      elif i!= len(x)-1:
        x_bar = np.append( x_bar , [x_bar[-1], 0.5*x[i]+0.5*x[i+1]])
        y_bar = np.append( y_bar , [y[i], y[i]])
      else:
        x_bar = np.append( x_bar , [x_bar[-1], x[i]])
        y_bar = np.append( y_bar , [y[i], y[i]])
        break
    if semilog:
      plt.semilogy(x_bar,y_bar)
    else:
      plt.plot(x_bar,y_bar)

def analysis_deltas_omegas(n_samples,base_dir,GS,write):
    import matplotlib.pyplot as plt
    current_n = 1
    
    theta = np.zeros((current_n,n_samples))
    omegas = np.zeros((current_n,n_samples))
    deltas = np.zeros((current_n,n_samples), dtype=np.complex128)
    
    if GS:
      base_dir_omega = base_dir + 'GS/omegas_GS_GS_'
      base_dir_delta = base_dir + 'GS/deltas_GS_GS_'
    else:
      base_dir_omega = base_dir + 'direct/omegas_direct_direct_'
      base_dir_delta = base_dir + 'direct/deltas_direct_direct_'
    for i in range(0,n_samples):
      tmp_delta = np.loadtxt(base_dir_delta+str(i)+'.txt')
      omegas_c = np.loadtxt(base_dir_omega+str(i)+'.txt')

      n = tmp_delta.shape[0]

      if current_n < n:
        
        theta_tmp = np.zeros((n,n_samples))
        omegas_tmp = np.zeros((n,n_samples))
        deltas_tmp = np.zeros((n,n_samples), dtype=np.complex128)
        
        theta_tmp[0:current_n,:] = theta
        omegas_tmp[0:current_n,:] = omegas
        deltas_tmp[0:current_n,:] = deltas

        theta = theta_tmp
        omegas = omegas_tmp
        deltas = deltas_tmp
    
        current_n = n

      deltas_c = np.zeros((n,), dtype=np.complex128)

      deltas_c.real = tmp_delta[:,0]
      deltas_c.imag = tmp_delta[:,1]

      deltas[0:n,i] = deltas_c

      omegas[0:n,i] = omegas_c

      theta_c = compute_theta(deltas_c)

      theta[0:n,i] = theta_c

      sort_indices = np.argsort(theta_c)
      theta_c = theta_c[sort_indices]
      deltas_c = deltas_c[sort_indices]
      omegas_c = omegas_c[sort_indices]

      N = 720
      theta_discrete, cum_delta, deriv_delta, cum_delta_omega, deriv_delta_omega, percent_delta, percent_delta_omega = compute_cum_delta(theta_c,deltas_c,omegas_c,N)

      indices = np.zeros((n,))
      sort_indices = np.argsort(-omegas_c)
      indices[sort_indices[0:30]] = 1
      plot_eigenvalues_on_unit_circle_highlight(deltas_c,indices)
      plot_polar_function(theta_discrete,deriv_delta,1.5,2.5)
      plot_polar_function(theta_discrete,deriv_delta_omega,3,4)
      plt.title('Sample ' + str(i))
      if GS:
        plt.savefig('fig_GS_'+str(i)+'.png')
      else:
        plt.savefig('fig_direct_'+str(i)+'.png')

      plot_as_bar_plot(theta_discrete,deriv_delta_omega)
      plt.title('Sample ' + str(i))
      if GS:
        plt.savefig('fig_GS_omegas_'+str(i)+'.png')
      else:
        plt.savefig('fig_direct_omegas_'+str(i)+'.png')
        
      tmp = np.ones((len(theta[0:n,i]),2))
      tmp[:,0] = theta[0:n,i]
      if write:
        if GS:
          np.savetxt('theta_GS_'+str(i)+'.txt',tmp)
        else:
          np.savetxt('theta_direct_'+str(i)+'.txt',tmp)

      tmp = np.ones((len(theta_discrete),2))
      tmp[:,0] = theta_discrete

      if write:
        if GS:
          np.savetxt('theta_GS_discrete_'+str(i)+'.txt',tmp)
          np.savetxt('deriv_GS_delta_'+str(i)+'.tex',deriv_delta)
          np.savetxt('deriv_GS_delta_omega_'+str(i)+'.tex',deriv_delta_omega)
        else:
          np.savetxt('theta_direct_discrete_'+str(i)+'.txt',tmp)
          np.savetxt('deriv_direct_delta_'+str(i)+'.tex',deriv_delta)
          np.savetxt('deriv_direct_delta_omega_'+str(i)+'.tex',deriv_delta_omega)
          
    if write:
      if GS:
        np.savetxt('omegas_GS_all.txt',omegas)
      else:
        np.savetxt('omegas_direct_all.txt',omegas)

    ensemble_sizes = np.array([8,16,32])

    for i_s in range(0,len(ensemble_sizes)):
      s = ensemble_sizes[i_s]
      for i in range(0,n_samples,s):
        i_sample_min = i
        i_sample_max = i_sample_min + s
        
        omegas_s = omegas[:,i_sample_min:i_sample_max].flatten('F')
        deltas_s = deltas[:,i_sample_min:i_sample_max].flatten('F')
        theta_s = theta[:,i_sample_min:i_sample_max].flatten('F')
        index = np.argwhere(np.absolute(deltas_s)>=0.9)
        
        omegas_s = omegas_s[index]
        deltas_s = deltas_s[index]
        theta_s = theta_s[index]
        
        sort_indices = np.argsort(theta_s)
        theta_s = theta_s[sort_indices]
        deltas_s = deltas_s[sort_indices]
        omegas_s = omegas_s[sort_indices]
        
        N = 720
        theta_discrete, cum_delta, deriv_delta, cum_delta_omega, deriv_delta_omega, percent_delta, percent_delta_omega = compute_cum_delta(theta_s,deltas_s,omegas_s,N)

        n = len(omegas_s)
        indices = np.zeros((n,))
        sort_indices = np.argsort(-omegas_s)
        indices[sort_indices[0:30]] = 1
        plot_eigenvalues_on_unit_circle_highlight(deltas_s,indices)
        plot_polar_function(theta_discrete,deriv_delta,1.5,2.5)
        plot_polar_function(theta_discrete,deriv_delta_omega,3,4)
        plt.title('Ensemble ' + str(int(old_div(i,s))) + ' of ensemble size ' + str(s))
        if GS:
          plt.savefig('fig_s_GS_'+ str(s) +'_' + str(int(old_div(i,s))) + '.png')
        else:
          plt.savefig('fig_s_direct_'+ str(s) +'_' + str(int(old_div(i,s))) + '.png')
        plt.clf()
        plt.close()
        plot_as_bar_plot(theta_discrete,deriv_delta_omega)
        plt.title('Ensemble ' + str(int(old_div(i,s))) + ' of ensemble size ' + str(s))
        if GS:
          plt.savefig('fig_s_GS_omegas_'+ str(s) +'_' + str(int(old_div(i,s))) + '.png')
        else:
          plt.savefig('fig_s_direct_omegas_'+ str(s) +'_' + str(int(old_div(i,s))) + '.png')
        plt.clf()
        plt.close()

def main():
    
    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()
    
    n_samples = 32
    
    base_dir = file_dir + '/../../../workspace/katropton_CMAME_tests_plate_deltas_omegas_'
    write = False

    analysis_deltas_omegas(n_samples,base_dir,True,write)
    analysis_deltas_omegas(n_samples,base_dir,False,write)
    


if __name__ == "__main__":
    main()
