#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div
import katropton as m
import tbox
import tbox.gmsh as gmsh

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print("info: MPI found")
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print("info: MPI not found => MPI disabled")



    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos
    import numpy as np

    mueluParams = Teuchos.ParameterList()
    #mueluParamsSub1 = Teuchos.ParameterList()
    #mueluParamsSub1['PDE equations'] = 4
    #mueluParams['Matrix'] = mueluParamsSub1
    mueluParams['number of equations'] = 4
    mueluParams['max levels'] = 5
    mueluParams['multigrid algorithm'] = "unsmoothed"
    mueluParams['smoother: type'] = "RELAXATION"
    mueluParams['coarse: max size'] = 3000

    mueluParamsSub2 = Teuchos.ParameterList()

    mueluParamsSub2['relaxation: type'] = "Symmetric Gauss-Seidel"
    mueluParamsSub2['relaxation: sweeps'] = 5
    mueluParamsSub2['relaxation: damping factor'] = 1.25

    mueluParams['coarse: type'] = "Klu"
    mueluParams['verbosity'] = "low"

    mueluParamsSub3 = Teuchos.ParameterList()

    mueluParamsSub3['relaxation: type'] = "Symmetric Gauss-Seidel"
    mueluParamsSub3['relaxation: sweeps'] = 4
    mueluParamsSub3['relaxation: damping factor'] = 1.25


    mueluParams['smoother: params'] = mueluParamsSub2
    #mueluParams['coarse: params'] = mueluParamsSub3

    solverList = Teuchos.ParameterList()
    solverList['type']= "BlockGmres"
    solverList['Convergence Tolerance'] = 10**(-7)
    solverList['Maximum Iterations'] = 2000
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 100
    solverList['mueluParams'] = mueluParams

    solverList['Use preconditioner'] = True
    
    solverList['Maximum active set iteration'] = 10

    solverList['Print Teuchos timers'] = True
    #solverList['Write matrix and vectors'] = True
    solverList['Write txt files'] = True
    

    geo_name = 'MIRROR_1_ASSEMBLY_mm.msh'
    mesh_name = 'MIRROR_1_ASSEMBLY_mm.msh'

    import shlex, subprocess
    import os

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    #solverList["MueLu xml file name"] = file_dir + '/finest_level_prec_2.xml'
    #solverList["MueLu xml file name"] = file_dir + '/finest_level_prec_simple_heat.xml'
    solverList["MueLu xml file name"] = file_dir + '/nested_multigrid_2_lvls_chebyshev_heat.xml'
    #solverList["MueLu xml file name"] = file_dir + '/nested_multigrid_2_lvls_old.xml'
    #solverList["MueLu xml file name"] = file_dir + '/nested_multigrid_2_lvls_chebyshev_heat.xml'
    #solverList["MueLu xml file name"] = file_dir + '/nested_multigrid_2_lvls_chebyshev.xml'
    
    
    if rank == 0:
        command_line = 'gmsh -3 ' + file_dir + '/mesh/' +  geo_name + ' -o ' + work_dir + '/' +  mesh_name +' -part ' + str(siz)
        print(command_line)
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log','w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
        retcode = p.wait()
        for i in range(1,siz):
            comm.send(1, dest=i, tag=11)
    else:
        data = comm.recv(source=0, tag=11)
    
    if siz == 1:
        msh = gmsh.MeshLoader(mesh_name,work_dir).execute()
    else:
        msh = gmsh.MeshLoader(mesh_name,work_dir).myrank_execute(rank)

    pbl = m.Problem(msh,comm)
    
    # Distance in the mesh are represented in minimeters

    ensemble_size = 1

    #379 000 MPa = 379 GPa = 379E9 Pa
    E_rhodium = 379000 * np.ones(ensemble_size)
    nu_rhodium = 0.26 * np.ones(ensemble_size)
    #0.150 kW/(m K) = 150 W/(m K) = 0.150 W/(mm K)
    k_rhodium = 0.150
    d_rhodium = 0.781E-05

    E_WCu = 280000 * np.ones(ensemble_size)
    nu_WCu = 0.26 * np.ones(ensemble_size)
    k_WCu = 0.180
    d_WCu = 0.88E-05
    
    E_SS = 180000 * np.ones(ensemble_size)
    nu_SS = 0.26 * np.ones(ensemble_size)
    k_SS = 0.020
    d_SS = 1.6E-05

    E_CuCrZr = 130000 * np.ones(ensemble_size)
    nu_CuCrZr = 0.26 * np.ones(ensemble_size)
    k_CuCrZr = 0.280
    d_CuCrZr = 1.6E-05

    E_I718 = 180000 * np.ones(ensemble_size)
    nu_I718 = 0.26 * np.ones(ensemble_size)
    k_I718 = 0.020
    d_I718 = 1.4E-05
    
    T_ref = 22.
    

    m.Medium(pbl, "Mirror", "rhodium",E_rhodium,nu_rhodium,k_rhodium,d_rhodium)
    m.Medium(pbl, "Substrate", "WCu",E_WCu,nu_WCu,k_WCu,d_WCu)
    
    m.Medium(pbl, "Holder", "SS",E_SS,nu_SS,k_SS,d_SS)

    m.Medium(pbl, "Washer 1", "I718",E_I718,nu_I718,k_I718,d_I718)
    m.Medium(pbl, "Washer 2", "I718",E_I718,nu_I718,k_I718,d_I718)
    m.Medium(pbl, "Washer 3", "I718",E_I718,nu_I718,k_I718,d_I718)
    
    m.Medium(pbl, "Shaft 1", "I718",E_I718,nu_I718,k_I718,d_I718)
    m.Medium(pbl, "Shaft 2", "I718",E_I718,nu_I718,k_I718,d_I718)
    m.Medium(pbl, "Shaft 3", "I718",E_I718,nu_I718,k_I718,d_I718)
    
    m.Medium(pbl, "Nut 1", "I718",E_I718,nu_I718,k_I718,d_I718)
    m.Medium(pbl, "Nut 2", "I718",E_I718,nu_I718,k_I718,d_I718)
    m.Medium(pbl, "Nut 3", "I718",E_I718,nu_I718,k_I718,d_I718)
    
    m.Medium(pbl, "Spacer 1 part 1", "CuCrZr",E_CuCrZr,nu_CuCrZr,k_CuCrZr,d_CuCrZr)
    m.Medium(pbl, "Spacer 2 part 1", "CuCrZr",E_CuCrZr,nu_CuCrZr,k_CuCrZr,d_CuCrZr)
    m.Medium(pbl, "Spacer 3 part 1", "CuCrZr",E_CuCrZr,nu_CuCrZr,k_CuCrZr,d_CuCrZr)

    m.Medium(pbl, "Spacer 1 part 2", "WCu",E_WCu,nu_WCu,k_WCu,d_WCu)
    m.Medium(pbl, "Spacer 2 part 2", "WCu",E_WCu,nu_WCu,k_WCu,d_WCu)
    m.Medium(pbl, "Spacer 3 part 2", "WCu",E_WCu,nu_WCu,k_WCu,d_WCu)
    
    if 1:
        m.uDirichlet(pbl, "Holder z clamping", "Clamped",0,0.,0,0.,1,0.,0,0.,ensemble_size)
        m.uDirichlet(pbl, "Holder x-y clamping", "Clamped",1,0.,1,0.,0,0.,0,0.,ensemble_size)
    else:
        m.uDirichlet(pbl, "Holder z clamping points", "Clamped",0,0.,0,0.,1,0.,0,0.,ensemble_size)
        m.uDirichlet(pbl, "Holder y clamping points", "Clamped",0,0.,1,0.,0,0.,0,0.,ensemble_size)
        m.uDirichlet(pbl, "Holder x clamping points", "Clamped",1,0.,0,0.,0,0.,0,0.,ensemble_size)

    m.uDirichlet(pbl, "Cooling channels", "Clamped",0,0.,0,0.,0,0.,1,70.-T_ref,ensemble_size)

    zero = np.zeros(ensemble_size)
    hf = (old_div(70,(0.0121*1000**2)))* np.ones(ensemble_size) #70 W / (0.0121 m^2)

    m.Neumann(pbl, "Mirror surface","heat fluxes",0,zero,0,zero,0,zero,1,hf,ensemble_size)

    m.Source(pbl, "Mirror", "ihg",70E-03)
    m.Source(pbl, "Substrate", "ihg",2.45E-03)
    m.Source(pbl, "Holder", "ihg",0.7E-03)

    m.Source(pbl, "Spacer 1 part 1", "ihg",0.7E-03)
    m.Source(pbl, "Spacer 2 part 1", "ihg",0.7E-03)
    m.Source(pbl, "Spacer 3 part 1", "ihg",0.7E-03)

    m.Source(pbl, "Spacer 1 part 2", "ihg",2.45E-03)
    m.Source(pbl, "Spacer 2 part 2", "ihg",2.45E-03)
    m.Source(pbl, "Spacer 3 part 2", "ihg",2.45E-03)

    m.Source(pbl, "Shaft 1", "ihg",0.7E-03)
    m.Source(pbl, "Shaft 2", "ihg",0.7E-03)
    m.Source(pbl, "Shaft 3", "ihg",0.7E-03)

    m.Source(pbl, "Nut 1", "ihg",0.7E-03)
    m.Source(pbl, "Nut 2", "ihg",0.7E-03)
    m.Source(pbl, "Nut 3", "ihg",0.7E-03)

    m.Source(pbl, "Washer 1", "ihg",0.7E-03)
    m.Source(pbl, "Washer 2", "ihg",0.7E-03)
    m.Source(pbl, "Washer 3", "ihg",0.7E-03)
    
    norm1 = tbox.Pt(0,0,-1)
    norm2 = tbox.Pt(0,0,1)

    ## Bolt 1
    cont = m.Contact(pbl,"Bolt 1 to holder","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    #cont.setInitialyOpen()
    cont.setMaster(pbl,"Holder to bolt 1",norm1)

    cont = m.Contact(pbl,"Spacer 1 to holder","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Holder to spacer 1",norm1)

    cont = m.Contact(pbl,"Bolt 1 to washer 1","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Washer 1 to bolt 1",norm1)

    cont = m.Contact(pbl,"Spacer 1 to washer 1 part 1","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Washer 1 to spacer 1 part 1",norm1)

    cont = m.Contact(pbl,"Spacer 1 to washer 1 part 2","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Washer 1 to spacer 1 part 2",norm1)
    
    cont = m.Contact(pbl,"Spacer 1 to substrate","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Substrate to spacer 1",norm1)
    
    ## Bolt 2
    cont = m.Contact(pbl,"Bolt 2 to holder","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Holder to bolt 2",norm1)

    cont = m.Contact(pbl,"Spacer 2 to holder","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Holder to spacer 2",norm1)

    cont = m.Contact(pbl,"Bolt 2 to washer 2","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Washer 2 to bolt 2",norm1)

    cont = m.Contact(pbl,"Spacer 2 to washer 2 part 1","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Washer 2 to spacer 2 part 1",norm1)

    cont = m.Contact(pbl,"Spacer 2 to washer 2 part 2","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Washer 2 to spacer 2 part 2",norm1)
    
    cont = m.Contact(pbl,"Spacer 2 to substrate","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Substrate to spacer 2",norm1)
    
    ## Bolt 3
    cont = m.Contact(pbl,"Bolt 3 to holder","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Holder to bolt 3",norm1)

    cont = m.Contact(pbl,"Spacer 3 to holder","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Holder to spacer 3",norm1)

    cont = m.Contact(pbl,"Bolt 3 to washer 3","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Washer 3 to bolt 3",norm1)

    cont = m.Contact(pbl,"Spacer 3 to washer 3 part 1","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Washer 3 to spacer 3 part 1",norm1)

    cont = m.Contact(pbl,"Spacer 3 to washer 3 part 2","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Washer 3 to spacer 3 part 2",norm1)
    
    cont = m.Contact(pbl,"Spacer 3 to substrate","contact",norm1,0.)
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Substrate to spacer 3",norm1)
    
    slv = m.IterativeSolver(pbl,args.k,solverList,4, False, ensemble_size)

    slv.start()

    '''
    ## Extract results
    if rank == 0:
        import katropton.utilities as u
        
        tag_name = "Mirror surface"
        name_sol = "x_mm.txt"

        DOFperNode = 4
        
        if siz > 1:
            x,y,z,tri,sol = u.extract_2D_all(msh2, tag_name, DOFperNode, ensemble_size, name_sol)
        else:
            x,y,z,tri,sol = u.extract_2D_all(msh, tag_name, DOFperNode, ensemble_size, name_sol)
        
        Zernike_coeff, Zernike_functions, theta, rho = u.compute_Zernike(x,y,z,tri,sol[:,2],10,np.array([-0.015,0.005]))

        colors = {0:(0,0.4470,0.7410), 1:(0.8500,0.3250,0.0980), 2:(0.9290,0.6940,0.1250), 3:(0.4940,0.1840,0.5560), 4:(0.4660,0.6740,0.1880), 5:(0.3010,0.7450,0.9330), 6:(0.6350,0.0780,0.1840), 7:(0,0.4470,0.7410)}
        
        import matplotlib.pyplot as plt
        from matplotlib import cm
        #plt.style.use('ggplot')

        plt_linewidth = 4.0
        
        print max(z)
        print min(z)

        print max(sol[:,2])
        print min(sol[:,2])
        
        print Zernike_coeff
        
        print u.check_optical_tolerance(Zernike_coeff, Zernike_functions, sol[:,2],"m")


        for which_dof in range(0,DOFperNode):
            levels = np.linspace(min(sol[:,which_dof]), max(sol[:,which_dof]), 15)
            # Velocity x-component
            fig = plt.figure(figsize=(8, 8))
            plt.set_cmap('jet')
            plt.tricontour(x, y, tri, sol[:,which_dof], levels, linestyles='-',
                        colors='black', linewidths=0.5)
            axtri = plt.tricontourf(x, y, tri, sol[:,which_dof], levels, cmap='coolwarm')
            #plt.triplot(x, y, tri)
            plt.gca().set_aspect('equal')
            plt.gca().set_axis_off()
            

            
            #coolwarm
            
            #cb = fig.colorbar(axtri, orientation="horizontal")
            #x = np.linspace(0.0, 1.0, 15)
            #rgb = cm.get_cmap('coolwarm')(x)[np.newaxis, :, :3]
            #print rgb
            
            if which_dof == 0:
                #cb.set_label('displacement along x [m]', rotation=0)
                plt.savefig("ux.pdf", transparent=True, bbox_inches='tight', pad_inches=0.05)
                print "ux"
                print max(sol[:,which_dof])
                print min(sol[:,which_dof])
            elif which_dof == 1:
                #cb.set_label('displacement along y [m]', rotation=0)
                plt.savefig("uy.pdf", transparent=True, bbox_inches='tight', pad_inches=0.05)
                print "uy"
                print max(sol[:,which_dof])
                print min(sol[:,which_dof])
            elif which_dof == 2:
                #cb.set_label('displacement along z [m]', rotation=0)
                plt.savefig("uz.pdf", transparent=True, bbox_inches='tight', pad_inches=0.05)
                print "uz"
                print max(sol[:,which_dof])
                print min(sol[:,which_dof])
            elif which_dof == 3:
                #cb.set_label('Temperature [C]', rotation=0)
                plt.savefig("T.pdf", transparent=True, bbox_inches='tight', pad_inches=0.05)
                print "T"
                print max(sol[:,which_dof])
                print min(sol[:,which_dof])
    
        #plt.minorticks_on()
        #plt.gca().set_xticklabels([])
        #plt.gca().set_yticklabels([])
        #plt.gca().set_axis_off()
        #plt.subplots_adjust(top = 1, bottom = 0, right = 1, left = 0, hspace = 0, wspace = 0)
        #plt.margins(0,0)
        #plt.gca().xaxis.set_major_locator(plt.NullLocator())
        #plt.gca().yaxis.set_major_locator(plt.NullLocator())

        zernike_fit = np.dot(Zernike_functions,Zernike_coeff)

        levels = np.linspace(min(zernike_fit), max(zernike_fit), 16)
        # Velocity x-component
        fig = plt.figure(figsize=(8, 8))
        plt.set_cmap('jet')
        plt.tricontour(x, y, tri, zernike_fit, levels, linestyles='-',
                    colors='black', linewidths=0.5)
        axtri = plt.tricontourf(x, y, tri, zernike_fit, levels)
        #plt.triplot(x, y, tri)
        plt.gca().set_aspect('equal')
        
        cb = fig.colorbar(axtri)
        
        levels = np.linspace(min(sol[:,2]-zernike_fit), max(sol[:,2]-zernike_fit), 16)
        # Velocity x-component
        fig = plt.figure(figsize=(8, 8))
        plt.set_cmap('jet')
        plt.tricontour(x, y, tri, sol[:,2]-zernike_fit, levels, linestyles='-',
                    colors='black', linewidths=0.5)
        axtri = plt.tricontourf(x, y, tri, sol[:,2]-zernike_fit, levels)
        #plt.triplot(x, y, tri)
        plt.gca().set_aspect('equal')
        
        cb = fig.colorbar(axtri)



        plt.show()
    '''
    from fwk.testing import *
    tests = CTests()
    tests.add(CTest('mean(Tbottom)', 0, 0))
    tests.run()

if __name__ == "__main__":
    main()
