// SWIG input file of the 'katropton' module

%feature("autodoc","1");

%module(docstring=
"'katroptonw' module: tests of Trilinos library
(c) ULg - A&M",
directors="1",
threads="1"
) katroptonw
%{

#include <string>
#include <sstream>
#include <typeinfo>
#include "katropton.h"

#include <mpi.h>
#include "Teuchos_ParameterList.hpp"

#include "fwkw.h"
#include "tboxw.h"

#include "Discrete_problem.h"
#include "wProblem.h"
//#include "wSolver.h"
#include "wMedium.h"
#include "wuDirichlet.h"
#include "wNeumann.h"
#include "RandomVariable.h"
#include "wSource.h"
#include "wContact.h"
#include "wDisplayHook.h"
#include "LinearSolver.h"
#include "DirectSolver.h"
#include "IterativeSolver.h"

#include <vector>

//Fix for missing SWIGPY_SLICE_ARG with some versions of swig.
#if PY_VERSION_HEX >= 0x03020000
# define SWIGPY_SLICE_ARG(obj) ((PyObject*) (obj))
#else
# define SWIGPY_SLICE_ARG(obj) ((PySliceObject*) (obj))
#endif

%}

%include "fwkw.swg"

// ----------- MODULES UTILISES ------------
%import "fwkw.i"
%import "tboxw.i"


// ----------- KATROPTON CLASSES ----------------

// Instantiate some std templates

%include mpi4py/mpi4py.i
%mpi4py_typemap(Comm, MPI_Comm);

%feature("notabstract") MpiComm_int;
//%include Teuchos.i


//%teuchos_rcp_pydict_overrides(CONST, CLASS...);

%include "katropton.h"

%shared_ptr(katropton::Medium);
%shared_ptr(katropton::uDirichlet);
%shared_ptr(katropton::Neumann);
%shared_ptr(katropton::RandomVariable);
%shared_ptr(katropton::Contact);
%shared_ptr(katropton::Source);

%feature("director") DisplayHook;
%include "wDisplayHook.h"

%feature("director:except") {
    if ($error != NULL) {
        std::cout << "[in director:except]\n";
        //throw Swig::DirectorMethodException();
        throw std::runtime_error("Director problem");
    }
}

%immutable katropton::Problem::msh; // avoid the creation of the setter method
%include "wProblem.h"
//%include "wSolver.h"
%immutable katropton::LinearSolver::tms;
%include "LinearSolver.h"
%include "DirectSolver.h"
%include "IterativeSolver.h"

%pythonappend katropton::Medium::Medium "self.thisown=0"
%include "wMedium.h"
%pythonappend katropton::uDirichlet::uDirichlet "self.thisown=0"
%include "wuDirichlet.h"
%pythonappend katropton::Neumann::Neumann "self.thisown=0"
%include "wNeumann.h"
%pythonappend katropton::RandomVariable::RandomVariable "self.thisown=0"
%include "RandomVariable.h"
%pythonappend katropton::Source::Source "self.thisown=0"
%include "wSource.h"
%pythonappend katropton::Contact::Contact "self.thisown=0"
%include "wContact.h"

%include "Discrete_problem.h"
namespace katropton {
    %template(Discrete_problem_d) Discrete_problem<double>;
}
//Problem::Problem(std::shared_ptr<MshData> _msh, MPI_Comm  _comm);



namespace std {
    %template(std_vector_KatProblem) vector<katropton::Problem*>;
}

