#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import print_function
from __future__ import division
from builtins import zip
from builtins import range
from past.utils import old_div
from builtins import object
import numpy as np

verb = False  # set to True for (a lot of) debug info

try:
    import mpi4py.MPI as mpi
    comm = mpi.COMM_WORLD
    rank = comm.rank
    siz  = comm.size
    name = mpi.Get_processor_name()
    status = mpi.Status()
except:
    comm = None
    rank = 0
    siz  = 1
    name = "noname"

def enum(*sequential, **named):
    enums = dict(list(zip(sequential, list(range(len(sequential))))), **named)
    return type('Enum', (), enums)

tags = enum('READY', 'DONE', 'EXIT', 'START', 'WAKEUP', 'SUICIDE')

class Index_no_repetition_2(object):
  def __init__(self, among_i):
    self.among_i = among_i
    self.n = len(among_i)
    self.i = np.zeros((2,), dtype=int)
    self.i_index = np.zeros((2,), dtype=int)
    self.is_valid = True
    for j in range(0,2):
      self.i_index[j] = j
      self.i[j] = among_i[self.i_index[j]]

  def compute_next(self):
    n = self.n
    if self.i_index[1] < n-1:
      self.i_index[1] = self.i_index[1]+1
      self.i[1] = self.among_i[self.i_index[1]]
    else:
      if self.i_index[0] < n-2:
        self.i_index[0] = self.i_index[0]+1
        self.i_index[1] = self.i_index[0]+1
        for j in range(0,2):
          self.i[j] = self.among_i[self.i_index[j]]
      else:
        self.is_valid = False
    return self.is_valid

  def get_i(self):
    return self.i

class Index_no_repetition(object):
  def __init__(self, k, n, n_0=0):
    self.k = k
    self.n = n
    self.i = np.zeros((k,), dtype=int)
    self.i_max = np.zeros((k,), dtype=int)
    self.is_valid = True
    for j in range(0,k):
      self.i[j] = n_0+j
    for j in range(0,k-1):
      self.i_max[j] = self.i[j+1] - 1
    self.i_max[-1] = n - 1

  def compute_next(self):
    k = self.k
    mask = (self.i<self.i_max-1)
    tmp = np.argwhere(mask == True)
    j = -1
    for i in range(0,k): j = (1-mask[i])*j + mask[i]*i
    if j == -1: self.is_valid = False
    else:
      self.i[j:k] = self.i[j]+1+np.arange(0,k-j)
      if j == 0: j_1 = 0
      else: j_1 = j - 1
      self.i_max[j_1:k-1] = self.i[j_1+1:k]
    return self.is_valid
    
  def get_is_valid(self):
    return self.is_valid

  def get_i(self):
    return self.i

def compute_convergence(k_max,n,omegas,deltas,A,b,x0,verbose=False):
  if siz > 1:
    convergence = compute_convergence_mpi(k_max,n,omegas,deltas,A,b,x0,verbose)
  else:
    convergence = compute_convergence_serial(k_max,n,omegas,deltas,A,b,x0,verbose)
  return convergence

def compute_sum_of_product_times_product(k,n,omegas,deltas):
  i1 = Index_no_repetition(k, n)
  results_sum = 0.
  while True:
    results_product_1 = 1.
    results_product_2 = 1.
    index = i1.get_i()
    print(index)
    results_product_1 = results_product_1 * np.prod(omegas[index])
    i2 = Index_no_repetition_2(index)
    while True:
      index2 = i2.get_i()
      tmp = np.absolute(deltas[index2[0]]-deltas[index2[1]])
      results_product_2 = results_product_2 * tmp**2
      if not i2.compute_next(): break
    results_sum = results_sum + results_product_1 * results_product_2
    if not i1.compute_next(): break
  return results_sum

def compute_convergence_serial(k_max,n,omegas,deltas,A,b,x0,verbose):
  convergence = np.ones((k_max,))
  convergence[0] = (np.linalg.norm(b - A.dot(x0)))**2
  old_sum = compute_sum_of_product_times_product(2,n,omegas,deltas)
  convergence[1] = old_sum
  for k in range(2,k_max):
    new_sum = compute_sum_of_product_times_product(k+1,n,omegas,deltas)
    convergence[k] = old_div(new_sum,old_sum)
    if verbose:
      print(convergence)
    old_sum = new_sum
  return convergence

def compute_convergence_mpi(k_max,n,omegas,deltas,A,b,x0,verbose):
  convergence = np.ones((k_max,))
  convergence[0] = (np.linalg.norm(b - A.dot(x0)))**2
  
  if rank == 0:
    master = Master()
    old_sum = master.start(2,n,omegas,deltas)
    convergence[1] = old_sum
  else:
    worker = Worker(rank)
    worker.start()

  for k in range(2,k_max):
    if rank == 0:
      master = Master()
      new_sum = master.start(k+1,n,omegas,deltas)
      convergence[k] = old_div(new_sum,old_sum)
      old_sum = new_sum
      if verbose:
        print(convergence)
    else:
      worker = Worker(rank)
      worker.start()
  return convergence

class Job(object):
    """ Class containing data transmitted between MPI procs
    """
    def __init__(self,omegas,deltas,index):
        # input
        self.omegas = omegas
        self.deltas = deltas
        self.index = index
        # output
        self.partial_sum = 0.

    def execute(self):
        """
        [executed by worker processes]
        solve a given job and calculate "partial_sum"
        """
        results_product_1 = 1.
        results_product_2 = 1.
        results_product_1 = results_product_1 * np.prod(self.omegas[self.index])
        i2 = Index_no_repetition_2(self.index)
        while True:
          index2 = i2.get_i()
          tmp = np.absolute(self.deltas[index2[0]]-self.deltas[index2[1]])
          results_product_2 = results_product_2 * tmp**2
          if not i2.compute_next(): break
        self.partial_sum = self.partial_sum + results_product_1 * results_product_2

class Master(object):
    """
    MPI Process with rank #0
    """
    def __init__(self):
        self.slaves = list(range(1,siz))

    def start(self,k,n,omegas,deltas):
        """
        master loop
        """
        global rank
        if rank!=0:
            raise Exception("this routine should be called with MPI rank=0")
        if verb: print("[%d] starting master" % rank)

        ##-----------------------------------------

        for s in self.slaves:
            if verb: print("[%d] sending wake-up signal to worker %d" % (rank,s))
            comm.send(None, dest=s, tag=tags.WAKEUP)

        #################################################

        i1 = Index_no_repetition(k, n)
        results_sum = 0

        #################################################

        num_workers = siz-1
        closed_workers = 0
        while closed_workers<num_workers:
            # get a msg from any source
            data = comm.recv(source=mpi.ANY_SOURCE,
                             tag=mpi.ANY_TAG, status=status)
            source = status.Get_source()
            tag = status.Get_tag()
            if tag == tags.READY:
                if verb: print("[0] worker %d is ready" % source)

                #################################################

                if i1.get_is_valid():
                    job = Job(omegas,deltas,i1.get_i())

                    if verb: print("[0] sending job to %d" % source)
                    comm.send(job, dest=source, tag=tags.START)

                    # -------------------

                    i1.compute_next()
                else:

                    #################################################

                    if verb: print("[%d] exit %d" % (rank,source))
                    comm.send(None, dest=source, tag=tags.EXIT)

            elif tag==tags.DONE:
                if verb: print("[0] worker %d gives me its results" % source)

                #################################################
                results_sum = results_sum + data.partial_sum
                #################################################

            elif tag==tags.EXIT:
                closed_workers+=1
                if verb: print("[0] worker %d exited (%d worker(s) still running)" % (source, num_workers-closed_workers))

        ##-----------------------------------------

        self.killslaves()

        return results_sum

    def killslaves(self):
        global rank
        for s in self.slaves:
            if verb: print("[%d] sending suicide job to %d" % (rank,s))
            comm.send(None, dest=s, tag=tags.SUICIDE)


class Worker(object):
    """
    MPI Process with rank #1-#n
    """
    def __init__(self, rank):
        self.rank = rank

    def start(self):
        """
        worker loop
        """
        if self.rank==0:
            raise Exception("this routine should be called with MPI rank!=0")

        if verb: print("[%d] starting worker" % self.rank)
        while True:
            comm.recv(source=0, tag=mpi.ANY_TAG, status=status)
            tag = status.Get_tag()
            if tag==tags.WAKEUP:
                if verb: print("[%d] waking up" % self.rank)
                while True:
                    comm.send(None, dest=0, tag=tags.READY)
                    job = comm.recv(source=0, tag=mpi.ANY_TAG, status=status)
                    tag = status.Get_tag()
                    if tag==tags.START:
                        if verb: print("[%d] starting job" % self.rank)
                        job.execute()
                        if verb: print("[%d] sending job results" % self.rank)
                        comm.send(job, dest=0, tag=tags.DONE)
                    elif tag==tags.EXIT:
                        if verb: print("[%d] sending exit confirmation" % self.rank)
                        comm.send(None, dest=0, tag=tags.EXIT)
                        break
            elif tag==tags.SUICIDE:
                if verb: print("[%d] I'm dying..." % self.rank)
                break
