#!/usr/bin/env python
# -*- coding: latin-1; -*-

from builtins import str
from builtins import range
import numpy as np

def read_timers(filename,querylines):
    file = open(filename, "r")
    times = np.zeros((len(querylines),));
    for line in file:
        words = line.split()
        if len(words)==0:
            continue
        for i in range(0,len(querylines)):
            if times[i] == 0.:
                if words[0] == querylines[i]:
                    times[i] = float(words[2])
                    continue
    file.close()
    return times

def read_teuchos_timers(filename,querylines):
    file = open(filename, "r")
    times = np.zeros((len(querylines),));
    for line in file:
        words = line.split()
        if len(words)==0:
            continue
        for i in range(0,len(querylines)):
            if times[i] == 0.:
                querylines_words = querylines[i].split()
                has_to_be_read = True
                for j in range(0,len(querylines_words)):
                    if words[j] != querylines_words[j]:
                        has_to_be_read = False
                        break
                if has_to_be_read:
                    times[i] = float(words[len(querylines_words)])
                    continue
    file.close()
    return times

def read_Belos(filename,ensemble_size):
    active_set_number = 1

    number_of_criterion = 0

    previously_an_iter = False

    iter_indices = np.array([])

    residuals = np.array([])

    file = open(filename, "r")
    for line in file:
        words = line.split()
        if len(words)==0:
            continue
        if words[0] == "Iter":
            previously_an_iter = True
            word = words[1]
            len_word = len(word)
            if iter_indices.size!=0:
                iter_indices = np.append(iter_indices,[int(word[:(len_word-1)])], axis=0)
            else:
                iter_indices = np.array([int(word[:(len_word-1)])])

            if number_of_criterion == 0:
                word = words[3]
                len_word = len(word)

                number_of_criterion = int(word[:(len_word-1)])
            if number_of_criterion == 1:
                if ensemble_size > 1:

                    residual = np.array([float(words[6])])
                    for i in range(1,ensemble_size):
                        residual = np.append(residual,[float(words[6+i])])
                    if residuals.size!=0:
                        residuals = np.append(residuals,[residual], axis=0)
                    else:
                        residuals = np.array([residual])

                else:
                    if residuals.size!=0:
                        residuals = np.append(residuals,[float(words[5])], axis=0)
                    else:
                        residuals = np.array([float(words[5])])
        elif previously_an_iter:
            previously_an_iter = False
            active_set_number = active_set_number + 1

    file.close()
    return iter_indices, residuals

def plot_Belos(iter_indices, residuals,ensemble_size=1,hold=False,color_offset=1):
    import matplotlib.pyplot as plt
    
    col = {'c1':(0.,0.4470,0.7410),     'c2':(0.8500,0.3250,0.0980),
    'c3':(0.9290,0.6940,0.1250), 'c4':(0.4940,0.1840,0.5560),
    'c5':(0.4660,0.6740,0.1880), 'c6':(0.3010,0.7450,0.9330),
    'c7':(0.6350,0.0780,0.1840), 'c8':(0,0.4470,0.7410)}

    plt_linewidth = 2
    plt.style.use('ggplot')

    if hold==False:
        plt.figure(figsize=(8, 8))
        plt.grid(True)
    
    if ensemble_size > 1:
        for i in range(0,ensemble_size):
            plt.semilogy(iter_indices,residuals[:,i], '-', color = col['c'+str(color_offset+i)], zorder=2,linewidth=plt_linewidth)
    else:
        plt.semilogy(iter_indices,residuals, '-', color = col['c'+str(color_offset)], zorder=2,linewidth=plt_linewidth)
