#ifndef WALGEBRAIC_H
#define WALGEBRAIC_H

#include "katropton.h"
#include "Map.h"
#include "Domain.h"
#include "Loads.h"
#include "Graph.h"
#include "Matrices.h"
#include "Vectors.h"

#include "LinearSolver.h"

#include "EnsembleTraits.h"

#include <limits>

namespace katropton
{
namespace discrete_problem
{
template <typename scalar>
class Algebraic
{
public:
    Teuchos::RCP<Map> map;
    Teuchos::RCP<Graph> graph;
    Teuchos::RCP<Matrices<scalar>> matrices;
    Teuchos::RCP<Vectors<scalar>> vectors;

    Algebraic();
    Algebraic(Teuchos::RCP<Map> map, Teuchos::RCP<Domain<scalar>> domain, Teuchos::RCP<Loads<scalar>> loads);

    void writeMatrixToFile(std::string &name);

    void writeCrsMatrixToFile(Teuchos::RCP<Tpetra::CrsMatrix<scalar, local_ordinal_type, global_ordinal_type>> A, std::string &name, size_t myRank);

    void writeCrsMatrixToFile(std::string &name, size_t myRank);

    void writeLocalCrsMatrixToFile(std::string &name, size_t myRank);

    void writeVectorToFile(std::string &name, size_t myRank);

    void writeVectorToFile(std::string &name, size_t myRank, Teuchos::RCP<Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type>> b);

    Teuchos::RCP<Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type>> readVectorFromFile(std::string &name, size_t myRank);
};
}; // namespace discrete_problem
}; // namespace katropton

#endif //WALGEBRAIC_H
