namespace katropton
{
namespace discrete_problem
{
template <typename scalar>
void Algebraic<scalar>::writeMatrixToFile(std::string &name)
{
    using Teuchos::RCP;
    using Teuchos::rcp;

    std::ofstream outputfile;
    outputfile.open(name);

    RCP<crs_graph_type> graph_AOut = rcp(new crs_graph_type(map->mapOutput, 0));
    Tpetra::Export<> exportOut(map->mapDofs, map->mapOutput);

    graph_AOut->doExport(*(graph->A), exportOut, Tpetra::INSERT);
    graph_AOut->fillComplete();

    RCP<Tpetra::CrsMatrix<scalar, local_ordinal_type, global_ordinal_type>> AOut = rcp(new Tpetra::CrsMatrix<scalar, local_ordinal_type, global_ordinal_type>(graph_AOut));

    AOut->doExport(*(matrices->A), exportOut, Tpetra::ADD);
    AOut->fillComplete();

    auto line_size = map->mapOutput->getGlobalNumElements();

    for (global_ordinal_type global_index = 0; global_index < line_size; ++global_index)
    {
        local_ordinal_type local_index = map->mapOutput->getLocalElement(global_index);

        Teuchos::ArrayView<const local_ordinal_type> local_indices;
        Teuchos::ArrayView<const scalar> local_values;
        AOut->getLocalRowView(local_index, local_indices, local_values);
        //auto D_ii = Kokkos::subview(D_inv, i, Kokkos::ALL(), Kokkos::ALL());

        global_ordinal_type current_index = 0;
        for (auto k = 0; k < local_indices.size(); ++k)
        {
            while (true)
            {
                if (current_index < local_indices[k])
                    outputfile << (scalar)0 << " ";
                else if (current_index == local_indices[k])
                    outputfile << (scalar)local_values[k] << " ";
                else
                    break;
                ++current_index;
            }
        }
        while (true)
        {
            if (current_index < line_size)
                outputfile << (scalar)0 << " ";
            else
                break;
            ++current_index;
        }
        outputfile << std::endl;
    }

    outputfile.close();
}

template <typename scalar>
void Algebraic<scalar>::writeCrsMatrixToFile(Teuchos::RCP<Tpetra::CrsMatrix<scalar, local_ordinal_type, global_ordinal_type>> A, std::string &name, size_t myRank)
{
    using Teuchos::RCP;
    using Teuchos::rcp;

    std::ofstream outputfile;
    if (myRank == 0)
        outputfile.open(name);

    typedef std::numeric_limits<double> dbl;
    outputfile.precision(dbl::max_digits10 + 2);

    RCP<const Tpetra::Map<>> mapDofs = A->getRowMap();
    const size_t numIndices = mapDofs->getGlobalNumElements();
    const size_t numLclIndices = (myRank == 0) ? numIndices : 0;
    RCP<const Tpetra::Map<>> mapOutput = rcp(new Tpetra::Map<>(numIndices, numLclIndices, mapDofs->getIndexBase(), mapDofs->getComm()));

    //RCP<const crs_graph_type> graph_A = A->getGraph();
    //RCP<crs_graph_type> graph_AOut = rcp(new crs_graph_type (mapOutput, 0));
    Tpetra::Export<> exportOut(mapDofs, mapOutput);

    //graph_AOut->doExport(*(graph_A), exportOut, Tpetra::INSERT);
    //graph_AOut->fillComplete();

    RCP<Tpetra::CrsMatrix<scalar, local_ordinal_type, global_ordinal_type>> AOut = rcp(new Tpetra::CrsMatrix<scalar, local_ordinal_type, global_ordinal_type>(mapOutput, 0));

    AOut->doExport(*(A), exportOut, Tpetra::ADD);
    AOut->fillComplete();

    if (myRank == 0)
    {
        auto line_size = mapOutput->getGlobalNumElements();

        for (global_ordinal_type global_index = 0; global_index < line_size; ++global_index)
        {
            local_ordinal_type local_index = mapOutput->getLocalElement(global_index);

            Teuchos::ArrayView<const local_ordinal_type> local_indices;
            Teuchos::ArrayView<const scalar> local_values;
            AOut->getLocalRowView(local_index, local_indices, local_values);

            for (auto k = 0; k < local_indices.size(); ++k)
                outputfile << local_index << "   " << A->getColMap()->getGlobalElement(local_indices[k]) << "   " << (scalar)local_values[k] << std::endl;
        }
        outputfile.close();
    }
}

template <typename scalar>
void Algebraic<scalar>::writeCrsMatrixToFile(std::string &name, size_t myRank)
{
    using Teuchos::RCP;
    using Teuchos::rcp;

    std::ofstream outputfile;
    if (myRank == 0)
        outputfile.open(name);

    typedef std::numeric_limits<double> dbl;
    outputfile.precision(dbl::max_digits10 + 2);

    RCP<crs_graph_type> graph_AOut = rcp(new crs_graph_type(map->mapOutput, 0));
    Tpetra::Export<> exportOut(map->mapDofs, map->mapOutput);

    graph_AOut->doExport(*(graph->A), exportOut, Tpetra::INSERT);
    graph_AOut->fillComplete();

    RCP<Tpetra::CrsMatrix<scalar, local_ordinal_type, global_ordinal_type>> AOut = rcp(new Tpetra::CrsMatrix<scalar, local_ordinal_type, global_ordinal_type>(graph_AOut));

    AOut->doExport(*(matrices->A), exportOut, Tpetra::ADD);
    AOut->fillComplete();

    if (myRank == 0)
    {
        auto line_size = map->mapOutput->getGlobalNumElements();

        for (global_ordinal_type global_index = 0; global_index < line_size; ++global_index)
        {
            local_ordinal_type local_index = map->mapOutput->getLocalElement(global_index);

            Teuchos::ArrayView<const local_ordinal_type> local_indices;
            Teuchos::ArrayView<const scalar> local_values;
            AOut->getLocalRowView(local_index, local_indices, local_values);

            for (auto k = 0; k < local_indices.size(); ++k)
                outputfile << local_index << "   " << local_indices[k] << "   " << (scalar)local_values[k] << std::endl;
        }
        outputfile.close();
    }
}

template <typename scalar>
void Algebraic<scalar>::writeLocalCrsMatrixToFile(std::string &name, size_t myRank)
{
    std::ofstream outputfile;
    outputfile.open(name + "_" + std::to_string(myRank) + ".text");

    typedef std::numeric_limits<double> dbl;
    outputfile.precision(dbl::max_digits10 + 2);

    auto line_size = map->mapDofs->getNodeNumElements();

    for (local_ordinal_type local_index = 0; local_index < line_size; ++local_index)
    {
        global_ordinal_type global_index = map->mapDofs->getGlobalElement(local_index);

        Teuchos::ArrayView<const local_ordinal_type> local_indices;
        Teuchos::ArrayView<const scalar> local_values;
        matrices->A->getLocalRowView(local_index, local_indices, local_values);

        for (auto k = 0; k < local_indices.size(); ++k)
            outputfile << global_index << "   " << matrices->A->getColMap()->getGlobalElement(local_indices[k]) << "   " << (scalar)local_values[k] << std::endl;
    }
    outputfile.close();
}

template <typename scalar>
void Algebraic<scalar>::writeVectorToFile(std::string &name, size_t myRank)
{
    using Teuchos::RCP;
    using Teuchos::rcp;

    std::ofstream outputfile;
    if (myRank == 0)
        outputfile.open(name);

    RCP<Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type>> bOut = rcp(new Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type>(map->mapOutput, true));
    Tpetra::Export<> exportOut(map->mapDofs, map->mapOutput);

    bOut->doExport(*(vectors->b), exportOut, Tpetra::ADD);

    if (myRank == 0)
    {
        bOut->template sync<Kokkos::HostSpace>();
        auto b_2d = bOut->template getLocalView<Kokkos::HostSpace>();
        auto B = Kokkos::subview(b_2d, Kokkos::ALL(), 0);

        const int ensemble_size = Kokkos::dimension_scalar(B) == 0 ? 1 : Kokkos::dimension_scalar(B);
        typedef decltype(B) B_View;
        typename Kokkos::FlatArrayType<B_View>::type B_flat = B;

        auto line_size = map->mapOutput->getGlobalNumElements();

        for (global_ordinal_type global_index = 0; global_index < line_size; ++global_index)
        {
            for (auto j = 0; j < ensemble_size; ++j)
                outputfile << B_flat[ensemble_size * global_index + j] << " ";
            outputfile << std::endl;
        }
        outputfile.close();
    }
}

template <typename scalar>
void Algebraic<scalar>::writeVectorToFile(std::string &name, size_t myRank, Teuchos::RCP<Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type>> b)
{
    using Teuchos::RCP;
    using Teuchos::rcp;

    std::ofstream outputfile;
    if (myRank == 0)
        outputfile.open(name);

    typedef std::numeric_limits<double> dbl;
    outputfile.precision(dbl::max_digits10 + 2);

    RCP<const Tpetra::Map<>> mapDofs = b->getMap();
    const size_t numIndices = mapDofs->getGlobalNumElements();
    const size_t numLclIndices = (myRank == 0) ? numIndices : 0;
    RCP<const Tpetra::Map<>> mapOutput = rcp(new Tpetra::Map<>(numIndices, numLclIndices, mapDofs->getIndexBase(), mapDofs->getComm()));

    RCP<Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type>> bOut = rcp(new Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type>(mapOutput, true));
    Tpetra::Export<> exportOut(mapDofs, mapOutput);

    bOut->doExport(*b, exportOut, Tpetra::ADD);

    if (myRank == 0)
    {
        bOut->template sync<Kokkos::HostSpace>();
        auto b_2d = bOut->template getLocalView<Kokkos::HostSpace>();
        auto B = Kokkos::subview(b_2d, Kokkos::ALL(), 0);

        const int ensemble_size = Kokkos::dimension_scalar(B) == 0 ? 1 : Kokkos::dimension_scalar(B);
        typedef decltype(B) B_View;
        typename Kokkos::FlatArrayType<B_View>::type B_flat = B;

        auto line_size = mapOutput->getGlobalNumElements();

        for (global_ordinal_type global_index = 0; global_index < line_size; ++global_index)
        {
            for (auto j = 0; j < ensemble_size; ++j)
                outputfile << B_flat[ensemble_size * global_index + j] << " ";
            outputfile << std::endl;
        }
        outputfile.close();
    }
}

template <typename scalar>
Teuchos::RCP<Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type>> Algebraic<scalar>::readVectorFromFile(std::string &name, size_t myRank)
{
    using Teuchos::RCP;
    using Teuchos::rcp;

    std::ifstream inputfile;
    if (myRank == 0)
        inputfile.open(name);

    RCP<Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type>> x_0 = rcp(new Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type>(map->mapOutput, true));
    RCP<Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type>> x = rcp(new Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type>(map->mapDofs, true));

    if (myRank == 0)
    {
        auto line_size = map->mapOutput->getGlobalNumElements();

        typedef EnsembleTraits<scalar> ET;
        const int ensemble_size = ET::size;

        for (global_ordinal_type global_index = 0; global_index < line_size; ++global_index)
        {
            scalar tmp;
            for (auto j = 0; j < ensemble_size; ++j)
                inputfile >> ET::coeff(tmp, j);

            x_0->replaceGlobalValue(global_index, tmp);
        }
        inputfile.close();
    }

    Tpetra::Export<> exportDof(map->mapOutput, map->mapDofs);

    x->doExport(*x_0, exportDof, Tpetra::ADD);
    return x;
}

template <typename scalar>
Algebraic<scalar>::Algebraic()
{
}

template <typename scalar>
Algebraic<scalar>::Algebraic(Teuchos::RCP<Map> _map, Teuchos::RCP<Domain<scalar>> domain, Teuchos::RCP<Loads<scalar>> loads)
{
    map = _map;

    LinearSolver::getTimers()["graph"].start();
    graph = Teuchos::rcp(new Graph(_map, domain->elements_list));
    LinearSolver::getTimers()["graph"].stop();
    matrices = Teuchos::rcp(new Matrices<scalar>(graph));
    vectors = Teuchos::rcp(new Vectors<scalar>(map));
}

}; // namespace discrete_problem
}; // namespace katropton
