#include "DirectSolver.h"
#include "wTimer.h"
#include "wProblem.h"
#include "wTag.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"
#include "wMedium.h"
#include "wuDirichlet.h"
#include "wContact.h"
#include "wPt.h"
#include <map>
#include "wTag.h"
#include <math.h>
#include "wHex8.h"
#include "wQuad4.h"
#include "wTri3.h"
#include "wCacheHex8.h"
#include "wGaussHex8.h"
#include "wResults.h"

#include "RandomVariable.h"

#include "wElemMatrix.h"
#include "mortar.h"
#include "wMshConvert.h"

//#include "wPreconditioner.h"

#include "wDisplayHook.h"

#include <math.h>

#include <mpi.h>

//#include <tbb/tbb.h>
//#include <tbb/parallel_for.h>
//#include <tbb/blocked_range.h>
//#include <tbb/task_scheduler_init.h>
//#include <tbb/spin_mutex.h>

#include <cstring>
#include <sstream>
#include <iostream>

// Tpetra
#include "Stokhos_Tpetra_MP_Vector.hpp"
#include "Stokhos_Tpetra_Utilities_MP_Vector.hpp"
#include "Tpetra_ConfigDefs.hpp"

#include "Tpetra_Map.hpp"
#include "Tpetra_MultiVector.hpp"
#include "Tpetra_Vector.hpp"
#include "Tpetra_CrsGraph.hpp"
#include "Tpetra_CrsMatrix.hpp"
#include "Stokhos_Tpetra_CG.hpp"

#include <Teuchos_Array.hpp>
#include <Teuchos_GlobalMPISession.hpp>
#include <Teuchos_oblackholestream.hpp>
#include <Teuchos_ScalarTraits.hpp>
#include <Teuchos_RCP.hpp>
#include "Teuchos_ParameterList.hpp"

#include "Stokhos_Amesos2_MP_Vector.hpp"
#include "Amesos2_Factory.hpp"

#include "Stokhos_Ifpack2_MP_Vector.hpp"
#include "Ifpack2_Factory.hpp"

#include "Belos_Tpetra_MP_Vector.hpp"
#include "BelosLinearProblem.hpp"
#include "BelosPseudoBlockGmresSolMgr.hpp"
#include "BelosPseudoBlockCGSolMgr.hpp"
#include "BelosBlockGmresSolMgr.hpp"

#include "Kokkos_ViewFactory.hpp"
#include "Kokkos_Vector.hpp"

#include <Epetra_Map.h>
#include "Epetra_MpiComm.h"

#include "BelosConfigDefs.hpp"

#include "Map.h"
#include "Discrete_problem.h"

/*
#include "BelosTpetraAdapter.hpp"
#include "BelosConfigDefs.hpp"
#include "BelosLinearProblem.hpp"
#include "BelosPseudoBlockGmresSolMgr.hpp"
#include "BelosPseudoBlockCGSolMgr.hpp"
#include "BelosBlockGmresSolMgr.hpp"
*/

#ifdef max
#undef max // pour std::numeric_limits<std::streamsize>::max()
#endif

using namespace katropton;
using namespace fwk;

void DirectSolver::write(std::ostream &out) const
{
    out << "katropton::Solver:\n";
    out << pbl; // wProblem.h
}

void DirectSolver::start()
{
    //if(ensemble_propagation)
    //{
    /*
        if (minimum_ensemble_size <= 1)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,1,Kokkos::Threads>> ensemble_scalar;
            DirectSolver::tstart<ensemble_scalar>();
        }
        else if (minimum_ensemble_size <= 2)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,2,Kokkos::Threads>> ensemble_scalar;
            DirectSolver::tstart<ensemble_scalar>();
        }
        */
    //else
    /*
        if (minimum_ensemble_size <= 4)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,4,Kokkos::Threads>> ensemble_scalar;
            DirectSolver::tstart<ensemble_scalar>();
        }
        else if (minimum_ensemble_size <= 8)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,8,Kokkos::Threads>> ensemble_scalar;
            DirectSolver::tstart<ensemble_scalar>();
        }
        else if (minimum_ensemble_size <= 16)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,16,Kokkos::Threads>> ensemble_scalar;
            DirectSolver::tstart<ensemble_scalar>();
        }
        else if (minimum_ensemble_size <= 32)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,32,Kokkos::Threads>> ensemble_scalar;
            DirectSolver::tstart<ensemble_scalar>();
        }
        */
    /*
        else if (minimum_ensemble_size <= 64)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,64,Kokkos::Threads>> ensemble_scalar;
            DirectSolver::tstart<ensemble_scalar>();
        }
        */
    //}

    //else
    DirectSolver::tstart<double>();
}

template <typename scalar>
void DirectSolver::tstart()
{
    // ------------------------------------------
    //
    //         Trilinos initialization
    //
    // ------------------------------------------
    using std::cerr;
    using std::cout;
    using std::endl;
    using Teuchos::arcp;
    using Teuchos::Array;
    using Teuchos::ArrayRCP;
    using Teuchos::ArrayView;
    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;

    using Kokkos::ALL;
    using Kokkos::subview;

    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    typedef Tpetra::MultiVector<scalar, local_ordinal_type, global_ordinal_type> multivector_type;
    typedef Tpetra::CrsMatrix<scalar, local_ordinal_type, global_ordinal_type> crs_matrix_type;

    typedef Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type> vector_type;
    typedef Tpetra::Export<> export_type;
    typedef Amesos2::Solver<crs_matrix_type, multivector_type> solver_type;

    Teuchos::oblackholestream blackhole;

    RCP<const Teuchos::Comm<int>> comm = rcp(new Teuchos::MpiComm<int>(pbl.comm));

    const size_t myRank = comm->getRank();

    // Make an output stream (for verbose output) that only prints on
    // Proc 0 of the communicator.
    Teuchos::oblackholestream blackHole;
    std::ostream &out = (myRank == 0) ? std::cout : blackHole;

    // Initiate Kokkos with the given number  of threads
    Kokkos::InitArguments args;
    args.num_threads = nthreads;

    Kokkos::initialize(args);
    {
        LinearSolver::getTimers()["total"].start();
        auto msh = pbl.msh;

        // assign a global row for each global node
        for (auto n : msh->nodes)
            n->row = n->no - 1;

        auto numNodes = msh->nodes.size();

        RCP<Teuchos::ParameterList> randomParams = solverList->get("randomParams", Teuchos::null);
        const int n_rv = pbl.RandomVariables.size();
        scalar m_rv_values[n_rv];
        for (int i = 0; i < n_rv; ++i)
        {
            typedef EnsembleTraits<scalar> ET;
            const int ensemble_size = ET::size;

            for (int j = 0; j < ensemble_size; ++j)
                ET::coeff(m_rv_values[i], j) = (double)pbl.RandomVariables[i]->values[j];
        }

        LinearSolver::getTimers()["dp"].start();
        Teuchos::RCP<Discrete_problem<scalar>> discrete_problem = Teuchos::rcp(new Discrete_problem<scalar>(pbl, numPrimalDPN, randomParams, m_rv_values));
        LinearSolver::getTimers()["dp"].stop();

        LinearSolver::getTimers()["computeMatrices"].start();
        discrete_problem->computeMatrices();

        LinearSolver::getTimers()["computeMatrices"].stop();
        LinearSolver::getTimers()["computeLoads"].start();
        discrete_problem->computeLoads();
        LinearSolver::getTimers()["computeLoads"].stop();

        std::string name_1, name_2, name_3, name_x;
        name_1 = "matrix_after_bc.text";
        name_2 = "matrix_after_bc";
        name_3 = "vector_after_bc.text";
        name_x = "solution";
        /*
        discrete_problem.algebraic.writeCrsMatrixToFile(name_1,myRank);
        discrete_problem.algebraic.writeLocalCrsMatrixToFile(name_2,myRank);
        discrete_problem.algebraic.writeVectorToFile(name_3,myRank);
*/
        const export_type exportOutput(discrete_problem->algebraic->map->mapDofs, discrete_problem->algebraic->map->mapOutput);

        RCP<vector_type> x(new vector_type(discrete_problem->algebraic->map->mapDofs, false));
        RCP<vector_type> xPost(new vector_type(discrete_problem->algebraic->map->mapOutput, false));

        LinearSolver::getTimers()["Amesos2"].start();
        RCP<solver_type> solverA;
        solverA = Amesos2::create<crs_matrix_type, multivector_type>("KLU2", discrete_problem->algebraic->matrices->A, x, discrete_problem->algebraic->vectors->b);
        solverA->solve();
        LinearSolver::getTimers()["Amesos2"].stop();

        xPost->doExport(*x, exportOutput, Tpetra::INSERT);

        if (myRank == 0)
            this->writeResults<scalar>(xPost, numNodes, name_x, 0, true, false, discrete_problem->algebraic->map, discrete_problem->domain->random_field);

        // ------------------------------------------
        //
        //        End of the timers and print
        //
        // ------------------------------------------
        LinearSolver::getTimers()["total"].stop();

        //tms = LinearSolver::getTimers();
        out << "\n---CPU statistics of proc " << myRank << "---\n";
        out << LinearSolver::getTimers();
        out << "---------------------------\n";
    }
    Kokkos::finalize();
}
