namespace katropton
{
namespace discrete_problem
{
template <typename scalar>
Dirichlet_list<scalar>::Dirichlet_list()
{
    dirichlet_number = 0;
    dirichlet_size = 0;
    dirichlet_nodes = 0;

    dirichlet_time = 0;
    dirichlet_dofs = 0;
    dirichlet_values = 0;

    dirichlet_nodes_list = Kokkos::View<int **, Kokkos::LayoutRight>("R", dirichlet_number, 0);
    dirichlet_values_list = Kokkos::View<scalar ***, Kokkos::LayoutRight>("R", dirichlet_number, 0, 0);
}

template <typename scalar>
Dirichlet_list<scalar>::Dirichlet_list(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<Elements_list> elements_list)
{
    //std::cout << "dbc 0" << std::endl;
    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    typedef EnsembleTraits<scalar> ET;
    const int ensemble_size = ET::size;

    local_ordinal_type numDPN = map->numPrimalDPN;
    dirichlet_number = pbl.duBCs.size();
    local_ordinal_type maxNumNodesPerBC = 0;

    std::vector<std::vector<global_ordinal_type>> bcNodes = {};

    //std::cout << "dbc 1" << std::endl;

    for (auto i = 0; i < dirichlet_number; ++i)
    {

        /*
                std::cout << "DC 0 elem size: "<<pbl.duBCs[i]->tag->elems.size() << std::endl;
                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[0]->nodes[0]->no << std::endl;
                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[0]->nodes[1]->no << std::endl;
                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[0]->nodes[2]->no << std::endl;
                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[0]->nodes[3]->no << std::endl;

                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[0]->nodes[0]->row << std::endl;
                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[0]->nodes[1]->row << std::endl;
                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[0]->nodes[2]->row << std::endl;
                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[0]->nodes[3]->row << std::endl;


                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[3]->nodes[0]->no << std::endl;
                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[3]->nodes[1]->no << std::endl;
                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[3]->nodes[2]->no << std::endl;
                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[3]->nodes[3]->no << std::endl;

                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[3]->nodes[0]->row << std::endl;
                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[3]->nodes[1]->row << std::endl;
                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[3]->nodes[2]->row << std::endl;
                std::cout << "DC 0_0: "<< pbl.duBCs[i]->tag->elems[3]->nodes[3]->row << std::endl;
                */
        std::vector<global_ordinal_type> tmpMyBcNodes = {};
        std::vector<global_ordinal_type> myBcNodes = {};
        //std::cout << "   dbc 1.1" << std::endl;

        for (auto j = 0; j < pbl.duBCs[i]->nodes.size(); ++j)
            tmpMyBcNodes.push_back(pbl.duBCs[i]->nodes[j]->row);

        //std::cout << "   dbc 1.2" << std::endl;
        std::sort(tmpMyBcNodes.begin(), tmpMyBcNodes.end());
        //std::cout << "   dbc 1.3" << std::endl;

        if (tmpMyBcNodes.size() >= 1)
            myBcNodes.push_back(tmpMyBcNodes[0]);
        //std::cout << "   dbc 1.4" << std::endl;

        for (auto ii = 1; ii < tmpMyBcNodes.size(); ++ii)
            if (tmpMyBcNodes[ii] != tmpMyBcNodes[ii - 1])
                myBcNodes.push_back(tmpMyBcNodes[ii]);

        //std::cout << "   dbc 1.5" << std::endl;
        /*
                for (auto j = 0; j < pbl.duBCs[i]->tag->elems.size(); ++j)
                {
                    local_ordinal_type e = map->mapElems->getLocalElement(pbl.duBCs[i]->tag->elems[j]->no);
                    if      (elements_list->Get_Element_Type(e) == EL_QUAD4);
                    else if (elements_list->Get_Element_Type(e) == EL_TRI3 );
                    else if (elements_list->Get_Element_Type(e) == EL_LINE2);
                    else if (elements_list->Get_Element_Type(e) == EL_POINT1);
                    else
                        continue;
                    for ( auto n1 = 0; n1 < elements_list->Get_Element_Size(e); ++n1)
                    {
                        global_ordinal_type i_global = pbl.duBCs[i]->tag->elems[j]->nodes[n1]->row; // elements_list->Get_Element_Node(e,n1); //pbl.duBCs[i]->tag->elems[j]->nodes[n1]->row;
                        //global_ordinal_type i_global = map->mapNodesWO->getGlobalElement(elements_list->Get_Element_Node(e,n1));
                        if(std::find(myBcNodes.begin(),myBcNodes.end(),i_global)==myBcNodes.end())
                            {
                                myBcNodes.push_back(i_global);
                                //std::cout << "DC 0: "<< i << " elem: "<< pbl.duBCs[i]->tag->elems[j]->no << " " << e <<" node DC: " << pbl.duBCs[i]->tag->elems[j]->nodes[n1]->row << " " << i_global << " " << elements_list->Get_Element_Node(e,n1)<< std::endl;
                            }
                    }
                }
                */
        //std::cout << "   dbc 1.6" << std::endl;

        bcNodes.push_back(myBcNodes);
        if (maxNumNodesPerBC < myBcNodes.size())
            maxNumNodesPerBC = myBcNodes.size();

        //std::cout << "   dbc 1.7" << std::endl;
    }

    //std::cout << "dbc 2" << std::endl;

    dirichlet_size = 0;
    dirichlet_nodes = 1;
    dirichlet_nodes_list = Kokkos::View<int **, Kokkos::LayoutRight>("R", dirichlet_number, dirichlet_nodes + maxNumNodesPerBC);

    for (auto i = 0; i < dirichlet_number; ++i)
    {
        dirichlet_nodes_list(i, dirichlet_size) = bcNodes[i].size();
        for (auto j = 0; j < bcNodes[i].size(); ++j)
        {
            dirichlet_nodes_list(i, dirichlet_nodes + j) = bcNodes[i][j];
            //std::cout << "DC 1: "<< i << " node DC: " << bcNodes[i][j] << std::endl;
        }
    }

    //std::cout << "dbc 3" << std::endl;

    dirichlet_time = 0;
    dirichlet_dofs = 1;
    dirichlet_values = 1 + numDPN;
    dirichlet_values_list = Kokkos::View<scalar ***, Kokkos::LayoutRight>("R", dirichlet_number, 1, 1 + 2 * numDPN);
    for (auto i = 0; i < dirichlet_number; ++i)
    {
        dirichlet_values_list(i, 0, dirichlet_time) = 0.; //to change latter
        for (auto j = 0; j < numDPN; ++j)
            if (pbl.duBCs[i]->which_dof[j])
                dirichlet_values_list(i, 0, dirichlet_dofs + j) = 1;
            else
                dirichlet_values_list(i, 0, dirichlet_dofs + j) = 0;

        for (int j = 0; j < ensemble_size; ++j)
        {
            ET::coeff(dirichlet_values_list(i, 0, dirichlet_values + 0), j) = pbl.duBCs[i]->x_values[j];
            ET::coeff(dirichlet_values_list(i, 0, dirichlet_values + 1), j) = pbl.duBCs[i]->y_values[j];
            ET::coeff(dirichlet_values_list(i, 0, dirichlet_values + 2), j) = pbl.duBCs[i]->z_values[j];
            if (numDPN == 4)
                ET::coeff(dirichlet_values_list(i, 0, dirichlet_values + 3), j) = pbl.duBCs[i]->T_values[j];
        }
    }

    //std::cout << "dbc 4" << std::endl;
}
}; // namespace discrete_problem
}; // namespace katropton
