#include <Xpetra_MatrixFactory.hpp>
#include <Xpetra_Matrix.hpp>
#include <Xpetra_MatrixMatrix.hpp>
#include <Xpetra_CrsMatrixWrap.hpp>
#include <Xpetra_BlockedCrsMatrix.hpp>
#include <Xpetra_CrsMatrix.hpp>
#include <MueLu_Utilities_decl.hpp>

namespace katropton
{
template <typename scalar>
bool Discrete_problem<scalar>::updateActivity(int active_set_iteration)
{
    using Teuchos::Array;
    using Teuchos::RCP;
    using Teuchos::rcp;

    const size_t myRank = algebraic->map->mapDofs->getComm()->getRank();

    // First, get the displacement solution,

    RCP<tpetra_mvector_type> mx = (Teuchos::rcp_dynamic_cast<xpetra_tmvector_type>(algebraic->vectors->solutionMultiVector))->getTpetra_MultiVector();
    RCP<tpetra_vector_type> xl = mx->getVectorNonConst(0);

    RCP<tpetra_vector_type> x = rcp(new tpetra_vector_type(algebraic->map->mapDofs, false));
    //RCP<tpetra_vector_type> l = rcp (new tpetra_vector_type(algebraic->map->mapLagrangeDofs, false));

    RCP<xpetra_bmvector_type> blockedSol =
        rcp(new xpetra_bmvector_type(algebraic->map->blockedMap, algebraic->vectors->solutionMultiVector));

    RCP<tpetra_vector_type> l = (Teuchos::rcp_dynamic_cast<xpetra_tmvector_type>(blockedSol->getMultiVector(1, true)))->getTpetra_MultiVector()->getVectorNonConst(0);

    RCP<tpetra_vector_type> gap = rcp(new tpetra_vector_type(*algebraic->vectors->initialGap, Teuchos::Copy));

    //rcp (new tpetra_vector_type(algebraic->map->mapLagrangeDofs, true));

    RCP<export_type> exportx = rcp(new export_type(algebraic->map->fullmap, algebraic->map->mapDofs));
    //RCP<export_type> exportl = rcp (new export_type(algebraic->map->fullmap,algebraic->map->mapLagrangeDofs));

    x->doExport(*xl, *exportx, Tpetra::INSERT);
    //l->doExport(*xl, *exportl, Tpetra::INSERT);

    //std::string name_1 = "x_sol_test_0.text";
    //algebraic->writeVectorToFile(name_1,myRank,xl);

    //std::string name_2 = "x_sol_test_1.text";
    //algebraic->writeVectorToFile(name_2,myRank,xx);

    // Multiply it with B_G and remove initial gap

    // gap = 1 * B_G * x - 1 * gap

    algebraic->matrices->B_G->apply(*x, *gap, Teuchos::NO_TRANS, ((scalar)1.), ((scalar)-1.));

    //std::string name_1 = "gap_0 "+std::to_string(active_set_iteration)+".text";
    //algebraic->writeVectorToFile(name_1,myRank,gap);

    //std::string name_2 = "l_0 "+std::to_string(active_set_iteration)+".text";
    //algebraic->writeVectorToFile(name_2,myRank,l);

    // sum lambda + c * the previously computed vector
    double c = 1.;
    l->update((scalar)c, *gap, (scalar)1.); // l = l + c*gap

    //std::string name_3 = "l_1 "+std::to_string(active_set_iteration)+".text";
    //algebraic->writeVectorToFile(name_3,myRank,l);

    // loop on the local lagrange and create the new activity

    l->template sync<Kokkos::HostSpace>();
    auto l_2d = l->template getLocalView<Kokkos::HostSpace>();

    //std::cout << "-----------------" << std::endl;
    //std::cout << "l_2d" << std::endl;
    //for(auto local_index=0; local_index<algebraic->map->mapLagrangeDofs->getNodeNumElements(); ++local_index)
    //    std::cout << l_2d(local_index,0) << std::endl;

    double hasConverged = 1.;
    double hasConverged_old = 0.;
    if (active_set_iteration > 1)
        hasConverged_old = 1.;

    //std::string name_C_1 = "C before correction "+std::to_string(active_set_iteration)+".txt";
    //algebraic->writeCrsMatrixToFile(algebraic->matrices->C, name_C_1, algebraic->map->mapDofs->getComm()->getRank());

    algebraic->matrices->B_2->resumeFill();
    algebraic->matrices->C->resumeFill();
    algebraic->matrices->Cb->resumeFill();

    const size_t numDPN = algebraic->map->numPrimalDPN;
    //const size_t numLMPN = (numDPN==4) ? 2: 1;

    size_t numLMPN;
    if (domain->contacts_list->hasAtLeastOneSticking() && numDPN == 4)
        numLMPN = 4;
    else if (domain->contacts_list->hasAtLeastOneSticking())
        numLMPN = 3;
    else if (numDPN == 4)
        numLMPN = 2;
    else
        numLMPN = 1;

    const int numMyLagrangeNum = algebraic->map->mapLagrangeDofs->getNodeNumElements();

    //const int numMyMechLagrangeNum = (numDPN==4) ? numMyLagrangeNum/2: numMyLagrangeNum;
    const int numMyMechLagrangeNum = numMyLagrangeNum / numLMPN;

    for (auto local_index = 0; local_index < numMyMechLagrangeNum; ++local_index)
    {
        global_ordinal_type lagrange_id = algebraic->map->mapLagrangeDofs->getGlobalElement(local_index);
        global_ordinal_type node_id = algebraic->map->lm_to_dof[lagrange_id];
        lagrange_id = lagrange_id * numLMPN;
        local_ordinal_type interface_id = domain->contacts_list->getInterfaceOfSlaveNode(node_id);

        if (domain->contacts_list->isNotUpdated(interface_id))
            continue;

        scalar old_mask = old_activity(local_index);
        scalar old_old_mask = old_old_activity(local_index);
        Mask<scalar> current_mask = (l_2d(local_index * numLMPN, 0) >= 0.);

        scalar oneMask, oneNotMask;
        mask_assign<scalar>(current_mask, oneMask) = {1., 0.};
        mask_assign<scalar>(current_mask, oneNotMask) = {0., 1.};

        //std::cout << l_2d(local_index,0) << std::endl;
        //std::cout << oneMask << std::endl;
        //std::cout << old_mask << std::endl;
        //std::cout << (oneMask == old_mask) << std::endl;
        //std::cout << (oneMask != old_mask) << std::endl;

        if (active_set_iteration > 1)
            if (!MaskLogic::AND(oneMask == old_old_mask))
                hasConverged_old = 0.;

        if (!MaskLogic::AND(oneMask == old_mask))
        {
            global_ordinal_type lagrange_id = algebraic->map->mapLagrangeDofs->getGlobalElement(local_index);

            lagrange_id = lagrange_id * numLMPN;

            hasConverged = 0.;
            // If the current Lagrange multiplier is active for every sample of the current ensemble, we can do the update without using the mask entries.
            if (MaskLogic::AND(current_mask))
            {
                Teuchos::ArrayView<const local_ordinal_type> local_indices;
                Teuchos::ArrayView<const scalar> values;

                if (domain->contacts_list->isSticking(interface_id))
                {
                    for (size_t j = 0; j < numLMPN; ++j)
                    {
                        algebraic->matrices->B->getLocalRowView(local_index * numLMPN + j, local_indices, values);

                        for (auto k = 0; k < local_indices.size(); ++k)
                        {
                            global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                            algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + j,
                                                                          Teuchos::tuple<global_ordinal_type>(column_index),
                                                                          Teuchos::tuple<scalar>(values[k]));
                        }

                        algebraic->matrices->C->replaceGlobalValues(lagrange_id + j,
                                                                    Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                    Teuchos::tuple<scalar>((scalar)0.));

                        algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + j,
                                                                     Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                     Teuchos::tuple<scalar>((scalar)1.));
                    }
                }
                else
                {
                    algebraic->matrices->B->getLocalRowView(local_index * numLMPN, local_indices, values);

                    for (auto k = 0; k < local_indices.size(); ++k)
                    {
                        global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                        algebraic->matrices->B_2->replaceGlobalValues(lagrange_id,
                                                                      Teuchos::tuple<global_ordinal_type>(column_index),
                                                                      Teuchos::tuple<scalar>(values[k]));
                    }

                    algebraic->matrices->C->replaceGlobalValues(lagrange_id,
                                                                Teuchos::tuple<global_ordinal_type>(lagrange_id),
                                                                Teuchos::tuple<scalar>((scalar)0.));

                    algebraic->matrices->Cb->replaceGlobalValues(lagrange_id,
                                                                 Teuchos::tuple<global_ordinal_type>(lagrange_id),
                                                                 Teuchos::tuple<scalar>((scalar)1.));

                    if (numLMPN == 2 || numLMPN == 4)
                    {
                        algebraic->matrices->B->getLocalRowView(local_index * numLMPN + numLMPN - 1, local_indices, values);

                        for (auto k = 0; k < local_indices.size(); ++k)
                        {
                            global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                            algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                          Teuchos::tuple<global_ordinal_type>(column_index),
                                                                          Teuchos::tuple<scalar>(values[k]));
                        }

                        algebraic->matrices->C->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                    Teuchos::tuple<global_ordinal_type>(lagrange_id + numLMPN - 1),
                                                                    Teuchos::tuple<scalar>((scalar)0.));

                        algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                     Teuchos::tuple<global_ordinal_type>(lagrange_id + numLMPN - 1),
                                                                     Teuchos::tuple<scalar>((scalar)1.));
                    }
                }
            }
            // If the current Lagrange multiplier is inactive for every sample of the current ensemble, we can do the update without using the mask entries.
            else if (!MaskLogic::AND(current_mask))
            {
                Teuchos::ArrayView<const local_ordinal_type> local_indices;
                Teuchos::ArrayView<const scalar> values;

                if (domain->contacts_list->isSticking(interface_id))
                {
                    for (size_t j = 0; j < numLMPN; ++j)
                    {
                        algebraic->matrices->B->getLocalRowView(local_index * numLMPN + j, local_indices, values);

                        for (auto k = 0; k < local_indices.size(); ++k)
                        {
                            global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                            algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + j,
                                                                          Teuchos::tuple<global_ordinal_type>(column_index),
                                                                          Teuchos::tuple<scalar>((scalar)0.));
                        }

                        algebraic->matrices->C->replaceGlobalValues(lagrange_id + j,
                                                                    Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                    Teuchos::tuple<scalar>((scalar)1.));

                        algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + j,
                                                                     Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                     Teuchos::tuple<scalar>((scalar)0.));
                    }
                }
                else
                {
                    algebraic->matrices->B->getLocalRowView(local_index * numLMPN, local_indices, values);

                    for (auto k = 0; k < local_indices.size(); ++k)
                    {
                        global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                        algebraic->matrices->B_2->replaceGlobalValues(lagrange_id,
                                                                      Teuchos::tuple<global_ordinal_type>(column_index),
                                                                      Teuchos::tuple<scalar>((scalar)0.));
                    }

                    algebraic->matrices->C->replaceGlobalValues(lagrange_id,
                                                                Teuchos::tuple<global_ordinal_type>(lagrange_id),
                                                                Teuchos::tuple<scalar>((scalar)1.));

                    algebraic->matrices->Cb->replaceGlobalValues(lagrange_id,
                                                                 Teuchos::tuple<global_ordinal_type>(lagrange_id),
                                                                 Teuchos::tuple<scalar>((scalar)0.));

                    if (numLMPN == 2 || numLMPN == 4)
                    {
                        algebraic->matrices->B->getLocalRowView(local_index * numLMPN + numLMPN - 1, local_indices, values);

                        for (auto k = 0; k < local_indices.size(); ++k)
                        {
                            global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                            algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                          Teuchos::tuple<global_ordinal_type>(column_index),
                                                                          Teuchos::tuple<scalar>((scalar)0.));
                        }

                        algebraic->matrices->C->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                    Teuchos::tuple<global_ordinal_type>(lagrange_id + numLMPN - 1),
                                                                    Teuchos::tuple<scalar>((scalar)1.));

                        algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                     Teuchos::tuple<global_ordinal_type>(lagrange_id + numLMPN - 1),
                                                                     Teuchos::tuple<scalar>((scalar)0.));
                    }
                }
            }
            // Else, if the activity of the current Lagrange multiplier depends on the samples, we use the mask to do the update.
            else
            {

                Teuchos::ArrayView<const local_ordinal_type> local_indices;
                Teuchos::ArrayView<const scalar> values;

                if (domain->contacts_list->isSticking(interface_id))
                {
                    for (size_t j = 0; j < numLMPN; ++j)
                    {
                        algebraic->matrices->B->getLocalRowView(local_index * numLMPN + j, local_indices, values);

                        for (auto k = 0; k < local_indices.size(); ++k)
                        {
                            global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                            algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + j,
                                                                          Teuchos::tuple<global_ordinal_type>(column_index),
                                                                          Teuchos::tuple<scalar>(oneMask * values[k]));
                        }

                        algebraic->matrices->C->replaceGlobalValues(lagrange_id + j,
                                                                    Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                    Teuchos::tuple<scalar>(oneNotMask));

                        algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + j,
                                                                     Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                     Teuchos::tuple<scalar>(oneMask));
                    }
                }
                else
                {
                    algebraic->matrices->B->getLocalRowView(local_index * numLMPN, local_indices, values);

                    for (auto k = 0; k < local_indices.size(); ++k)
                    {
                        global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                        algebraic->matrices->B_2->replaceGlobalValues(lagrange_id,
                                                                      Teuchos::tuple<global_ordinal_type>(column_index),
                                                                      Teuchos::tuple<scalar>(oneMask * values[k]));
                    }

                    algebraic->matrices->C->replaceGlobalValues(lagrange_id,
                                                                Teuchos::tuple<global_ordinal_type>(lagrange_id),
                                                                Teuchos::tuple<scalar>(oneNotMask));

                    algebraic->matrices->Cb->replaceGlobalValues(lagrange_id,
                                                                 Teuchos::tuple<global_ordinal_type>(lagrange_id),
                                                                 Teuchos::tuple<scalar>(oneMask));

                    if (numLMPN == 2 || numLMPN == 4)
                    {
                        algebraic->matrices->B->getLocalRowView(local_index * numLMPN + numLMPN - 1, local_indices, values);

                        for (auto k = 0; k < local_indices.size(); ++k)
                        {
                            global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                            algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                          Teuchos::tuple<global_ordinal_type>(column_index),
                                                                          Teuchos::tuple<scalar>(oneMask * values[k]));
                        }

                        algebraic->matrices->C->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                    Teuchos::tuple<global_ordinal_type>(lagrange_id + numLMPN - 1),
                                                                    Teuchos::tuple<scalar>(oneNotMask));

                        algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                     Teuchos::tuple<global_ordinal_type>(lagrange_id + numLMPN - 1),
                                                                     Teuchos::tuple<scalar>(oneMask));
                    }
                }
            }
        }

        old_old_activity(local_index) = old_activity(local_index);
        old_activity(local_index) = oneMask;
    }
    //std::cout << "-----------------" << std::endl;

    algebraic->matrices->B_2->fillComplete(algebraic->map->mapDofs, algebraic->map->mapLagrangeDofs);
    algebraic->matrices->C->fillComplete();
    algebraic->matrices->Cb->fillComplete();

    //algebraic->matrices->xB = rcp (new xpetra_tcrs_type(algebraic->matrices->B_2));
    //algebraic->matrices->xC = rcp (new xpetra_tcrs_type(algebraic->matrices->C));
    //algebraic->matrices->xwB = rcp (new xpetra_wcrs_type(algebraic->matrices->xB));
    //algebraic->matrices->xwC = rcp (new xpetra_wcrs_type(algebraic->matrices->xC));

    /*
        RCP<xpetra_crs_type > xB =
        rcp (new xpetra_tcrs_type(algebraic->matrices->B_2));
        RCP<xpetra_crs_type > xC =
            rcp (new xpetra_tcrs_type(algebraic->matrices->C));

        RCP<xpetra_wcrs_type > xwB =
            rcp (new xpetra_wcrs_type(xB));


        RCP<xpetra_wcrs_type > xwC =
            rcp (new xpetra_wcrs_type(xC));
*/

    //algebraic->matrices->blockedMatrix->resumeFill();

    //algebraic->matrices->blockedMatrix->setMatrix(1,0,algebraic->matrices->xwB);
    //algebraic->matrices->blockedMatrix->setMatrix(1,1,algebraic->matrices->xwC);

    //algebraic->matrices->blockedMatrix->fillComplete();

    /*
        algebraic->matrices->blockedMatrix =
            rcp (new xpetra_bcrs_type(algebraic->map->blockedMap,algebraic->map->blockedMap,8));


        RCP<xpetra_crs_type > xA =
            rcp (new xpetra_tcrs_type(algebraic->matrices->A));
        RCP<xpetra_crs_type > xB =
            rcp (new xpetra_tcrs_type(algebraic->matrices->B_2));
        RCP<xpetra_crs_type > xB_T =
            rcp (new xpetra_tcrs_type(algebraic->matrices->B_T));
        RCP<xpetra_crs_type > xC =
            rcp (new xpetra_tcrs_type(algebraic->matrices->C));


        RCP<xpetra_wcrs_type > xwA =
            rcp (new xpetra_wcrs_type(xA));

        RCP<xpetra_wcrs_type > xwB =
            rcp (new xpetra_wcrs_type(xB));

        RCP<xpetra_wcrs_type > xwB_T =
            rcp (new xpetra_wcrs_type(xB_T));

        RCP<xpetra_wcrs_type > xwC =
            rcp (new xpetra_wcrs_type(xC));

        algebraic->matrices->blockedMatrix->setMatrix(0,0,xwA);
        algebraic->matrices->blockedMatrix->setMatrix(0,1,xwB_T);
        algebraic->matrices->blockedMatrix->setMatrix(1,0,xwB);
        algebraic->matrices->blockedMatrix->setMatrix(1,1,xwC);

        algebraic->matrices->blockedMatrix->fillComplete();
        */

    //std::string name_C_2 = "C after correction "+std::to_string(active_set_iteration)+".txt";
    //algebraic->writeCrsMatrixToFile(algebraic->matrices->C, name_C_2, algebraic->map->mapDofs->getComm()->getRank());

    RCP<tpetra_vector_type> initialActiveGap(new tpetra_vector_type(algebraic->map->mapLagrangeDofs, true));
    algebraic->matrices->Cb->apply(*algebraic->vectors->initialGap, *initialActiveGap);

    RCP<xpetra_mvector_type> xg =
        rcp(new xpetra_tmvector_type(initialActiveGap));
    algebraic->vectors->rhsBlockedMultiVector->setMultiVector(1, xg, true);
    algebraic->vectors->rhsMultiVector = algebraic->vectors->rhsBlockedMultiVector->Merge();

    //size_t myRank = algebraic->map->mapDofs->getComm()->getRank();
    size_t numtasks = algebraic->map->mapDofs->getComm()->getSize();
    if (myRank == 0)
    {

        double globalHasConverged = hasConverged;
        double globalHasConverged_old = hasConverged_old;
        for (size_t i = 1; i < numtasks; ++i)
        {
            MPI_Recv(&hasConverged, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            if (hasConverged == 0.)
                globalHasConverged = 0.;

            MPI_Recv(&hasConverged_old, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            if (hasConverged_old == 0.)
                globalHasConverged_old = 0.;
        }
        hasConverged = globalHasConverged;
        hasConverged_old = globalHasConverged_old;
        for (size_t i = 1; i < numtasks; ++i)
        {
            MPI_Send(&hasConverged, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD);
            MPI_Send(&hasConverged_old, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD);
        }
    }
    else
    {
        MPI_Send(&hasConverged, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
        MPI_Send(&hasConverged_old, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
        MPI_Recv(&hasConverged, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        MPI_Recv(&hasConverged_old, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    }

    if (hasConverged_old == 1.)
    {
        if (myRank == 0)
        {
            std::cout << "-----------------------" << std::endl;
            std::cout << "---Has not converged---" << std::endl;
            std::cout << "-----------------------" << std::endl;
        }
        return true;
    }

    //std::string name_A = "A " +std::to_string(active_set_iteration)+".txt";
    //std::string name_B = "B" +std::to_string(active_set_iteration)+".txt";
    //std::string name_B_T = "B_T" +std::to_string(active_set_iteration)+".txt";
    //std::string name_C = "C" +std::to_string(active_set_iteration)+".txt";
    //algebraic->writeCrsMatrixToFile(algebraic->matrices->A, name_A, algebraic->map->mapDofs->getComm()->getRank());
    //algebraic->writeCrsMatrixToFile(algebraic->matrices->B_2, name_B, algebraic->map->mapDofs->getComm()->getRank());
    //algebraic->writeCrsMatrixToFile(algebraic->matrices->B_T, name_B_T, algebraic->map->mapDofs->getComm()->getRank());
    //algebraic->writeCrsMatrixToFile(algebraic->matrices->C, name_C, algebraic->map->mapDofs->getComm()->getRank());

    return hasConverged == 1.;
}

template <typename scalar>
void Discrete_problem<scalar>::createBlockMatrix(scalar gamma, bool write)
{
    using Teuchos::Array;
    using Teuchos::RCP;
    using Teuchos::rcp;

    const std::vector<RCP<const tpetra_map_type>> maps = {algebraic->map->mapDofs, algebraic->map->mapLagrangeDofs};

    const size_t numDofs = algebraic->map->mapDofs->getGlobalNumElements();
    const size_t numMyDofs = algebraic->map->mapDofs->getNodeNumElements();
    const size_t numMyDofsWO = algebraic->map->mapDofsWO->getNodeNumElements();
    Teuchos::ArrayView<const global_ordinal_type> myDofs = algebraic->map->mapDofs->getNodeElementList();
    Teuchos::ArrayView<const global_ordinal_type> myDofsWO = algebraic->map->mapDofsWO->getNodeElementList();

    const size_t numLagrangeDofs = algebraic->map->mapLagrangeDofs->getGlobalNumElements();
    const size_t numMyLagrangeDofs = algebraic->map->mapLagrangeDofs->getNodeNumElements();
    Teuchos::ArrayView<const global_ordinal_type> myLagrangeDofs = algebraic->map->mapLagrangeDofs->getNodeElementList();

    Array<global_ordinal_type> myFullDofs(numMyDofs + numMyLagrangeDofs);
    for (auto i = 0; i < numMyDofs; ++i)
        myFullDofs[i] = myDofs[i];
    for (auto i = 0; i < numMyLagrangeDofs; ++i)
        myFullDofs[numMyDofs + i] = numDofs + myLagrangeDofs[i];

    Array<global_ordinal_type> myFullDofsWO(numMyDofsWO + numMyLagrangeDofs);
    for (auto i = 0; i < numMyDofsWO; ++i)
        myFullDofsWO[i] = myDofsWO[i];
    for (auto i = 0; i < numMyLagrangeDofs; ++i)
        myFullDofsWO[numMyDofsWO + i] = numDofs + myLagrangeDofs[i];

    algebraic->map->fullmap = rcp(new tpetra_map_type(numDofs + numLagrangeDofs, myFullDofs, algebraic->map->mapDofs->getIndexBase(), algebraic->map->mapDofs->getComm()));
    algebraic->map->fullmapWO = rcp(new tpetra_map_type(numDofs + numLagrangeDofs, myFullDofsWO, algebraic->map->mapDofs->getIndexBase(), algebraic->map->mapDofs->getComm()));

    RCP<xpetra_map_type> xfullmap = rcp(new xpetra_tmap_type(algebraic->map->fullmap));

    const size_t myRank = algebraic->map->mapDofs->getComm()->getRank();
    const size_t numLclIndices = (myRank == 0) ? numDofs + numLagrangeDofs : 0;

    algebraic->map->fullmapOutput = rcp(new tpetra_map_type(numDofs + numLagrangeDofs, numLclIndices, algebraic->map->mapDofs->getIndexBase(), algebraic->map->mapDofs->getComm()));

    std::vector<RCP<const xpetra_map_type>> xsubmaps = {rcp(new xpetra_tmap_type(algebraic->map->mapDofs)), rcp(new xpetra_tmap_type(algebraic->map->mapLagrangeDofs))};

    /*
        std::cout << "=========== Full map ===========" << std::endl;
        std::cout << *algebraic->map->fullmap << std::endl;

        {
            std::ofstream outputfile;
            std::string name = "fullMap";
            outputfile.open (name + "_" + std::to_string(algebraic->map->mapDofs->getComm()->getRank()) + ".txt");
            outputfile << *algebraic->map->fullmap << std::endl;
            size_t numElement = algebraic->map->fullmap->getGlobalNumElements();
            for (auto i = 0; i < numElement; ++i)
            {
                outputfile << algebraic->map->fullmap->isNodeGlobalElement(i) << std::endl;
            }
            outputfile.close();
        }


        std::cout << "=========== Dofs map ===========" << std::endl;
        std::cout << *algebraic->map->mapDofs << std::endl;

        {
            std::ofstream outputfile;
            std::string name = "dofsMap";
            outputfile.open (name + "_" + std::to_string(algebraic->map->mapDofs->getComm()->getRank()) + ".txt");
            outputfile << *algebraic->map->mapDofs << std::endl;
            size_t numElement = algebraic->map->mapDofs->getGlobalNumElements();
            for (auto i = 0; i < numElement; ++i)
            {
                outputfile << algebraic->map->mapDofs->isNodeGlobalElement(i) << std::endl;
            }
            outputfile.close();
        }

        std::cout << "=========== Lagrange Dofs map ===========" << std::endl;
        std::cout << *algebraic->map->mapLagrangeDofs << std::endl;

        {
            std::ofstream outputfile;
            std::string name = "LagrangeDofsMap";
            outputfile.open (name + "_" + std::to_string(algebraic->map->mapDofs->getComm()->getRank()) + ".txt");
            outputfile << *algebraic->map->mapLagrangeDofs << std::endl;
            size_t numElement = algebraic->map->mapLagrangeDofs->getGlobalNumElements();
            for (auto i = 0; i < numElement; ++i)
            {
                outputfile << algebraic->map->mapLagrangeDofs->isNodeGlobalElement(i) << std::endl;
            }
            outputfile.close();
        }

        std::cout << "=========================================" << std::endl;
        */

    //Xpetra::UnderlyingLib lib = Xpetra::useTpetra;
    //RCP< xpetra_map_type > xfullmap = Xpetra::MapFactory<>::Build(lib,)

    algebraic->map->blockedMap =
        rcp(new xpetra_bmap_type(xfullmap, xsubmaps, true));

    //RCP<xpetra_bcrs_type >
    algebraic->matrices->blockedMatrix =
        rcp(new xpetra_bcrs_type(algebraic->map->blockedMap, algebraic->map->blockedMap, 8));

    /*
        std::string name_A = "A before blocked.txt";
        std::string name_B = "B before blocked.txt";
        std::string name_B_T = "B_T before blocked.txt";
        std::string name_B_G = "B_G before blocked.txt";
        std::string name_C = "C before blocked.txt";
        algebraic->writeCrsMatrixToFile(algebraic->matrices->A, name_A, algebraic->map->mapDofs->getComm()->getRank());
        algebraic->writeCrsMatrixToFile(algebraic->matrices->B, name_B, algebraic->map->mapDofs->getComm()->getRank());
        algebraic->writeCrsMatrixToFile(algebraic->matrices->B_T, name_B_T, algebraic->map->mapDofs->getComm()->getRank());
        algebraic->writeCrsMatrixToFile(algebraic->matrices->B_G, name_B_G, algebraic->map->mapDofs->getComm()->getRank());
        algebraic->writeCrsMatrixToFile(algebraic->matrices->C, name_C, algebraic->map->mapDofs->getComm()->getRank());
        */

    //Update B based on the initial activity

    algebraic->matrices->B_2->resumeFill();
    algebraic->matrices->C->resumeFill();
    algebraic->matrices->Cb->resumeFill();

    //const int numOfMechLagrange = algebraic->map->lm_to_dof.size();
    const size_t numDPN = algebraic->map->numPrimalDPN;
    //const size_t numLMPN = (numDPN==4) ? 2: 1;

    size_t numLMPN;
    if (domain->contacts_list->hasAtLeastOneSticking() && numDPN == 4)
        numLMPN = 4;
    else if (domain->contacts_list->hasAtLeastOneSticking())
        numLMPN = 3;
    else if (numDPN == 4)
        numLMPN = 2;
    else
        numLMPN = 1;

    const int numMyLagrangeNum = algebraic->map->mapLagrangeDofs->getNodeNumElements();

    //const int numMyMechLagrangeNum = (numDPN==4) ? numMyLagrangeNum/2: numMyLagrangeNum;
    const int numMyMechLagrangeNum = numMyLagrangeNum / numLMPN;

    old_activity = Kokkos::View<scalar *, Kokkos::LayoutRight>("R", numMyMechLagrangeNum);
    old_old_activity = Kokkos::View<scalar *, Kokkos::LayoutRight>("R", numMyMechLagrangeNum);

    for (auto local_index = 0; local_index < numMyMechLagrangeNum; ++local_index)
    {
        global_ordinal_type lagrange_id = algebraic->map->mapLagrangeDofs->getGlobalElement(local_index);
        //std::cout << "lagrange id " <<lagrange_id <<std::endl;
        global_ordinal_type node_id = algebraic->map->lm_to_dof[lagrange_id];
        lagrange_id = lagrange_id * numLMPN;
        //std::cout << "node id " <<node_id <<std::endl;
        local_ordinal_type interface_id = domain->contacts_list->getInterfaceOfSlaveNode(node_id);
        //std::cout << "interface id " <<interface_id <<std::endl;
        if (domain->contacts_list->isInitiallyOpen(interface_id))
        {
            if (domain->contacts_list->isNodeInitiallyClosed(interface_id, node_id))
            {
                old_activity(local_index) = (scalar)1.;
                continue;
            }
            old_activity(local_index) = (scalar)0.;
            //std::cout << "replace row " << local_index << "/" << algebraic->map->mapLagrangeDofs->getNodeNumElements() << std::endl;
            Teuchos::ArrayView<const local_ordinal_type> local_indices;
            Teuchos::ArrayView<const scalar> values;

            for (size_t j = 0; j < numLMPN; ++j)
            {
                algebraic->matrices->B_2->getLocalRowView(local_index * numLMPN + j, local_indices, values);

                for (auto k = 0; k < local_indices.size(); ++k)
                {
                    global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                    algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + j,
                                                                  Teuchos::tuple<global_ordinal_type>(column_index),
                                                                  Teuchos::tuple<scalar>((scalar)0.));
                }

                algebraic->matrices->C->replaceGlobalValues(lagrange_id + j,
                                                            Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                            Teuchos::tuple<scalar>((scalar)1.));

                algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + j,
                                                             Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                             Teuchos::tuple<scalar>((scalar)0.));
            }
            /*
                if (numLMPN==2 || numLMPN==4)
                {
                    algebraic->matrices->B_2->getLocalRowView (local_index*numLMPN+numLMPN-1, local_indices, values);

                    for ( auto k = 0; k < local_indices.size() ; ++k)
                    {
                        global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                        algebraic->matrices->B_2->replaceGlobalValues(lagrange_id+numLMPN-1,
                            Teuchos::tuple<global_ordinal_type> (column_index),
                            Teuchos::tuple<scalar> ((scalar) 0.));
                    }

                    algebraic->matrices->C->replaceGlobalValues(lagrange_id+numLMPN-1,
                        Teuchos::tuple<global_ordinal_type> (lagrange_id+numLMPN-1),
                        Teuchos::tuple<scalar> ((scalar) 1.));

                    algebraic->matrices->Cb->replaceGlobalValues(lagrange_id+numLMPN-1,
                        Teuchos::tuple<global_ordinal_type> (lagrange_id+numLMPN-1),
                        Teuchos::tuple<scalar> ((scalar) 0.));
                }
                */
        }
        else
        {
            old_activity(local_index) = (scalar)1.;
            if (domain->contacts_list->isSticking(interface_id) == false && domain->contacts_list->hasAtLeastOneSticking() == true)
            {
                Teuchos::ArrayView<const local_ordinal_type> local_indices;
                Teuchos::ArrayView<const scalar> values;

                for (size_t j = 1; j < numLMPN; ++j)
                {
                    algebraic->matrices->B_2->getLocalRowView(local_index * numLMPN + j, local_indices, values);

                    for (auto k = 0; k < local_indices.size(); ++k)
                    {
                        global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                        algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + j,
                                                                      Teuchos::tuple<global_ordinal_type>(column_index),
                                                                      Teuchos::tuple<scalar>((scalar)0.));
                    }

                    algebraic->matrices->C->replaceGlobalValues(lagrange_id + j,
                                                                Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                Teuchos::tuple<scalar>((scalar)1.));

                    algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + j,
                                                                 Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                 Teuchos::tuple<scalar>((scalar)0.));
                }
            }
        }
    }
    algebraic->matrices->B_2->fillComplete(algebraic->map->mapDofs, algebraic->map->mapLagrangeDofs);
    algebraic->matrices->C->fillComplete();
    algebraic->matrices->Cb->fillComplete();

    algebraic->matrices->xA = rcp(new xpetra_tcrs_type(algebraic->matrices->A));
    algebraic->matrices->xB = rcp(new xpetra_tcrs_type(algebraic->matrices->B_2));
    algebraic->matrices->xB_T = rcp(new xpetra_tcrs_type(algebraic->matrices->B_T));
    algebraic->matrices->xC = rcp(new xpetra_tcrs_type(algebraic->matrices->C));

    algebraic->matrices->xwA = rcp(new xpetra_wcrs_type(algebraic->matrices->xA));
    algebraic->matrices->xwB = rcp(new xpetra_wcrs_type(algebraic->matrices->xB));
    algebraic->matrices->xwB_T = rcp(new xpetra_wcrs_type(algebraic->matrices->xB_T));
    algebraic->matrices->xwC = rcp(new xpetra_wcrs_type(algebraic->matrices->xC));

    RCP<Xpetra::Matrix<scalar, local_ordinal_type, global_ordinal_type, node_type>> CTC = Teuchos::null;
    RCP<Xpetra::Matrix<scalar, local_ordinal_type, global_ordinal_type, node_type>> ApCTC = Teuchos::null;
    std::string MatrixMatrix_output_file = "MatrixMatrix_out.txt";
    std::ofstream MatrixMatrix_ofstream(MatrixMatrix_output_file);
    RCP<Teuchos::FancyOStream> verbOut = Teuchos::getFancyOStream(Teuchos::rcpFromRef(MatrixMatrix_ofstream));
    CTC = Xpetra::MatrixMatrix<scalar, local_ordinal_type, global_ordinal_type, node_type>::Multiply(*(algebraic->matrices->xwB_T), false, *(algebraic->matrices->xwB), false, *verbOut);
    Xpetra::MatrixMatrix<scalar, local_ordinal_type, global_ordinal_type, node_type>::TwoMatrixAdd(*CTC, false, gamma, *(algebraic->matrices->xwA), false, (scalar)1., ApCTC, *verbOut);
    ApCTC->fillComplete();

    /*

        RCP<xpetra_crs_type > xA =
            rcp (new xpetra_tcrs_type(algebraic->matrices->A));
        RCP<xpetra_crs_type > xB =
            rcp (new xpetra_tcrs_type(algebraic->matrices->B_2));
        RCP<xpetra_crs_type > xB_T =
            rcp (new xpetra_tcrs_type(algebraic->matrices->B_T));
        RCP<xpetra_crs_type > xC =
            rcp (new xpetra_tcrs_type(algebraic->matrices->C));


        RCP<xpetra_wcrs_type > xwA =
            rcp (new xpetra_wcrs_type(xA));

        RCP<xpetra_wcrs_type > xwB =
            rcp (new xpetra_wcrs_type(xB));

        RCP<xpetra_wcrs_type > xwB_T =
            rcp (new xpetra_wcrs_type(xB_T));

        RCP<xpetra_wcrs_type > xwC =
            rcp (new xpetra_wcrs_type(xC));

        */
    //algebraic->matrices->blockedMatrix->setMatrix(0,0,algebraic->matrices->xwA);
    algebraic->matrices->blockedMatrix->setMatrix(0, 0, ApCTC);
    algebraic->matrices->blockedMatrix->setMatrix(0, 1, algebraic->matrices->xwB_T);
    algebraic->matrices->blockedMatrix->setMatrix(1, 0, algebraic->matrices->xwB);
    algebraic->matrices->blockedMatrix->setMatrix(1, 1, algebraic->matrices->xwC);

    algebraic->matrices->blockedMatrix->fillComplete();
    RCP<const xpetra_wcrs_type> tmpOp = Teuchos::rcp_dynamic_cast<const xpetra_wcrs_type>(ApCTC);
    const RCP<const xpetra_tcrs_type> &tmp_ECrsMtx = Teuchos::rcp_dynamic_cast<const xpetra_tcrs_type>(tmpOp->getCrsMatrix());
    RCP<const tpetra_crs_type> tpetra_ApCTC = tmp_ECrsMtx->getTpetra_CrsMatrix();

    if (write)
        Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeSparseFile(std::string("ApCTC_mm.txt"), tpetra_ApCTC);
}
template <typename scalar>
void Discrete_problem<scalar>::createBlockMVector(bool write)
{
    using Teuchos::Array;
    using Teuchos::RCP;
    using Teuchos::rcp;

    /*
        std::cout << "algebraic->map->mapLagrangeDofs" << std::endl;
        std::cout << *algebraic->map->mapLagrangeDofs << std::endl;

        std::cout << "algebraic->matrices->B_G->getRangeMap()" << std::endl;
        std::cout << *algebraic->matrices->B_G->getRangeMap() << std::endl;

        std::cout << "algebraic->matrices->B_G->getDomainMap()" << std::endl;
        std::cout << *algebraic->matrices->B_G->getDomainMap() << std::endl;
        */

    RCP<tpetra_vector_type> intialContactPosition(new tpetra_vector_type(algebraic->map->mapDofs, true));
    algebraic->vectors->initialGap = rcp(new tpetra_vector_type(algebraic->map->mapLagrangeDofs, true));
    RCP<tpetra_vector_type> initialActiveGap(new tpetra_vector_type(algebraic->map->mapLagrangeDofs, true));

    algebraic->vectors->lagrange = rcp(new tpetra_vector_type(algebraic->map->mapLagrangeDofs, true));

    {
        size_t numMyContacts = domain->contacts_list->getContactNumber();
        size_t numPrimalDPN = algebraic->map->numPrimalDPN;
        for (auto i = 0; i < numMyContacts; ++i)
        {
            size_t numMySlaveNodes = domain->contacts_list->getSlaveNodesSize(i);
            for (auto j = 0; j < numMySlaveNodes; ++j)
            {
                // get unique node id
                auto global_nodeid = domain->contacts_list->getSlaveNode(i, j);
                if (algebraic->map->mapNodes->isNodeGlobalElement(global_nodeid))
                {
                    auto local_nodeid = algebraic->map->mapNodesWO->getLocalElement(global_nodeid);
                    for (auto k = 0; k < 3; ++k)
                        intialContactPosition->replaceGlobalValue(global_nodeid * numPrimalDPN + k,
                                                                  (((scalar)domain->nodes_list->nodes(local_nodeid, k)) + domain->contacts_list->getSlaveNormal(i, k) * loads->preload_list->getPreloadValue(i)));
                }
            }

            size_t numMyMasterNodes = domain->contacts_list->getMasterNodesSize(i);
            for (auto j = 0; j < numMyMasterNodes; ++j)
            {
                // get unique node id
                auto global_nodeid = domain->contacts_list->getMasterNode(i, j);
                if (algebraic->map->mapNodes->isNodeGlobalElement(global_nodeid))
                {
                    auto local_nodeid = algebraic->map->mapNodesWO->getLocalElement(global_nodeid);
                    for (auto k = 0; k < 3; ++k)
                        intialContactPosition->replaceGlobalValue(global_nodeid * numPrimalDPN + k,
                                                                  ((scalar)domain->nodes_list->nodes(local_nodeid, k)));
                }
            }
        }
    }

    /*
        std::string name_initial_0 = "initial gap 0.txt";
        algebraic->writeVectorToFile( name_initial_0, algebraic->map->mapDofs->getComm()->getRank(), initialGap);

        std::string name_lag = "lagrange 0.txt";
        algebraic->writeVectorToFile( name_lag, algebraic->map->mapDofs->getComm()->getRank(), algebraic->vectors->lagrange);
        */
    algebraic->matrices->B_G->apply(*intialContactPosition, *algebraic->vectors->initialGap, Teuchos::NO_TRANS, ((scalar)-1.));

    algebraic->matrices->Cb->apply(*algebraic->vectors->initialGap, *initialActiveGap);
    /*
        std::string name_initial_pos = "initial pos.txt";
        algebraic->writeVectorToFile( name_initial_pos, algebraic->map->mapDofs->getComm()->getRank(), intialContactPosition);

        std::string name_initial = "initial gap.txt";
        algebraic->writeVectorToFile( name_initial, algebraic->map->mapDofs->getComm()->getRank(), initialGap);

        std::string name_disp = "displacement rhs.txt";
        algebraic->writeVectorToFile( name_disp, algebraic->map->mapDofs->getComm()->getRank(), algebraic->vectors->b);
        */

    // TODO: fix me when there is mesh tying and contact together!
    if (domain->contacts_list->isTying(0))
        initialActiveGap->scale((scalar)0.);

    if (write)
        Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("initialActiveGap_mm.txt"), initialActiveGap);

    RCP<xpetra_mvector_type> xb =
        rcp(new xpetra_tmvector_type(algebraic->vectors->b));

    RCP<xpetra_mvector_type> xg =
        rcp(new xpetra_tmvector_type(initialActiveGap));

    RCP<xpetra_mvector_type> xx =
        rcp(new xpetra_tmvector_type(algebraic->vectors->x));

    RCP<xpetra_mvector_type> xlagrange =
        rcp(new xpetra_tmvector_type(algebraic->vectors->lagrange));

    algebraic->vectors->rhsBlockedMultiVector =
        rcp(new xpetra_bmvector_type(algebraic->map->blockedMap, 1, true));

    algebraic->vectors->solutionBlockedMultiVector =
        rcp(new xpetra_bmvector_type(algebraic->map->blockedMap, 1, true));

    algebraic->vectors->rhsBlockedMultiVector->setMultiVector(0, xb, true);
    algebraic->vectors->rhsBlockedMultiVector->setMultiVector(1, xg, true);

    algebraic->vectors->solutionBlockedMultiVector->setMultiVector(0, xx, true);
    algebraic->vectors->solutionBlockedMultiVector->setMultiVector(1, xlagrange, true);

    //std::cout << *(algebraic->vectors->rhsBlockedMultiVector) <<std::endl;

    algebraic->vectors->rhsMultiVector = algebraic->vectors->rhsBlockedMultiVector->Merge();
    algebraic->vectors->solutionMultiVector = algebraic->vectors->solutionBlockedMultiVector->Merge();
}

template <typename scalar>
void Discrete_problem<scalar>::computeContactMatrices(MPI_Comm comm)
{
    std::cout << "Before compute_B<scalar>" << std::endl;
    compute_B<scalar>(algebraic->matrices,
                      domain->elements_list,
                      domain->nodes_list,
                      domain->contacts_list,
                      algebraic->map, comm);
    std::cout << "After compute_B<scalar>" << std::endl;
}
template <typename scalar>
void Discrete_problem<scalar>::computeMatrices(void)
{

    //std::cout << " A " << std::endl;
    //typedef Kokkos::CrsMatrix<scalar,local_ordinal_type,tpetra_crs_type::execution_space> local_matrix_type;
    typedef KokkosSparse::CrsMatrix<scalar, local_ordinal_type, typename tpetra_crs_type::execution_space> local_matrix_type;

    typedef Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace> elem_matrix_view_type;

    //std::cout << " B " << std::endl;

    algebraic->matrices->AWO->resumeFill();
    algebraic->matrices->AWO->setAllToScalar((scalar)0.0);
    algebraic->matrices->AWO->fillComplete();

    algebraic->matrices->A->resumeFill();
    algebraic->matrices->A->setAllToScalar((scalar)0.0);
    algebraic->matrices->A->fillComplete();

    algebraic->matrices->AWO->resumeFill();

    //std::cout << " C " << std::endl;

    local_matrix_type local_AWO = algebraic->matrices->AWO->getLocalMatrix();

    auto numMyElems = domain->elements_list->Get_Element_Number();

    //std::cout << " D " << std::endl;

    elem_matrix_view_type K, L, S;

#ifdef KOKKOS_ENABLE_DEPRECATED_CODE
    int pool_size = Kokkos::DefaultExecutionSpace::thread_pool_size();
#else
    int pool_size = Kokkos::DefaultExecutionSpace::impl_thread_pool_size();
#endif
    int max_element_size = 1;

    for (int i_elem = 0; i_elem < numMyElems; ++i_elem)
    {
        if (domain->elements_list->Get_Element_Type(i_elem) == static_cast<int>(tbox::ELTYPE::HEX8))
        {
            max_element_size = 8;
            break;
        }
        if (domain->elements_list->Get_Element_Type(i_elem) == static_cast<int>(tbox::ELTYPE::TETRA4))
        {
            max_element_size = 4;
            break;
        }
    }

    K = Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A", pool_size, 3 * max_element_size, 4 * max_element_size);
    L = Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A", pool_size, max_element_size, max_element_size);
    S = Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A", pool_size, 3 * max_element_size, max_element_size);

    ElementMatrices<scalar, static_cast<int>(tbox::ELTYPE::HEX8), 8> *elementMatrices_HEX8 = new ElementMatrices<scalar, static_cast<int>(tbox::ELTYPE::HEX8), 8>(domain, numPrimalDPN, pool_size);
    ElementMatrices<scalar, static_cast<int>(tbox::ELTYPE::TETRA4), 4> *elementMatrices_TETRA4 = new ElementMatrices<scalar, static_cast<int>(tbox::ELTYPE::TETRA4), 4>(domain, numPrimalDPN, pool_size);

    Kokkos::parallel_for(numMyElems, KOKKOS_LAMBDA(const int i_elem) {
        bool bContinue = true;

#ifdef KOKKOS_ENABLE_DEPRECATED_CODE
        int i_thread = Kokkos::DefaultExecutionSpace::thread_pool_rank();
#else
        int i_thread = Kokkos::DefaultExecutionSpace::impl_thread_pool_rank();
#endif

        //std::cout << " D1 " << std::endl;

        auto K_e = subview(K, i_thread, Kokkos::ALL(), Kokkos::ALL());
        auto L_e = subview(L, i_thread, Kokkos::ALL(), Kokkos::ALL());
        auto S_e = subview(S, i_thread, Kokkos::ALL(), Kokkos::ALL());

        if (domain->elements_list->Get_Element_Type(i_elem) == static_cast<int>(tbox::ELTYPE::HEX8))
        {
            elementMatrices_HEX8->compute(i_elem, i_thread);
            Kokkos::deep_copy(K_e, elementMatrices_HEX8->getK(i_thread));
            Kokkos::deep_copy(L_e, elementMatrices_HEX8->getL(i_thread));
            Kokkos::deep_copy(S_e, elementMatrices_HEX8->getS(i_thread));
        }

        else if (domain->elements_list->Get_Element_Type(i_elem) == static_cast<int>(tbox::ELTYPE::TETRA4))
        {
            elementMatrices_TETRA4->compute(i_elem, i_thread);
            Kokkos::deep_copy(K_e, elementMatrices_TETRA4->getK(i_thread));
            Kokkos::deep_copy(L_e, elementMatrices_TETRA4->getL(i_thread));
            Kokkos::deep_copy(S_e, elementMatrices_TETRA4->getS(i_thread));
        }
        else
            bContinue = false;

        //std::cout << " D2 " << std::endl;

        if (bContinue)
            for (auto i = 0; i < domain->elements_list->Get_Element_Size(i_elem); ++i)
            {
                local_ordinal_type local_i = numPrimalDPN * (domain->elements_list->Get_Element_Node(i_elem, i));
                for (auto j = 0; j < domain->elements_list->Get_Element_Size(i_elem); ++j)
                {
                    local_ordinal_type local_j = numPrimalDPN * (domain->elements_list->Get_Element_Node(i_elem, j));
                    // K
                    for (auto k = 0; k < 3; ++k)
                        for (auto l = 0; l < 3; ++l)
                        {
                            local_ordinal_type cols = local_j + l;
                            scalar vals = K_e(3 * i + k, 3 * j + l);
                            local_AWO.sumIntoValues(local_i + k, &cols, 1, &vals, false, true);
                        }
                    if (numPrimalDPN == 4)
                    {
                        // partial K /partial T
                        //for ( auto k = 0; k < 3 ; ++k)
                        //{
                        //    local_ordinal_type cols = local_j+3;
                        //    scalar vals = K_e(3*i+k,3*mesh_Elems(i_elem,elems_size)+j);
                        //    local_AWO.sumIntoValues(local_i+k,&cols,1,&vals);
                        //}

                        // L
                        local_ordinal_type cols = local_j + 3;
                        scalar vals = L_e(i, j);
                        //if (local_i == local_j)
                        //    std::cout << local_i <<"  : values : " <<vals << std::endl;
                        local_AWO.sumIntoValues(local_i + 3, &cols, 1, &vals, false, true);

                        // S
                        for (auto k = 0; k < 3; ++k)
                        {
                            scalar vals = S_e(3 * i + k, j);

                            local_AWO.sumIntoValues(local_i + k, &cols, 1, &vals, false, true);
                        }
                    }
                }
            }
        //std::cout << " D3 " << std::endl;
    });

    //std::cout << " E " << std::endl;

    algebraic->matrices->AWO->fillComplete();

    if (1)
    {
        Tpetra::Export<> exportAWO(algebraic->map->mapDofsWO, algebraic->map->mapDofs);

        LinearSolver::getTimers()["assembly-MPI"].start();
        algebraic->matrices->A->resumeFill();
        algebraic->matrices->A->doExport(*(algebraic->matrices->AWO), exportAWO, Tpetra::ADD);
        algebraic->vectors->b->doExport(*(algebraic->vectors->bWO), exportAWO, Tpetra::ADD);
        algebraic->matrices->A->fillComplete();
        LinearSolver::getTimers()["assembly-MPI"].stop();
    }
    //Tpetra::Export<> exportAWO(algebraic->map->mapDofsWO,algebraic->map->mapDofs);

    //algebraic->matrices->A->resumeFill();
    //algebraic->matrices->A->doExport(*(algebraic->matrices->AWO), exportAWO, Tpetra::ADD);
    //algebraic->matrices->A->fillComplete ();
}
template <typename scalar>
void Discrete_problem<scalar>::computeLoads(void)
{
    //std::cout << "discrete_problem.computeLoads" << domain->elements_list->Get_Element_Node(17,0) << " " << domain->elements_list->Get_Element_Node(17,1) << " " << domain->elements_list->Get_Element_Node(17,2) << " " << domain->elements_list->Get_Element_Node(17,3) << std::endl;

    //const size_t myRank = algebraic->map->mapDofs->getComm()->getRank();

    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;

    typedef Kokkos::View<scalar **, Kokkos::LayoutRight> elem_matrix_view_type;
    // ------------------------------------------
    //
    //          volumetric heat sources
    //
    // ------------------------------------------

    auto numMySources = loads->sources_list->getSourceNumber();

    for (auto i = 0; i < numMySources; ++i)
    {
        scalar integrated_values = 0;

        int number_of_cell_computed = 0;

        auto currentSourceSize = loads->sources_list->getSourceSize(i);
        for (auto e1 = 0; e1 < currentSourceSize; ++e1)
        {
            local_ordinal_type i_elem = loads->sources_list->getSourceElement(i, e1);

            elem_matrix_view_type N_e;

            bool bContinue = true;

            if (domain->elements_list->Get_Element_Type(i_elem) == static_cast<int>(tbox::ELTYPE::HEX8))
            {
                ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::HEX8), 8> elementVectors = ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::HEX8), 8>(i_elem, domain, loads, numPrimalDPN);
                N_e = elementVectors.N;
            }
            else if (domain->elements_list->Get_Element_Type(i_elem) == static_cast<int>(tbox::ELTYPE::TETRA4))
            {
                ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::TETRA4), 4> elementVectors = ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::TETRA4), 4>(i_elem, domain, loads, numPrimalDPN);
                N_e = elementVectors.N;
            }
            else
                bContinue = false;

            if (bContinue)
            {
                ++number_of_cell_computed;

                auto j = numPrimalDPN - 1;

                //std::cout << " myRank " << myRank << " i_elem " << i_elem << " " << domain->elements_list->Get_Element_Size(i_elem) << std::endl;

                for (auto ii = 0; ii < domain->elements_list->Get_Element_Size(i_elem); ++ii)
                {
                    local_ordinal_type local_i = numPrimalDPN * (domain->elements_list->Get_Element_Node(i_elem, ii)) + j;
                    global_ordinal_type global_i = algebraic->map->mapDofsWO->getGlobalElement(local_i);

                    for (auto jj = 0; jj < domain->elements_list->Get_Element_Size(i_elem); ++jj)
                    {
                        algebraic->vectors->bWO->sumIntoGlobalValue(global_i, ((scalar)loads->sources_list->getSourceValue(i)) * N_e(ii, jj));
                        integrated_values += ((scalar)loads->sources_list->getSourceValue(i)) * N_e(ii, jj);
                    }
                }
            }
        }
        //std::cout << " myRank " << myRank << " currentSourceSize " << currentSourceSize << " number_of_cell_computed " << number_of_cell_computed << " integrated source values " << integrated_values << endl;
    }

    //algebraic->vectors->b->doExport(*(algebraic->vectors->bWO), exportAWO, Tpetra::ADD);

    // ------------------------------------------
    //
    //       von Neumann boundary conditions
    //
    // ------------------------------------------

    auto numMynBCs = loads->neumann_list->getNeumannNumber();

    for (auto i = 0; i < numMynBCs; ++i)
    {
        scalar integrated_values = 0;

        auto currentBCSize = loads->neumann_list->getNeumannSize(i);
        //std::cout << i << " currentBCSize: " << currentBCSize << std::endl;
        for (auto e1 = 0; e1 < currentBCSize; ++e1)

        {

            local_ordinal_type i_elem = loads->neumann_list->getNeumannElement(i, e1);

            elem_matrix_view_type N_e;

            /*
                //elem_subview_type currentElem = subview(mesh_Elems,e1,ALL());

                elem_matrix_view_type N_e;

                bool bContinue = true;

                if (currentElem(elems_type) == EL_QUAD4)
                    N_e = trilinosQuad4BuildN<scalar>(currentElem,mesh_Nodes);
                else if (currentElem(elems_type) == static_cast<int>(tbox::ELTYPE::TRI3))
                    N_e = trilinosTri3BuildN<scalar>(currentElem,mesh_Nodes);
                else
                    bContinue = false;
                */

            bool bContinue = true;

            if (domain->elements_list->Get_Element_Type(i_elem) == static_cast<int>(tbox::ELTYPE::TRI3))
            {
                ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::TRI3), 3> elementVectors = ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::TRI3), 3>(i_elem, domain, loads, numPrimalDPN);
                N_e = elementVectors.N;
                //std::cout << "Hi i'm a triangle"<< std::endl;
                //std::cout << N_e(0,0)<< std::endl;
            }
            else if (domain->elements_list->Get_Element_Type(i_elem) == static_cast<int>(tbox::ELTYPE::QUAD4))
            {
                ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::QUAD4), 4> elementVectors = ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::QUAD4), 4>(i_elem, domain, loads, numPrimalDPN);
                N_e = elementVectors.N;
                //std::cout << "Hi i'm a quadrangle"<< std::endl;
                //std::cout << N_e(0,0)<< std::endl;
            }
            else
                bContinue = false;

            /*
                if (bContinue)
                {
                    for ( auto j = 0; j < numPrimalDPN; ++j)
                        if (values_nBCs(i,0,nBCs_dofs+j)==1)
                            for ( auto ii = 0; ii < currentElem(elems_size) ; ++ii)
                            {
                                local_ordinal_type local_i = numPrimalDPN*currentElem(elems_nodes+ii)+j;
                                global_ordinal_type global_i = mapDofsWO->getGlobalElement(local_i);

                                for ( auto jj = 0; jj < currentElem(elems_size) ; ++jj)
                                    {
                                        bWO->sumIntoGlobalValue(global_i, (scalar(values_nBCs(i,0,nBCs_values+j)*N_e(ii,jj))));
                                        if (j==3)
                                            integrated_values += values_nBCs(i,0,nBCs_values+j)*N_e(ii,jj);
                                    }
                            }
                }
                */

            if (bContinue)
            {
                for (auto j = 0; j < numPrimalDPN; ++j)
                {
                    scalar which_dof = loads->neumann_list->getNeumannDof(i, j);
                    scalar value = loads->neumann_list->getNeumannValue(i, j);
                    //std::cout << j << " " << which_dof << " " << value << std::endl;
                    if (which_dof == (scalar)1)
                    {
                        for (auto ii = 0; ii < domain->elements_list->Get_Element_Size(i_elem); ++ii)
                        {
                            local_ordinal_type local_i = numPrimalDPN * (domain->elements_list->Get_Element_Node(i_elem, ii)) + j;
                            global_ordinal_type global_i = algebraic->map->mapDofsWO->getGlobalElement(local_i);

                            for (auto jj = 0; jj < domain->elements_list->Get_Element_Size(i_elem); ++jj)
                            {
                                algebraic->vectors->bWO->sumIntoGlobalValue(global_i, value * N_e(ii, jj));
                                //std::cout << loads->neumann_list->getNeumannValue(i,j)<< std::endl;
                                //std::cout << N_e(ii,jj)<< std::endl;
                                //std::cout << ((scalar) loads->neumann_list->getNeumannValue(i,j))*N_e(ii,jj)<< std::endl;
                                integrated_values += value * N_e(ii, jj);
                            }
                        }
                    }
                }
            }
        }

        //out << mesh_nBCs(i,nBCs_size) << endl;
        //std::cout << " currentBCSize " << currentBCSize << " integrated VN values " << integrated_values << std::endl;
    }

    // ------------------------------------------
    //
    //        Dirichlet boundary conditions
    //
    // ------------------------------------------
    /*
        algebraic->matrices->A->resumeFill();
        auto numMyBCs = loads->dirichlet_list->getDirichletNumber();
        for (auto i = 0; i < numMyBCs; ++i)
        {
            auto currentBCSize = loads->dirichlet_list->getDirichletSize(i);
            Kokkos::parallel_for (currentBCSize, KOKKOS_LAMBDA (const int i_node)
            {
                global_ordinal_type n1 = loads->dirichlet_list->getDirichletNode(i,i_node);

                for ( auto j = 0; j < numPrimalDPN; ++j)
                {
                    scalar which_dof = loads->dirichlet_list->getDirichletDof(i,j);

                    if (which_dof==(scalar) 1)
                    {


                        if ( algebraic->map->mapNodes->isNodeGlobalElement (n1) )
                        {

                            global_ordinal_type global_index = numPrimalDPN*n1+j;
                            local_ordinal_type local_index = algebraic->map->mapDofsWO->getLocalElement(global_index);

                            // Get the initialized elements of the current line using the graph
                            // Attention: the colums indeces "local_indices" obtained are
                            // local even if we did not seperate the columns on the different
                            // mpi process. It's why the
                            //    algebraic->map->mapDofsWO->getGlobalElement(local_indices[j])
                            // is needed.
                            Teuchos::ArrayView<const local_ordinal_type> local_indices ;
                            algebraic->graph->A->getLocalRowView (local_index, local_indices);

                            // Replace the initialized elements by zero
                            for ( auto k = 0; k < local_indices.size() ; ++k)
                                algebraic->matrices->A->replaceGlobalValues(global_index,
                                                         tuple<global_ordinal_type> (algebraic->map->mapDofs->getGlobalElement(local_indices[k])),
                                                         tuple<scalar> (scalar(0.0)));

                            // Put one on the diagonal element
                            algebraic->matrices->A->replaceGlobalValues(global_index,
                                                     tuple<global_ordinal_type> (global_index),
                                                     tuple<scalar> (scalar(1.0)));

                            // Put the corresponding value in the right hand side
                            algebraic->vectors->b->replaceGlobalValue(global_index, (scalar(loads->dirichlet_list->getDirichletValue(i,j))));
                        }
                    }
                }
            });
        }
*/
    /*
        Tpetra::Import<> importA(algebraic->map->mapDofs,algebraic->map->mapDofsWO);

        LinearSolver::getTimers()["assembly-MPI"].start();
        algebraic->matrices->AWO->resumeFill();
        algebraic->matrices->AWO->doImport(*(algebraic->matrices->A), importA, Tpetra::ADD);
        algebraic->matrices->AWO->fillComplete ();
        LinearSolver::getTimers()["assembly-MPI"].stop();
*/

    Tpetra::Export<> exportAWO(algebraic->map->mapDofsWO, algebraic->map->mapDofs);

    LinearSolver::getTimers()["assembly-MPI"].start();
    //algebraic->matrices->A->resumeFill();
    //algebraic->matrices->A->doExport(*(algebraic->matrices->AWO), exportAWO, Tpetra::ADD);
    algebraic->vectors->b->doExport(*(algebraic->vectors->bWO), exportAWO, Tpetra::ADD);
    //algebraic->matrices->A->fillComplete ();
    LinearSolver::getTimers()["assembly-MPI"].stop();

    algebraic->matrices->A->resumeFill();
    //algebraic->matrices->A->setAllToScalar((scalar) 0.0);
    //algebraic->matrices->A->fillComplete ();

    //algebraic->matrices->AWO->resumeFill();
    /*
        if (domain->contacts_list->getContactNumber() >= 1)
            algebraic->matrices->B_T->resumeFill();
*/
    auto numMyBCs = loads->dirichlet_list->getDirichletNumber();
    //std::cout << "numMyBCs " << numMyBCs << std::endl;
    for (auto i = 0; i < numMyBCs; ++i)
    {
        auto currentBCSize = loads->dirichlet_list->getDirichletSize(i);
        for (auto i_node = 0; i_node < currentBCSize; ++i_node)
        //Kokkos::parallel_for (currentBCSize, KOKKOS_LAMBDA (const int i_node) //correct this bug !
        {
            global_ordinal_type n1 = loads->dirichlet_list->getDirichletNode(i, i_node);

            //std::cout << "DC 2: "<< i << " node DC: " << n1 << std::endl;
            //std::cout << "Dealing with DC: "<< i << " node DC: " << n1+1 << " node row: " << n1 << std::endl;

            for (auto j = 0; j < numPrimalDPN; ++j)
            {
                scalar which_dof = loads->dirichlet_list->getDirichletDof(i, j);

                //std::cout << " j: "<< j << " which_dof " << which_dof << std::endl;

                if (which_dof == (scalar)1)
                {
                    global_ordinal_type global_index = numPrimalDPN * n1 + j;
                    local_ordinal_type local_index = algebraic->map->mapDofs->getLocalElement(global_index);

                    // Get the initialized elements of the current line using the graph
                    // Attention: the colums indeces "local_indices" obtained are
                    // local even if we did not seperate the columns on the different
                    // mpi process. It's why the
                    //    mapDofsWO->getGlobalElement(local_indices[j])
                    // is needed.
                    Teuchos::ArrayView<const local_ordinal_type> local_indices;
                    algebraic->graph->A->getLocalRowView(local_index, local_indices);

                    // Replace the initialized elements by zero
                    for (auto k = 0; k < local_indices.size(); ++k)
                        algebraic->matrices->A->replaceGlobalValues(global_index,
                                                                    tuple<global_ordinal_type>(algebraic->matrices->A->getColMap()->getGlobalElement(local_indices[k])),
                                                                    tuple<scalar>(scalar(0.0)));

                    if (algebraic->map->mapNodes->isNodeGlobalElement(n1))
                    {
                        // Put one on the diagonal element
                        algebraic->matrices->A->replaceGlobalValues(global_index,
                                                                    tuple<global_ordinal_type>(global_index),
                                                                    tuple<scalar>(scalar(1.0)));

                        // Put the corresponding value in the right hand side
                        algebraic->vectors->b->replaceGlobalValue(global_index, (scalar(loads->dirichlet_list->getDirichletValue(i, j))));
                        algebraic->vectors->x->replaceGlobalValue(global_index, (scalar(loads->dirichlet_list->getDirichletValue(i, j))));
                    }
                    /*
                        if (domain->contacts_list->getContactNumber() >= 1)
                        {
                            local_ordinal_type local_index = algebraic->map->mapDofs->getLocalElement(global_index);
                            algebraic->matrices->B_T->getGraph()->getLocalRowView (local_index, local_indices);
                            for ( auto k = 0; k < local_indices.size() ; ++k)
                                algebraic->matrices->B_T->replaceGlobalValues(local_index,
                                                     tuple<global_ordinal_type> (algebraic->matrices->B_T->getColMap()->getGlobalElement(local_indices[k])),
                                                     tuple<scalar> (scalar(0.0)));
                        }
*/
                }
            }
        } //);
    }
    algebraic->matrices->A->fillComplete();
    /*
        if (domain->contacts_list->getContactNumber() >= 1)
            algebraic->matrices->B_T->fillComplete (algebraic->map->mapLagrangeDofs,algebraic->map->mapDofs);
*/
}
}; // namespace katropton
