
template <typename scalar, int element_type>
double ElementComputation<scalar, element_type>::buildJ(int k, tMatrix<double, 3, 3> &J, local_ordinal_type e, const katropton::discrete_problem::Elements_list &elements_list, const katropton::discrete_problem::Nodes_list &nodes_list)
{
    tbox::Cache *cache;
    size_t element_size;

    if (element_type == static_cast<int>(tbox::ELTYPE::HEX8))
    {
        tbox::CacheHex8 &cachehex8 = trilinosHex8GetCache();
        cache = &cachehex8;
        element_size = 8;
    }
    else if (element_type == static_cast<int>(tbox::ELTYPE::TETRA4))
    {
        tbox::CacheTetra4 &cachetetra4 = trilinosTetra4GetCache();
        cache = &cachetetra4;
        element_size = 4;
    }
    else if (element_type == static_cast<int>(tbox::ELTYPE::TRI3))
    {
        tbox::CacheTri3 &cachetri3 = trilinosTri3GetCache();
        cache = &cachetri3;
        element_size = 3;

        tVector<double, 3> tg1, tg2, n;

        int i = 0;
        for (auto n1 = 0; n1 < element_size; ++n1, ++i)
        {
            for (auto j = 0; j < 3; ++j)
            {
                tg1(j) += (static_cast<tbox::CacheTri3 *>(cache)->dff)[k][i][0] * nodes_list.nodes(elements_list.Get_Element_Node(e, n1), j);
                tg2(j) += (static_cast<tbox::CacheTri3 *>(cache)->dff)[k][i][1] * nodes_list.nodes(elements_list.Get_Element_Node(e, n1), j);
            }
        }

        n(0) = tg1(1) * tg2(2) - tg1(2) * tg2(1);
        n(1) = tg1(2) * tg2(0) - tg1(0) * tg2(2);
        n(2) = tg1(0) * tg2(1) - tg1(1) * tg2(0);
        double detJ = sqrt(pow(n(0), 2) + pow(n(1), 2) + pow(n(2), 2));
        return detJ;
    }
    else if (element_type == static_cast<int>(tbox::ELTYPE::QUAD4))
    {
        tbox::CacheQuad4 &cachequad4 = trilinosQuad4GetCache();
        cache = &cachequad4;
        element_size = 4;

        tVector<double, 3> tg1, tg2, n;

        int i = 0;
        for (auto n1 = 0; n1 < element_size; ++n1, ++i)
        {
            for (auto j = 0; j < 3; ++j)
            {
                tg1(j) += (static_cast<tbox::CacheQuad4 *>(cache)->dff)[k][i][0] * nodes_list.nodes(elements_list.Get_Element_Node(e, n1), j);
                tg2(j) += (static_cast<tbox::CacheQuad4 *>(cache)->dff)[k][i][1] * nodes_list.nodes(elements_list.Get_Element_Node(e, n1), j);
            }
        }

        n(0) = tg1(1) * tg2(2) - tg1(2) * tg2(1);
        n(1) = tg1(2) * tg2(0) - tg1(0) * tg2(2);
        n(2) = tg1(0) * tg2(1) - tg1(1) * tg2(0);
        double detJ = sqrt(pow(n(0), 2) + pow(n(1), 2) + pow(n(2), 2));
        return detJ;
    }

    tMatrix<double, 3, 3> Jn;
    J.clean();

    tVector<double, 3> tmp, tmp2;

    int i = 0;
    for (auto n1 = 0; n1 < elements_list.Get_Element_Size(e); ++n1, ++i)
    {
        for (auto j = 0; j < 3; ++j)
        {
            tmp(j) = nodes_list.nodes(elements_list.Get_Element_Node(e, n1), j);
            if (element_type == static_cast<int>(tbox::ELTYPE::HEX8))
                tmp2(j) = (static_cast<tbox::CacheHex8 *>(cache)->dff)[k][i][j];
            else if (element_type == static_cast<int>(tbox::ELTYPE::TETRA4))
                tmp2(j) = (static_cast<tbox::CacheTetra4 *>(cache)->dff)[k][i][j];
            //else if (element_type==static_cast<int>(tbox::ELTYPE::TRI3))
            //    tmp2(j) = (static_cast<tbox::CacheTri3*>(cache)->dff)[k][i][j];
            //else if (element_type==static_cast<int>(tbox::ELTYPE::QUAD4))
            //    tmp2(j) = (static_cast<tbox::CacheQuad4*>(cache)->dff)[k][i][j];
        }

        Jn = tmp2.ddotproduct(tmp);
        J += Jn;
    }

    double detJ = J.det();
    return detJ;
}

template <typename scalar, int element_type>
tbox::CacheHex8 &ElementComputation<scalar, element_type>::trilinosHex8GetCache()
{
    static tbox::Lazy<tbox::CacheHex8> cache(2);
    return cache.get();
}

template <typename scalar, int element_type>
tbox::CacheTetra4 &ElementComputation<scalar, element_type>::trilinosTetra4GetCache()
{
    static tbox::Lazy<tbox::CacheTetra4> cache(2);
    return cache.get();
}

template <typename scalar, int element_type>
tbox::CacheTri3 &ElementComputation<scalar, element_type>::trilinosTri3GetCache()
{
    static tbox::Lazy<tbox::CacheTri3> cache(2);
    return cache.get();
}

template <typename scalar, int element_type>
tbox::CacheQuad4 &ElementComputation<scalar, element_type>::trilinosQuad4GetCache()
{
    static tbox::Lazy<tbox::CacheQuad4> cache(2);
    return cache.get();
}
