template <typename scalar, int element_type, int element_size>
ElementVectors<scalar, element_type, element_size>::ElementVectors(local_ordinal_type e, Teuchos::RCP<katropton::discrete_problem::Domain<scalar>> domain, Teuchos::RCP<katropton::discrete_problem::Loads<scalar>> loads, size_t numPrimalDPN)
{
    tbox::Cache *cache;

    std::vector<tbox::Pt> gauss_p;
    std::vector<double> gauss_w;

    /* @todo could probably be refactored into cache = element->getVCache(); */
    if (element_type == static_cast<int>(tbox::ELTYPE::HEX8))
    {
        tbox::CacheHex8 &cachehex8 = this->trilinosHex8GetCache();
        cache = &cachehex8;

        gauss_p = cache->getVGauss().p;
        gauss_w = cache->getVGauss().w;
    }
    else if (element_type == static_cast<int>(tbox::ELTYPE::TETRA4))
    {
        tbox::CacheTetra4 &cachetetra4 = this->trilinosTetra4GetCache();
        cache = &cachetetra4;

        gauss_p = cache->getVGauss().p;
        gauss_w = cache->getVGauss().w;
    }
    else if (element_type == static_cast<int>(tbox::ELTYPE::TRI3))
    {
        tbox::CacheTri3 &cachetri3 = this->trilinosTri3GetCache();
        cache = &cachetri3;

        gauss_p = cache->getVGauss().p;
        gauss_w = cache->getVGauss().w;
    }
    else if (element_type == static_cast<int>(tbox::ELTYPE::QUAD4))
    {
        tbox::CacheQuad4 &cachequad4 = this->trilinosQuad4GetCache();
        cache = &cachequad4;

        gauss_p = cache->getVGauss().p;
        gauss_w = cache->getVGauss().w;
    }

    N = Kokkos::View<scalar **, Kokkos::LayoutRight>("A", element_size, element_size);
    tVector<double, element_size> ff;
    tMatrix<double, 3, 3> J;

    // Looping on gauss points
    for (auto a = 0; a < gauss_p.size(); ++a)
    {
        for (auto j = 0; j < element_size; ++j)
            if (element_type == static_cast<int>(tbox::ELTYPE::HEX8))
                ff(j) = (static_cast<tbox::CacheHex8 *>(cache)->ff)[a][j];
            else if (element_type == static_cast<int>(tbox::ELTYPE::TETRA4))
                ff(j) = (static_cast<tbox::CacheTetra4 *>(cache)->ff)[a][j];
            else if (element_type == static_cast<int>(tbox::ELTYPE::TRI3))
                ff(j) = (static_cast<tbox::CacheTri3 *>(cache)->ff)[a][j];
            else if (element_type == static_cast<int>(tbox::ELTYPE::QUAD4))
                ff(j) = (static_cast<tbox::CacheQuad4 *>(cache)->ff)[a][j];

        double detJ = this->buildJ(a, J, e, *(domain->elements_list), *(domain->nodes_list));
        //std::cout << detJ << std::endl;

        for (auto i = 0; i < element_size; ++i)
            for (auto j = 0; j < element_size; ++j)
            {
                N(i, j) += ff(i) * ff(j) * gauss_w[a] * detJ;
                //std::cout << "i: "<< i<<", j: " << j << ", ff(i): " << ff(i) << ", ff(j): " << ff(j) << ", gauss_w[a]: " << gauss_w[a] << ", detJ: " << detJ << std::endl;
            }
    }
}
