#include "Elements_list.h"

#include "wProblem.h"
#include "wTag.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"
#include "wMedium.h"

using namespace katropton;
using namespace discrete_problem;

Elements_list::Elements_list()
{
    elements_number = 0;
    elements_type = 0;
    elements_size = 0;
    elements_material = 0;
    elements_nodes = 0;

    elements = Kokkos::View<int **, Kokkos::LayoutRight>("R", elements_number, 0);
}

Elements_list::Elements_list(Problem &pbl, Teuchos::RCP<Map> map)
{
    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;

    elements_number = pbl.msh->elems.size();
    local_ordinal_type materials_number = pbl.media.size();
    elements_type = 0;
    elements_size = 1;
    elements_material = 2;
    elements_nodes = 3;

    elements = Kokkos::View<int **, Kokkos::LayoutRight>("R", elements_number, elements_nodes + 8);

    for (auto i = 0; i < elements_number; ++i)
    {
        elements(i, elements_type) = static_cast<int>(pbl.msh->elems[i]->type());
        elements(i, elements_size) = pbl.msh->elems[i]->nodes.size();
        for (auto j = 0; j < materials_number; ++j)
            if (pbl.msh->elems[i]->ptag == (pbl.media[j]->tag))
            {
                elements(i, elements_material) = j;
                break;
            }
        for (auto j = 0; j < pbl.msh->elems[i]->nodes.size(); ++j)
            elements(i, elements_nodes + j) = map->mapNodesWO->getLocalElement(pbl.msh->elems[i]->nodes[j]->row);

        /*
        if (i==17)
        {
            std::cout<<"Element " << i << ": " << elements(i,elements_nodes+0) << " " << elements(i,elements_nodes+1) << " " << elements(i,elements_nodes+2) << " " << elements(i,elements_nodes+3) << std::endl;
            std::cout<<"Element " << i << ": " << pbl.msh->elems[i]->nodes[0]->row << " " << pbl.msh->elems[i]->nodes[1]->row << " " << pbl.msh->elems[i]->nodes[2]->row << " " << pbl.msh->elems[i]->nodes[3]->row << std::endl;
            std::cout<<"Element " << i << ": " << map->mapNodesWO->getLocalElement(pbl.msh->elems[i]->nodes[0]->row) << " " << map->mapNodesWO->getLocalElement(pbl.msh->elems[i]->nodes[1]->row) << " " << map->mapNodesWO->getLocalElement(pbl.msh->elems[i]->nodes[2]->row) << " " << map->mapNodesWO->getLocalElement(pbl.msh->elems[i]->nodes[3]->row) << std::endl;
        }
        */
    }
}
