#ifndef WELEMENTS_LIST_H
#define WELEMENTS_LIST_H

#include "katropton.h"

#include "Map.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Kokkos_ViewFactory.hpp>

typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;

namespace katropton
{
namespace discrete_problem
{
class Elements_list
{
private:
    Kokkos::View<int **, Kokkos::LayoutRight> elements;

    Tpetra::Vector<>::local_ordinal_type elements_number;
    Tpetra::Vector<>::local_ordinal_type elements_type;
    Tpetra::Vector<>::local_ordinal_type elements_size;
    Tpetra::Vector<>::local_ordinal_type elements_material;
    Tpetra::Vector<>::local_ordinal_type elements_nodes;

public:
    Elements_list();
    Elements_list(Problem &pbl, Teuchos::RCP<Map> map);

    inline local_ordinal_type Get_Element_Number(void) const { return elements_number; };
    inline local_ordinal_type Get_Element_Type(const local_ordinal_type e) const { return elements(e, elements_type); };
    inline local_ordinal_type Get_Element_Size(const local_ordinal_type e) const { return elements(e, elements_size); };
    inline local_ordinal_type Get_Element_Material(const local_ordinal_type e) const { return elements(e, elements_material); };
    inline local_ordinal_type Get_Element_Node(const local_ordinal_type e, local_ordinal_type i) const { return elements(e, elements_nodes + i); };
};
}; // namespace discrete_problem
}; // namespace katropton

#endif //WELEMENTS_LIST_H
