#include "IterativeSolver.h"
#include "wTimer.h"
#include "wProblem.h"
#include "wTag.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"
#include "wMedium.h"
#include "wuDirichlet.h"
#include "wContact.h"
#include "wPt.h"
#include <map>
#include "wTag.h"
#include <math.h>
#include "wHex8.h"
#include "wQuad4.h"
#include "wTri3.h"
#include "wCacheHex8.h"
#include "wGaussHex8.h"
#include "wResults.h"

#include "RandomVariable.h"

#include "wElemMatrix.h"
#include "mortar.h"
#include "wMshConvert.h"

//#include "wPreconditioner.h"

#include "wDisplayHook.h"

#include <math.h>

#include <mpi.h>

//#include <tbb/tbb.h>
//#include <tbb/parallel_for.h>
//#include <tbb/blocked_range.h>
//#include <tbb/task_scheduler_init.h>
//#include <tbb/spin_mutex.h>

#include <cstring>
#include <sstream>
#include <iostream>
#include <locale> // std::locale

// Tpetra
#include "Stokhos_Tpetra_MP_Vector.hpp"
#include "Stokhos_Tpetra_Utilities_MP_Vector.hpp"
#include "Tpetra_ConfigDefs.hpp"

#include "Tpetra_Map.hpp"
#include "Tpetra_MultiVector.hpp"
#include "Tpetra_Vector.hpp"
#include "Tpetra_CrsGraph.hpp"
#include "Tpetra_CrsMatrix.hpp"
#include "Stokhos_Tpetra_CG.hpp"
//#include "Sacado_Fad_MP_Vector.hpp"

#include <Teuchos_Array.hpp>
#include <Teuchos_GlobalMPISession.hpp>
#include <Teuchos_oblackholestream.hpp>
#include <Teuchos_ScalarTraits.hpp>
#include <Teuchos_RCP.hpp>
#include "Teuchos_ParameterList.hpp"

#include "Stokhos_Amesos2_MP_Vector.hpp"
#include "Amesos2_Factory.hpp"

#include "Stokhos_Ifpack2_MP_Vector.hpp"
#include "Ifpack2_Factory.hpp"

#include "Belos_Tpetra_MP_Vector.hpp"
#include "BelosStatusTestCombo.hpp"

#include <Xpetra_BlockedCrsMatrix.hpp>
#include <Xpetra_Matrix.hpp>
#include <BelosXpetraAdapter.hpp> // => This header defines Belos::XpetraOp
#include <BelosMueLuAdapter.hpp>  // => This header defines Belos::MueLuOp
#include <Xpetra_TpetraMultiVector.hpp>

//#include "BelosGmresSolMgr.hpp"

#include "Kokkos_ViewFactory.hpp"
#include "Kokkos_Vector.hpp"

#include <Epetra_Map.h>
#include "Epetra_MpiComm.h"

#include "BelosConfigDefs.hpp"

#include "Map.h"
#include "Discrete_problem.h"

#include "Preconditioner.h"

#include <fstream>
/*
#include "BelosTpetraAdapter.hpp"
#include "BelosConfigDefs.hpp"
#include "BelosLinearProblem.hpp"
#include "BelosPseudoBlockGmresSolMgr.hpp"
#include "BelosPseudoBlockCGSolMgr.hpp"
#include "BelosBlockGmresSolMgr.hpp"


// MueLu main header: include most common header files in one line
#include <MueLu.hpp>
#include <MueLu_TpetraOperator.hpp>
#include <MueLu_CreateTpetraPreconditioner.hpp>
#include <MueLu_Utilities.hpp>
*/

#include <MueLu.hpp>

#include <MueLu_Exceptions.hpp>
#include <MueLu_Hierarchy.hpp>
#include <MueLu_MasterList.hpp>
#include <MueLu_MLParameterListInterpreter.hpp>
#include <MueLu_ParameterListInterpreter.hpp>
#include <MueLu_Utilities.hpp>
#include <MueLu_HierarchyUtils.hpp>
//#include <MueLu_ParameterListInterpreter_decl.hpp>

#include "ElementMatrices.h"

#include "Nodes_list.h"

#ifdef max
#undef max // pour std::numeric_limits<std::streamsize>::max()
#endif

using namespace katropton;
using namespace fwk;

#define HAVE_BELOS_TPETRA_TIMERS

template <typename scalar>
Teuchos::ParameterList convertParamList(Teuchos::ParameterList &paramList)
{
    for (Teuchos::ParameterList::ConstIterator i = paramList.begin(); i != paramList.end(); ++i)
    {
        const std::string &name_i = paramList.name(i);
        //std::cout << name_i << std::endl;

        if (Teuchos::isParameterType<double>(paramList, name_i))
        {
            //std::cout << name_i << std::endl;
            double value = Teuchos::getParameter<double>(paramList, name_i);
            std::string name_ii = name_i;

            //std::cout << paramList << std::endl;
            //paramList.remove(name_i);
            //std::cout << paramList << std::endl;

            std::cout << "try to set a scalar parameter" << std::endl;
            paramList.set<scalar>(name_ii, ((scalar)value));
            //std::cout << paramList << std::endl;
        }
        else if (Teuchos::isParameterType<Teuchos::ParameterList>(paramList, name_i))
        {
            Teuchos::ParameterList subparamList = Teuchos::getParameter<Teuchos::ParameterList>(paramList, name_i);
            subparamList = convertParamList<scalar>(subparamList);
            paramList.set(name_i, subparamList);
        }
    }

    return paramList;
}

void IterativeSolver::write(std::ostream &out) const
{
    out << "katropton::Solver:\n";
    out << pbl; // wProblem.h
}

void IterativeSolver::start()
{
    std::locale::global(std::locale::classic());
    /*

    #ifdef STOKHOS_SACADO_KOKKOS_HPP
        std::cout<<"defined STOKHOS_SACADO_KOKKOS_HPP"<<std::endl;
    #else
        std::cout<<"undefined STOKHOS_SACADO_KOKKOS_HPP"<<std::endl;
    #endif

    #ifdef HAVE_STOKHOS_ENSEMBLE_SCALAR_TYPE
        std::cout<<"defined HAVE_STOKHOS_ENSEMBLE_SCALAR_TYPE"<<std::endl;
    #else
        std::cout<<"undefined HAVE_STOKHOS_ENSEMBLE_SCALAR_TYPE"<<std::endl;
    #endif

    #ifdef STOKHOS_SACADO_KOKKOS_MP_VECTOR_HPP
        std::cout<<"defined STOKHOS_SACADO_KOKKOS_MP_VECTOR_HPP"<<std::endl;
    #else
        std::cout<<"undefined STOKHOS_SACADO_KOKKOS_MP_VECTOR_HPP"<<std::endl;
    #endif

    #ifdef STOKHOS_STATIC_STORAGE_HPP
        std::cout<<"defined STOKHOS_STATIC_STORAGE_HPP"<<std::endl;
    #else
        std::cout<<"undefined STOKHOS_STATIC_STORAGE_HPP"<<std::endl;
    #endif

    #ifdef STOKHOS_STATIC_FIXED_STORAGE_HPP
        std::cout<<"defined STOKHOS_STATIC_FIXED_STORAGE_HPP"<<std::endl;
    #else
        std::cout<<"undefined STOKHOS_STATIC_FIXED_STORAGE_HPP"<<std::endl;
    #endif

    #ifdef STOKHOS_STATIC_ARRAY_TRAITS_HPP
        std::cout<<"defined STOKHOS_STATIC_ARRAY_TRAITS_HPP"<<std::endl;
    #else
        std::cout<<"undefined STOKHOS_STATIC_ARRAY_TRAITS_HPP"<<std::endl;
    #endif

    #ifdef SACADO_MP_VECTOR_HPP
        std::cout<<"defined SACADO_MP_VECTOR_HPP"<<std::endl;
    #else
        std::cout<<"undefined SACADO_MP_VECTOR_HPPP"<<std::endl;
    #endif

    #ifdef HAVE_STOKHOS_SACADO
        std::cout<<"defined HAVE_STOKHOS_SACADO"<<std::endl;
    #else
        std::cout<<"undefined HAVE_STOKHOS_SACADO"<<std::endl;
    #endif
    */

    //Kokkos::OpenMP::initialize (nthreads);
    //Kokkos::OpenMP::initialize (1);
    {

        if (ensemble_propagation)
        {
            std::cout << "Ensemble propagation enable" << std::endl;
            /*
            if (minimum_ensemble_size <= 1)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,1,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            //else
            if (minimum_ensemble_size <= 2)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,2,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            */
            //else
            /*
            if (minimum_ensemble_size <= 4)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,4,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            */
            /*else if (minimum_ensemble_size <= 6)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,6,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            else
    */
            //else

            if (minimum_ensemble_size <= 8)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int, double, 8, Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }

            /*      else if (minimum_ensemble_size <= 10)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,10,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            */
            /*
            else if (minimum_ensemble_size <= 12)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,12,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            */
            /*
            else if (minimum_ensemble_size <= 14)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,14,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            */

            else if (minimum_ensemble_size <= 16)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int, double, 16, Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }

            /*

            else if (minimum_ensemble_size <= 18)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,18,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            */
            /*
            else if (minimum_ensemble_size <= 20)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,20,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            */
            /*
            else if (minimum_ensemble_size <= 22)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,22,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            */
            /*
            else if (minimum_ensemble_size <= 24)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,24,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar, Kokkos::OpenMP>();
            }
            */
            /*
            else if (minimum_ensemble_size <= 26)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,26,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            */
            /*
            else if (minimum_ensemble_size <= 28)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,28,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            */
            /*
            else if (minimum_ensemble_size <= 30)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,30,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            */

            else if (minimum_ensemble_size <= 32)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int, double, 32, Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }

            /*
            else if (minimum_ensemble_size <= 64)
            {
                typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,64,Kokkos::OpenMP>> ensemble_scalar;
                IterativeSolver::tstart<ensemble_scalar>();
            }
            */
        }
        else
        {
            std::cout << "Ensemble propagation disable" << std::endl;
            IterativeSolver::tstart<double>();
        }
        std::cout << "Number of used threads: " << nthreads << std::endl;
    }
    //Kokkos::Threads::finalize ();

    //Kokkos::finalize ();
}

template <typename scalar>
void IterativeSolver::tstart()
{
    LinearSolver::getTimers().reset();
    // ------------------------------------------
    //
    //         Trilinos initialization
    //
    // ------------------------------------------
    using std::cerr;
    using std::cout;
    using std::endl;
    using Teuchos::arcp;
    using Teuchos::Array;
    using Teuchos::ArrayRCP;
    using Teuchos::ArrayView;
    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;

    using Kokkos::ALL;
    using Kokkos::subview;

    typedef discrete_problem::Map::local_ordinal_type local_ordinal_type;
    typedef discrete_problem::Map::global_ordinal_type global_ordinal_type;
    typedef discrete_problem::Map::node_type node_type;

    typedef typename discrete_problem::Vectors<scalar>::tpetra_vector_type tpetra_vector_type;
    typedef typename discrete_problem::Vectors<scalar>::tpetra_mvector_type tpetra_mvector_type;
    typedef typename discrete_problem::Vectors<scalar>::xpetra_mvector_type xpetra_mvector_type;

    typedef typename discrete_problem::Matrices<scalar>::tpetra_crs_type tpetra_crs_type;

    typedef Tpetra::Export<local_ordinal_type, global_ordinal_type, node_type> export_type;
    typedef Tpetra::Import<local_ordinal_type, global_ordinal_type, node_type> import_type;

    typedef MueLu::Hierarchy<scalar, local_ordinal_type, global_ordinal_type, node_type> Hierarchy;
    typedef MueLu::ParameterListInterpreter<scalar, local_ordinal_type, global_ordinal_type, node_type> ParameterListInterpreter;

    typedef Teuchos::ScalarTraits<scalar> STS;
    typedef typename STS::magnitudeType magnitude_type;

    Teuchos::oblackholestream blackhole;

    RCP<const Teuchos::Comm<int>> comm = rcp(new Teuchos::MpiComm<int>(pbl.comm));

    const size_t myRank = comm->getRank();

    // Make an output stream (for verbose output) that only prints on
    // Proc 0 of the communicator.
    Teuchos::oblackholestream blackHole;
    std::ostream &out = (myRank == 0) ? std::cout : blackHole;

    // Initiate Kokkos with the given number  of threads
    Kokkos::InitArguments args;
    args.num_threads = nthreads;

    // Create some timers used by Belos so we can access them
    Teuchos::RCP<Teuchos::Time> time_mat_vec =
        Teuchos::TimeMonitor::getNewTimer("Belos: Operation Op*x");
    Teuchos::RCP<Teuchos::Time> time_prec_apply =
        Teuchos::TimeMonitor::getNewTimer("Belos: Operation Prec*x");
    Teuchos::RCP<Teuchos::Time> time_prec_setup =
        Teuchos::TimeMonitor::getNewTimer("Total MueLu setup time");

    std::string ortho_strategy = solverList->get("Orthogonalization", "DGKS");
    std::string belos_ortho_timer_name = "Belos: " + ortho_strategy;

    if (ortho_strategy == "IMGS")
        belos_ortho_timer_name = belos_ortho_timer_name + "[1]";
    else
        belos_ortho_timer_name = belos_ortho_timer_name + "[2]";

    Teuchos::RCP<Teuchos::Time> time_ortho =
        Teuchos::TimeMonitor::getNewTimer(belos_ortho_timer_name + ": Orthogonalization");
    Teuchos::RCP<Teuchos::Time> time_ortho_inner =
        Teuchos::TimeMonitor::getNewTimer(belos_ortho_timer_name + ": Ortho (Inner Product)");
    Teuchos::RCP<Teuchos::Time> time_ortho_norm =
        Teuchos::TimeMonitor::getNewTimer(belos_ortho_timer_name + ": Ortho (Norm)");
    Teuchos::RCP<Teuchos::Time> time_ortho_update =
        Teuchos::TimeMonitor::getNewTimer(belos_ortho_timer_name + ": Ortho (Update)");

    Teuchos::RCP<Teuchos::Time> time_belos_total =
        Teuchos::TimeMonitor::getNewTimer("Belos: PseudoBlockGmresSolMgr total solve time");

    size_t assembly_nthreads = solverList->get("Matrix assembly: number of threads", nthreads);

    bool write_intermediate_files = solverList->get("Write intermediate files", false);
    bool write_txt_files = solverList->get("Write txt files", false);

#ifdef KOKKOS_ENABLE_DEPRECATED_CODE
    Kokkos::OpenMP::initialize(assembly_nthreads);
#else
    Kokkos::OpenMP::impl_initialize(assembly_nthreads);
#endif
    {
        LinearSolver::getTimers()["total"].start();
        auto msh = pbl.msh;

        // assign a global row for each global node
        for (auto n : msh->nodes)
            n->row = n->no - 1;

        auto numNodes = msh->nodes.size();

        RCP<Teuchos::ParameterList> randomParams = rcp(new Teuchos::ParameterList(solverList->get("randomParams", Teuchos::ParameterList())));
        const int n_rv = pbl.RandomVariables.size();
        scalar m_rv_values[n_rv];
        for (int i = 0; i < n_rv; ++i)
        {
            typedef EnsembleTraits<scalar> ET;
            const int ensemble_size = ET::size;

            for (int j = 0; j < ensemble_size; ++j)
                ET::coeff(m_rv_values[i], j) = (double)pbl.RandomVariables[i]->values[j];
        }

        LinearSolver::getTimers()["dp"].start();
        Teuchos::RCP<Discrete_problem<scalar>> discrete_problem = Teuchos::rcp(new Discrete_problem<scalar>(pbl, numPrimalDPN, randomParams, m_rv_values));
        LinearSolver::getTimers()["dp"].stop();

        out << "Before matrix computation" << std::endl;
        LinearSolver::getTimers()["computeMatrices"].start();
        if (discrete_problem->domain->contacts_list->getContactNumber() >= 1)
            discrete_problem->computeContactMatrices(pbl.comm);
        discrete_problem->computeMatrices();
        LinearSolver::getTimers()["computeMatrices"].stop();
        out << "After matrix computation" << std::endl;

        //std::cout << "discrete_problem->algebraic->map->lm_to_dof " << discrete_problem->algebraic->map->lm_to_dof.size() << std::endl;

        LinearSolver::getTimers()["computeMatrices"].start();
        //if (discrete_problem->domain->contacts_list->getContactNumber() >= 1)
        //out<< *(discrete_problem->algebraic->matrices->B) << std::endl;

        bool writeMatrixAndVectors = solverList->get("Write matrix and vectors", false);

        if (writeMatrixAndVectors)
        {
            std::string name_0;
            name_0 = "matrix_before_bc.text";

            LinearSolver::getTimers()["write"].start();
            discrete_problem->algebraic->writeCrsMatrixToFile(name_0, myRank);
            Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeSparseFile(std::string("A_mm.txt"), discrete_problem->algebraic->matrices->A);
            if (discrete_problem->domain->contacts_list->getContactNumber() >= 1)
            {
                Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeSparseFile(std::string("B_mm.txt"), discrete_problem->algebraic->matrices->B);
                Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeSparseFile(std::string("B_T_mm.txt"), discrete_problem->algebraic->matrices->B_T);
            }
            LinearSolver::getTimers()["write"].stop();
        }

        LinearSolver::getTimers()["computeLoads"].start();
        discrete_problem->computeLoads();
        LinearSolver::getTimers()["computeLoads"].stop();

        if (discrete_problem->domain->contacts_list->getContactNumber() >= 1)
        {
            scalar gamma = solverList->get("gamma", 0.);
            discrete_problem->createBlockMatrix(gamma, writeMatrixAndVectors);
            //out<< *(discrete_problem->algebraic->matrices->blockedMatrix) << std::endl;
        }

        //std::cout << "discrete_problem->algebraic->map->lm_to_dof " << discrete_problem->algebraic->map->lm_to_dof.size() << std::endl;

        LinearSolver::getTimers()["computeMatrices"].stop();

        LinearSolver::getTimers()["computeLoads"].start();
        //discrete_problem->computeLoads();
        if (discrete_problem->domain->contacts_list->getContactNumber() >= 1)
        {
            discrete_problem->createBlockMVector(writeMatrixAndVectors);
            //std::cout << *(discrete_problem->algebraic->vectors->rhsBlockedMultiVector) <<std::endl;
        }
        LinearSolver::getTimers()["computeLoads"].stop();

        if (writeMatrixAndVectors)
        {
            std::string name_1, name_2, name_3;
            name_1 = "matrix_after_bc.text";
            name_2 = "matrix_after_bc";
            name_3 = "vector_after_bc.text";

            LinearSolver::getTimers()["write"].start();
            //discrete_problem->algebraic->writeCrsMatrixToFile(name_1,myRank);
            Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeSparseFile(std::string("A_mm_after_bc.txt"), discrete_problem->algebraic->matrices->A);
            Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("b_mm_after_bc.txt"), discrete_problem->algebraic->vectors->b);
            //discrete_problem->algebraic->writeLocalCrsMatrixToFile(name_2,myRank);
            //discrete_problem->algebraic->writeVectorToFile(name_3,myRank);
            LinearSolver::getTimers()["write"].stop();
        }

        std::string name_4, defaultName, name_x;
        defaultName = "none";
        name_4 = solverList->get("Initial guess", defaultName);
        ;
        name_x = "solution";

        RCP<export_type> exportOutput;
        RCP<import_type> importOutput, importOutputMM;
        RCP<tpetra_vector_type> xPost, xPostMM;
        if (discrete_problem->domain->contacts_list->getContactNumber() >= 1)
        {
            importOutput = rcp(new import_type(discrete_problem->algebraic->map->fullmap, discrete_problem->algebraic->map->fullmapWO));
            importOutputMM = rcp(new import_type(discrete_problem->algebraic->map->fullmap, discrete_problem->algebraic->map->fullmapOutput));
            //exportOutput = rcp (new export_type(discrete_problem->algebraic->map->fullmap,discrete_problem->algebraic->map->fullmapOutput));
            xPost = rcp(new tpetra_vector_type(discrete_problem->algebraic->map->fullmapWO, false));
            xPostMM = rcp(new tpetra_vector_type(discrete_problem->algebraic->map->fullmapOutput, false));
        }
        else
        {
            //exportOutput = rcp (new export_type(discrete_problem->algebraic->map->mapDofs,discrete_problem->algebraic->map->mapDofsWO));
            importOutput = rcp(new import_type(discrete_problem->algebraic->map->mapDofs, discrete_problem->algebraic->map->mapDofsWO));
            importOutputMM = rcp(new import_type(discrete_problem->algebraic->map->mapDofs, discrete_problem->algebraic->map->mapOutput));
            xPost = rcp(new tpetra_vector_type(discrete_problem->algebraic->map->mapDofsWO, false));
            xPostMM = rcp(new tpetra_vector_type(discrete_problem->algebraic->map->mapOutput, false));
        }

        if (name_4.compare(defaultName) != 0)
            discrete_problem->algebraic->vectors->x = discrete_problem->algebraic->readVectorFromFile(name_4, myRank);

        RCP<tpetra_vector_type> r(new tpetra_vector_type(discrete_problem->algebraic->map->mapDofs, true));

        //r = discrete_problem->algebraic->vectors->b;

        discrete_problem->algebraic->matrices->A->apply(*(discrete_problem->algebraic->vectors->x), *r, Teuchos::NO_TRANS, (scalar)-1., (scalar)1.);

        if (writeMatrixAndVectors)
        {
            LinearSolver::getTimers()["write"].start();
            std::string name_5 = "x_0.text";
            //discrete_problem->algebraic->writeVectorToFile(name_5,myRank,discrete_problem->algebraic->vectors->x);
            Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string(name_5), discrete_problem->algebraic->vectors->x);
            std::string name_6 = "r_0.text";
            //discrete_problem->algebraic->writeVectorToFile(name_6,myRank,r);
            Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string(name_6), r);
            LinearSolver::getTimers()["write"].stop();
        }

        bool usePrec = solverList->get("Use preconditioner", false);

        //LinearSolver::getTimers()["belos"].start();

        std::string belos_output_file = solverList->get("Output Stream file name belos", "belos_out.txt");

        std::ofstream belos_ofstream(belos_output_file);

        RCP<Teuchos::FancyOStream> verbOut = Teuchos::getFancyOStream(Teuchos::rcpFromRef(belos_ofstream));
        solverList->set("Output Stream", verbOut->getOStream());

        solverList->set("Convergence Tolerance", (magnitude_type)solverList->get("Ensemble Convergence Tolerance", 1E-08));

        size_t active_set_iteration = 0;
        size_t max_active_set_iteration = solverList->get("Maximum active set iteration", 1);

        LinearSolver::getTimers()["solver"].start();
        while (true)
        {

            // Solve
            //typedef typename Sacado::ValueType<scalar>::type belos_scalar;
            RCP<tpetra_vector_type> xx;

            if (discrete_problem->domain->contacts_list->getContactNumber() >= 1)
            {
#if 0
                {
                    std::ofstream outputfile;
                    std::string name = "lm_to_dof";
                    outputfile.open (name + "_" + std::to_string(discrete_problem->algebraic->map->mapDofs->getComm()->getRank()) + ".txt");
                    for (auto i = discrete_problem->algebraic->map->lm_to_dof.begin(); i != discrete_problem->algebraic->map->lm_to_dof.end(); ++i)
                        outputfile << i->second << std::endl;
                        //outputfile << i->first << " : " << i->second << std::endl;;
                    outputfile.close();
                }
#endif

                RCP<xpetra_mvector_type> nullspace1 = Teuchos::null;
                RCP<xpetra_mvector_type> nullspace2 = Teuchos::null;

                string xmlFileName = solverList->get("MueLu xml file name", " ");

                bool doIconver = solverList->get("convert MueLu xml file", false);

                RCP<Teuchos::ParameterList> paramList = Teuchos::getParametersFromXmlFile(xmlFileName);

                Teuchos::ParameterList paramList2;

                RCP<std::map<local_ordinal_type, local_ordinal_type>> lagr2dof = Teuchos::rcpFromRef(discrete_problem->algebraic->map->lm_to_dof);

                RCP<Teuchos::ParameterList> paramsF, paramsI;
                paramsF = sublist(paramList, "Factories");
                paramsI = sublist(paramsF, "myInterfaceAggs2");
                paramsI->set<RCP<std::map<local_ordinal_type, local_ordinal_type>>>("DualNodeID2PrimalNodeID - level 0", lagr2dof);

                if (doIconver)
                {
                    //out << *paramList << std::endl;
                    //out << "start the parameter list conversion" << std::endl;
                    paramList2 = convertParamList<scalar>(*paramList);
                    //out << "end the parameter list conversion" << std::endl;
                    //out << paramList2 << std::endl;
                }
                else
                    paramList2 = (*paramList);

                RCP<Tpetra::MultiVector<double>> inCoords(new Tpetra::MultiVector<double>(discrete_problem->algebraic->map->mapNodesWO, 3, true));
                for (auto i = 0; i < discrete_problem->algebraic->map->mapNodesWO->getNodeNumElements(); ++i)
                {
                    auto global_index = discrete_problem->algebraic->map->mapNodesWO->getGlobalElement(i);
                    for (auto j = 0; j < 3; ++j)
                        inCoords->replaceGlobalValue(global_index, j, pbl.msh->nodes[global_index]->pos.x[j]);
                }

                ParameterListInterpreter mueLuFactory(paramList2, comm);
                RCP<Hierarchy> H = mueLuFactory.CreateHierarchy();
                H->GetLevel(0)->Set("A", Teuchos::rcp_dynamic_cast<Xpetra::Matrix<scalar, local_ordinal_type, global_ordinal_type, node_type>>(discrete_problem->algebraic->matrices->blockedMatrix));
                //H->GetLevel(0)->Set("Nullspace1",   nullspace1); // null space for displacement dofs (rigid body modes)
                //H->GetLevel(0)->Set("Nullspace2",   nullspace2); // null space for Lagrange multipliers (just standard constant vectors)

                //std::cout << "discrete_problem->algebraic->map->lm_to_dof " << discrete_problem->algebraic->map->lm_to_dof.size() << std::endl;

                H->IsPreconditioner(true);
                //H->GetLevel(0)->Set("Coordinates", inCoords);
                H->SetDefaultVerbLevel(MueLu::Extreme);

                mueLuFactory.SetupHierarchy(*H);

                typedef Xpetra::MultiVector<scalar, local_ordinal_type, global_ordinal_type, node_type> MV;
                typedef typename tpetra_mvector_type::dot_type belos_scalar;
                typedef Belos::OperatorT<MV> OP;

                typedef Belos::LinearProblem<belos_scalar, MV, OP> BLinProb;
                RCP<OP> belosOp = rcp(new Belos::XpetraOp<scalar, local_ordinal_type, global_ordinal_type, node_type>(discrete_problem->algebraic->matrices->blockedMatrix)); // Turns a Xpetra::Matrix object into a Belos operator
                RCP<OP> belosPrec = rcp(new Belos::MueLuOp<scalar, local_ordinal_type, global_ordinal_type, node_type>(H));                                                   // Turns a MueLu::Hierarchy object into a Belos operator

                RCP<BLinProb> blinproblem = rcp(new BLinProb(belosOp, discrete_problem->algebraic->vectors->solutionMultiVector, discrete_problem->algebraic->vectors->rhsMultiVector));

                blinproblem->setRightPrec(belosPrec);
                blinproblem->setProblem();
                RCP<Teuchos::ParameterList> belosParams = solverList; //rcp(new Teuchos::ParameterList(solverList));
                RCP<Belos::SolverManager<belos_scalar, MV, OP>> blinsolver =
                    rcp(new Belos::PseudoBlockGmresSolMgr<belos_scalar, MV, OP>(blinproblem, belosParams));
                out << "Belos informations are written in file " << belos_output_file << endl;

                if (writeMatrixAndVectors)
                {
                    LinearSolver::getTimers()["write"].start();
                    RCP<tpetra_mvector_type> mxx = (Teuchos::rcp_dynamic_cast<Xpetra::TpetraMultiVector<scalar, local_ordinal_type, global_ordinal_type, node_type>>(discrete_problem->algebraic->vectors->solutionMultiVector))->getTpetra_MultiVector();
                    RCP<tpetra_mvector_type> mbb = (Teuchos::rcp_dynamic_cast<Xpetra::TpetraMultiVector<scalar, local_ordinal_type, global_ordinal_type, node_type>>(discrete_problem->algebraic->vectors->rhsMultiVector))->getTpetra_MultiVector();
                    Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("blocked_x_0_mm.txt"), mxx->getVectorNonConst(0));
                    Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("blocked_b_mm.txt"), mbb->getVectorNonConst(0));
                    LinearSolver::getTimers()["write"].stop();
                }

                //RCP<Belos::StatusTestGenResNorm<belos_scalar,MV,OP> > ensembleStatusTest = rcp(new Belos::StatusTestGenResNorm<belos_scalar, MV, OP>((scalar) solverList->get("Ensemble Convergence Tolerance",1E-08)));

                //blinsolver->setUserConvStatusTest(ensembleStatusTest);

                //blinsolver->setUserConvStatusTest(Belos::StatusTestGenResNorm<belos_scalar, MV, OP>(), Belos::StatusTestCombo<belos_scalar,MV,OP>::StatusTestCombo(Belos::ComboType::AND));

                if (solverList->get("Create Preconditioned Matrix", false))
                {
                    LinearSolver::getTimers()["create AM^{-1}"].start();
                    typedef Xpetra::TpetraMultiVector<scalar, local_ordinal_type, global_ordinal_type, node_type> TMV;

                    Teuchos::RCP<const Xpetra::Map<local_ordinal_type, global_ordinal_type, node_type>> merged_map = discrete_problem->algebraic->vectors->rhsMultiVector->getMap();

                    Teuchos::RCP<TMV> MVx1 = Teuchos::rcp(new TMV(merged_map, merged_map->getGlobalNumElements(), true));
                    Teuchos::RCP<TMV> MVx2 = Teuchos::rcp(new TMV(merged_map, merged_map->getGlobalNumElements(), true));
                    Teuchos::RCP<TMV> MVy1 = Teuchos::rcp(new TMV(merged_map, merged_map->getGlobalNumElements(), true));
                    Teuchos::RCP<TMV> MVy2 = Teuchos::rcp(new TMV(merged_map, merged_map->getGlobalNumElements(), true));

                    for (size_t local_i = 0; local_i < merged_map->getNodeNumElements(); ++local_i)
                        MVx1->sumIntoGlobalValue(merged_map->getGlobalElement(local_i), merged_map->getGlobalElement(local_i), ((scalar)1.));

                    belosPrec->Apply(*MVx1, *MVx2);
                    belosOp->Apply(*MVx2, *MVy2);
                    belosOp->Apply(*MVx1, *MVy1);

                    Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("Applied_Matrix_mm.txt"), MVy1->getTpetra_MultiVector());
                    Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("Applied_Preconditioner_mm.txt"), MVx2->getTpetra_MultiVector());
                    Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("Applied_Preconditioned_Matrix_mm.txt"), MVy2->getTpetra_MultiVector());

                    LinearSolver::getTimers()["create AM^{-1}"].stop();
                }

                out << "--------- Belos start" << endl;
                blinsolver->solve();
                out << "--------- Belos stop" << endl;

                //std::cout << "discrete_problem->algebraic->map->lm_to_dof " << discrete_problem->algebraic->map->lm_to_dof.size() << std::endl;

                RCP<tpetra_mvector_type> mxx = (Teuchos::rcp_dynamic_cast<Xpetra::TpetraMultiVector<scalar, local_ordinal_type, global_ordinal_type, node_type>>(discrete_problem->algebraic->vectors->solutionMultiVector))->getTpetra_MultiVector();
                xx = mxx->getVectorNonConst(0);

                if (write_intermediate_files)
                {
                    LinearSolver::getTimers()["write results"].start();

                    std::string name_5 = "x_sol" + std::to_string(active_set_iteration) + ".txt";
                    //discrete_problem->algebraic->writeVectorToFile(name_5,myRank,xx);
                    Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string(name_5), xx);
                    LinearSolver::getTimers()["write results"].stop();
                }
                /*
                discrete_problem->algebraic->vectors->x = (Teuchos::rcp_dynamic_cast<Xpetra::TpetraMultiVector<scalar, local_ordinal_type, global_ordinal_type, node_type>>
                    (discrete_problem->algebraic->vectors->solutionBlockedMultiVector->getMultiVector(0)))->getTpetra_MultiVector()->getVectorNonConst(0);

                std::string name_5 = "x_sol.text";
                discrete_problem->algebraic->writeVectorToFile(name_5,myRank,discrete_problem->algebraic->vectors->x);

                discrete_problem->algebraic->vectors->lagrange = (Teuchos::rcp_dynamic_cast<Xpetra::TpetraMultiVector<scalar, local_ordinal_type, global_ordinal_type, node_type>>
                    (discrete_problem->algebraic->vectors->solutionBlockedMultiVector->getMultiVector(1)))->getTpetra_MultiVector()->getVectorNonConst(0);
    */
            }
            else
            {
                typedef tpetra_mvector_type MV;
                typedef typename tpetra_mvector_type::dot_type belos_scalar;
                typedef Tpetra::Operator<scalar> OP;
                typedef Belos::LinearProblem<belos_scalar, MV, OP> BLinProb;
                RCP<BLinProb> blinproblem = rcp(new BLinProb(discrete_problem->algebraic->matrices->A, discrete_problem->algebraic->vectors->x, discrete_problem->algebraic->vectors->b));
                if (usePrec)
                {
                    //LinearSolver::getTimers()["belos"].stop();
                    //LinearSolver::getTimers()["prec"].start();
                    std::string muelu_output_file = solverList->get("Output Stream file name muelu", "muelu_out.txt");

                    bool useXML = solverList->get("use xml file", false);

                    string xmlFileName = solverList->get("MueLu xml file name", " ");

                    RCP<Teuchos::ParameterList> mueluParams;
                    if (useXML)
                        mueluParams = Teuchos::getParametersFromXmlFile(xmlFileName);
                    else
                        mueluParams = rcp(new Teuchos::ParameterList(solverList->get("mueluParams", Teuchos::ParameterList())));

                    RCP<Tpetra::MultiVector<double>> inCoords(new Tpetra::MultiVector<double>(discrete_problem->algebraic->map->mapNodesWO, 3, true));
                    for (auto i = 0; i < discrete_problem->algebraic->map->mapNodesWO->getNodeNumElements(); ++i)
                    {
                        auto global_index = discrete_problem->algebraic->map->mapNodesWO->getGlobalElement(i);
                        for (auto j = 0; j < 3; ++j)
                            inCoords->replaceGlobalValue(global_index, j, pbl.msh->nodes[global_index]->pos.x[j]);
                    }

                    RCP<Tpetra::MultiVector<scalar>> nullspace = Teuchos::null;
                    RCP<MueLu::TpetraOperator<scalar>> M;
                    M = mueluPreconditioner<scalar>(*mueluParams, discrete_problem->algebraic->matrices->A, inCoords, nullspace, muelu_output_file);
                    //LinearSolver::getTimers()["prec"].stop();
                    //LinearSolver::getTimers()["belos"].start();
                    blinproblem->setRightPrec(M);
                }

                blinproblem->setProblem();
                RCP<Teuchos::ParameterList> belosParams = solverList; //rcp(new Teuchos::ParameterList(solverList));
                RCP<Belos::SolverManager<belos_scalar, MV, OP>> blinsolver =
                    rcp(new Belos::PseudoBlockGmresSolMgr<belos_scalar, MV, OP>(blinproblem, belosParams));
                out << "Belos informations are written in file " << belos_output_file << endl;

                //RCP<Belos::StatusTestGenResNorm<belos_scalar,MV,OP> > ensembleStatusTest = rcp(new Belos::StatusTestGenResNorm<belos_scalar, MV, OP>((scalar) solverList->get("Ensemble Convergence Tolerance",1E-08)));

                //blinsolver->setUserConvStatusTest(ensembleStatusTest);

                //blinsolver->setUserConvStatusTest(Belos::StatusTestGenResNorm<belos_scalar, MV, OP>(), Belos::StatusTestCombo<belos_scalar,MV,OP>::StatusTestCombo(Belos::ComboType::AND));

                out << "--------- Belos start" << endl;
                blinsolver->solve();
                out << "--------- Belos stop" << endl;
            }

            //LinearSolver::getTimers()["belos"].stop();

            //this->belos_iter = (int) blinsolver->getNumIters ();
            if (discrete_problem->domain->contacts_list->getContactNumber() >= 1)
            {
                xPost->doImport(*xx, *importOutput, Tpetra::INSERT);
                xPostMM->doImport(*xx, *importOutputMM, Tpetra::INSERT);
                //Tpetra::MatrixMarket::Writer< tpetra_crs_type >::writeDenseFile(std::string("x_mm.txt"),xx);
            }
            else
            {
                xPost->doImport(*(discrete_problem->algebraic->vectors->x), *importOutput, Tpetra::INSERT);
                xPostMM->doImport(*(discrete_problem->algebraic->vectors->x), *importOutputMM, Tpetra::INSERT);
                //Tpetra::MatrixMarket::Writer< tpetra_crs_type >::writeDenseFile(std::string("x_mm.txt"),discrete_problem->algebraic->vectors->x);
            }
            ++active_set_iteration;

            if (write_intermediate_files)
            {
                LinearSolver::getTimers()["write results"].start();
                if (write_txt_files)
                    Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("x_mm.txt"), xPostMM);
                this->writeResults<scalar>(xPost, numNodes, name_x, active_set_iteration, false, false, discrete_problem->algebraic->map, discrete_problem->domain->random_field);
                LinearSolver::getTimers()["write results"].stop();
            }
            bool hasConverged = true;

            if (active_set_iteration == max_active_set_iteration)
            {
                if (discrete_problem->domain->contacts_list->getContactNumber() >= 1)
                {
                    LinearSolver::getTimers()["update"].start();
                    hasConverged = discrete_problem->updateActivity(active_set_iteration);
                    LinearSolver::getTimers()["update"].stop();

                    if (hasConverged == false)
                    {
                        std::cout << "--------------------------------------------------------" << std::endl;
                        std::cout << "---Has not converged before reaching maximum iteration---" << std::endl;
                        std::cout << "--------------------------------------------------------" << std::endl;
                    }
                }
                break;
            }
            else if (discrete_problem->domain->contacts_list->getContactNumber() >= 1)
            {
                LinearSolver::getTimers()["update"].start();
                hasConverged = discrete_problem->updateActivity(active_set_iteration);
                LinearSolver::getTimers()["update"].stop();
            }
            if (hasConverged)
                break;
        }

        //std::cout << "discrete_problem->algebraic->map->lm_to_dof " << discrete_problem->algebraic->map->lm_to_dof.size() << std::endl;

        LinearSolver::getTimers()["write results"].start();
        //if (discrete_problem->domain->contacts_list->getContactNumber() >= 1)
        //    this->writeResults<scalar>(xPost, numNodes, name_x,active_set_iteration,true,write_txt_files,discrete_problem->algebraic->map->lm_to_dof,discrete_problem->algebraic->map);
        //else
        if (write_txt_files)
            Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("x_mm.txt"), xPostMM);
        std::cout << " before writing vtk" << std::endl;
        this->writeResults<scalar>(xPost, numNodes, name_x, active_set_iteration, true, false, discrete_problem->algebraic->map, discrete_problem->domain->random_field);
        LinearSolver::getTimers()["write results"].stop();
        // ------------------------------------------
        //
        //        End of the timers and print
        //
        // ------------------------------------------
        LinearSolver::getTimers()["solver"].stop();
        LinearSolver::getTimers()["total"].stop();

        discrete_problem->domain->contacts_list = Teuchos::null;

        //tms = LinearSolver::getTimers();

        double time_mat_vec_sec = time_mat_vec->totalElapsedTime();
        double time_prec_setup_sec = time_prec_setup->totalElapsedTime();
        double time_prec_apply_sec = time_prec_apply->totalElapsedTime();
        double time_ortho_sec = time_ortho->totalElapsedTime();

        double time_belos_total_sec = time_belos_total->totalElapsedTime();

        double time_ortho_inner_sec = time_ortho_inner->totalElapsedTime();
        double time_ortho_norm_sec = time_ortho_norm->totalElapsedTime();
        double time_ortho_update_sec = time_ortho_update->totalElapsedTime();

        static clock_t freq = CLOCKS_PER_SEC;

        UnixClockTime zero_time = UnixClockTime(0);
        UnixClockTime time_mat_vec_time = UnixClockTime(time_mat_vec_sec * static_cast<double>(freq) / 10000);
        UnixClockTime time_prec_setup_time = UnixClockTime(time_prec_setup_sec * static_cast<double>(freq) / 10000);
        UnixClockTime time_prec_apply_time = UnixClockTime(time_prec_apply_sec * static_cast<double>(freq) / 10000);
        UnixClockTime time_ortho_time = UnixClockTime(time_ortho_sec * static_cast<double>(freq) / 10000);
        UnixClockTime time_ortho_inner_time = UnixClockTime(time_ortho_inner_sec * static_cast<double>(freq) / 10000);
        UnixClockTime time_ortho_norm_time = UnixClockTime(time_ortho_norm_sec * static_cast<double>(freq) / 10000);
        UnixClockTime time_ortho_update_time = UnixClockTime(time_ortho_update_sec * static_cast<double>(freq) / 10000);

        UnixClockTime time_belos_total_time = UnixClockTime(time_belos_total_sec * static_cast<double>(freq) / 10000);

        Times time_mat_vec_times, time_prec_setup_times, time_prec_apply_times, time_ortho_times, time_ortho_inner_times, time_ortho_norm_times, time_ortho_update_times;
        Times zero_times;

        Times time_belos_total_times;

        time_mat_vec_times.setUser(zero_time);
        time_mat_vec_times.setReal(time_mat_vec_time);
        time_mat_vec_times.setKernel(zero_time);

        time_prec_setup_times.setUser(zero_time);
        time_prec_setup_times.setReal(time_prec_setup_time);
        time_prec_setup_times.setKernel(zero_time);

        time_prec_apply_times.setUser(zero_time);
        time_prec_apply_times.setReal(time_prec_apply_time);
        time_prec_apply_times.setKernel(zero_time);

        time_ortho_times.setUser(zero_time);
        time_ortho_times.setReal(time_ortho_time);
        time_ortho_times.setKernel(zero_time);

        time_ortho_inner_times.setUser(zero_time);
        time_ortho_inner_times.setReal(time_ortho_inner_time);
        time_ortho_inner_times.setKernel(zero_time);

        time_ortho_norm_times.setUser(zero_time);
        time_ortho_norm_times.setReal(time_ortho_norm_time);
        time_ortho_norm_times.setKernel(zero_time);

        time_ortho_update_times.setUser(zero_time);
        time_ortho_update_times.setReal(time_ortho_update_time);
        time_ortho_update_times.setKernel(zero_time);

        zero_times.setUser(zero_time);
        zero_times.setReal(zero_time);
        zero_times.setKernel(zero_time);

        time_belos_total_times.setUser(zero_time);
        time_belos_total_times.setReal(time_belos_total_time);
        time_belos_total_times.setKernel(zero_time);

        LinearSolver::getTimers()["time_belos_total"].setTStart(zero_times);
        LinearSolver::getTimers()["time_belos_total"].setTElaps(time_belos_total_times);

        LinearSolver::getTimers()["time_mat_vec"].setTStart(zero_times);
        LinearSolver::getTimers()["time_mat_vec"].setTElaps(time_mat_vec_times);

        LinearSolver::getTimers()["time_prec_setup"].setTStart(zero_times);
        LinearSolver::getTimers()["time_prec_setup"].setTElaps(time_prec_setup_times);

        LinearSolver::getTimers()["time_prec_apply"].setTStart(zero_times);
        LinearSolver::getTimers()["time_prec_apply"].setTElaps(time_prec_apply_times);

        LinearSolver::getTimers()["time_ortho"].setTStart(zero_times);
        LinearSolver::getTimers()["time_ortho"].setTElaps(time_ortho_times);

        LinearSolver::getTimers()["time_ortho_inner"].setTStart(zero_times);
        LinearSolver::getTimers()["time_ortho_inner"].setTElaps(time_ortho_inner_times);

        LinearSolver::getTimers()["time_ortho_norm"].setTStart(zero_times);
        LinearSolver::getTimers()["time_ortho_norm"].setTElaps(time_ortho_norm_times);

        LinearSolver::getTimers()["time_ortho_update"].setTStart(zero_times);
        LinearSolver::getTimers()["time_ortho_update"].setTElaps(time_ortho_update_times);

        //tms = LinearSolver::getTimers();
        out << "\n---CPU statistics of proc " << myRank << "---\n";
        out << LinearSolver::getTimers();
        out << "---------------------------\n";

        bool print_teuchos_timers = solverList->get("Print Teuchos timers", false);
        if (print_teuchos_timers)
        {
            std::streambuf *psbuf, *backup;
            std::ofstream filestr;
            filestr.open("teuchos_timers.txt");

            backup = std::cout.rdbuf(); // back up cout's streambuf

            psbuf = filestr.rdbuf(); // get file's streambuf
            std::cout.rdbuf(psbuf);  // assign streambuf to cout

            Teuchos::TimeMonitor::summarize();

            std::cout.rdbuf(backup); // restore cout's original streambuf

            filestr.close();
        }

        Teuchos::TimeMonitor::zeroOutTimers();
    }
    Kokkos::finalize();
}
