#ifndef WLOADS_H
#define WLOADS_H

#include "katropton.h"
#include "Map.h"
#include "Domain.h"
#include "Dirichlet_list.h"
#include "Neumann_list.h"
#include "Sources_list.h"
#include "Elements_list.h"
#include "Preload_list.h"

namespace katropton
{
namespace discrete_problem
{
template <typename scalar>
class Loads
{
public:
    Teuchos::RCP<Dirichlet_list<scalar>> dirichlet_list;
    Teuchos::RCP<Neumann_list<scalar>> neumann_list;
    Teuchos::RCP<Sources_list<scalar>> sources_list;
    Teuchos::RCP<Preload_list<scalar>> preload_list;

    Loads();
    Loads(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<Domain<scalar>> domain);
};

template <typename scalar>
Loads<scalar>::Loads()
{
}

template <typename scalar>
Loads<scalar>::Loads(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<Domain<scalar>> domain)
{
    dirichlet_list = Teuchos::rcp(new Dirichlet_list<scalar>(pbl, map, domain->elements_list));
    neumann_list = Teuchos::rcp(new Neumann_list<scalar>(pbl, map, domain->elements_list));
    sources_list = Teuchos::rcp(new Sources_list<scalar>(pbl, map, domain->elements_list));
    preload_list = Teuchos::rcp(new Preload_list<scalar>(pbl, map, domain->elements_list));
}

}; // namespace discrete_problem
}; // namespace katropton

#endif //WLOADS_H
