#ifndef WMAP_H
#define WMAP_H

#include "katropton.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>

#include <Xpetra_Map.hpp>
#include <Xpetra_TpetraMap.hpp>
#include <Xpetra_BlockedMap.hpp>

namespace katropton
{
namespace discrete_problem
{
class Map
{
public:
    //
    // Convenience typedefs
    //
    typedef int local_ordinal_type;
    typedef int global_ordinal_type;
    typedef KokkosClassic::DefaultNode::DefaultNodeType node_type;
    typedef Tpetra::Map<local_ordinal_type, global_ordinal_type, node_type> tpetra_map_type;
    typedef Xpetra::Map<local_ordinal_type, global_ordinal_type, node_type> xpetra_map_type;
    typedef Xpetra::TpetraMap<local_ordinal_type, global_ordinal_type, node_type> xpetra_tmap_type;
    typedef Xpetra::BlockedMap<local_ordinal_type, global_ordinal_type, node_type> xpetra_bmap_type;

    Teuchos::RCP<const tpetra_map_type> mapNodesWO;
    Teuchos::RCP<const tpetra_map_type> mapNodes;
    Teuchos::RCP<const tpetra_map_type> mapElems;
    Teuchos::RCP<const tpetra_map_type> mapDofsWO;
    Teuchos::RCP<const tpetra_map_type> mapDofs;
    Teuchos::RCP<const tpetra_map_type> mapOutput;
    Teuchos::RCP<const tpetra_map_type> mapLagrangeDofs;
    Teuchos::RCP<const tpetra_map_type> fullmap;
    Teuchos::RCP<const tpetra_map_type> fullmapWO;
    Teuchos::RCP<const tpetra_map_type> fullmapOutput;

    Teuchos::RCP<const xpetra_bmap_type> blockedMap;

    std::map<int, int> lm_to_dof;        //to change
    std::map<int, int> lm_to_dof_global; //to change
    std::map<int, int> lagrange_to_dof;
    std::map<int, int> dof_to_lm;

    const global_ordinal_type indexBase = 0;
    size_t numPrimalDPN;
    size_t numDualDPN = 1;

    Map();
    Map(Problem &pbl, size_t _numPrimalDPN);
    Map &operator=(const Map &rhs);
};

}; // namespace discrete_problem
}; // namespace katropton

#endif //WMAP_H
