#ifndef WMATERIAL_H
#define WMATERIAL_H

#include "katropton.h"

#include "Map.h"
#include "wProblem.h"
#include "wTag.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"
#include "wMedium.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Kokkos_ViewFactory.hpp>

namespace katropton
{
namespace discrete_problem
{
template <typename scalar>
class Material
{
private:
    scalar E;
    scalar nu;
    scalar k;
    scalar beta;

public:
    Material();
    Material(scalar _E, scalar _nu, scalar _k, scalar _beta);

    inline scalar getE() { return E; };
    inline scalar getNu() { return nu; };
    inline scalar getK() { return k; };
    inline scalar getBeta() { return beta; };
};

template <typename scalar>
Material<scalar>::Material()
{
    E = (scalar)1.;
    nu = (scalar)0.;
    k = (scalar)0.;
    beta = (scalar)0.;
}

template <typename scalar>
Material<scalar>::Material(scalar _E, scalar _nu, scalar _k, scalar _beta) : E(_E), nu(_nu), k(_k), beta(_beta)
{
}
}; // namespace discrete_problem
}; // namespace katropton

#endif //WMATERIAL_H
