#ifndef WMATRICES_H
#define WMATRICES_H

#include "katropton.h"

#include "Map.h"
#include "Graph.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Tpetra_CrsGraph.hpp>
#include <Tpetra_CrsMatrix.hpp>
#include <Teuchos_RCP.hpp>

#include <Xpetra_BlockedMap.hpp>
#include <Xpetra_BlockedCrsMatrix.hpp>

namespace katropton
{
namespace discrete_problem
{
template <typename scalar>
class Matrices
{
    /*
    2x2 block matrix:
    ----------------

    *-----*-----*
    |  A  | B_1 |
    *-----*-----*
    | B_2 |  C  |
    *-----*-----*

    B includes the values of B_2 if the contact is fully closed.
    B_T includes the values of B_1 if the contact is fully closed.
    B_G includes the values of B with the M matrix included regardless if we have a Signorini problem or not.
    */
public:
    //
    // Convenience typedefs
    //
    typedef Map::local_ordinal_type local_ordinal_type;
    typedef Map::global_ordinal_type global_ordinal_type;
    typedef Map::node_type node_type;
    typedef Map::tpetra_map_type tpetra_map_type;
    typedef Map::xpetra_bmap_type xpetra_bmap_type;
    typedef Tpetra::CrsGraph<local_ordinal_type, global_ordinal_type, node_type> tpetra_graph_type;
    typedef Tpetra::CrsMatrix<scalar, local_ordinal_type, global_ordinal_type, node_type> tpetra_crs_type;

    typedef Xpetra::CrsMatrix<scalar, local_ordinal_type, global_ordinal_type, node_type> xpetra_crs_type;
    typedef Xpetra::CrsMatrixWrap<scalar, local_ordinal_type, global_ordinal_type, node_type> xpetra_wcrs_type;
    typedef Xpetra::BlockedCrsMatrix<scalar, local_ordinal_type, global_ordinal_type, node_type> xpetra_bcrs_type;
    typedef Xpetra::TpetraCrsMatrix<scalar, local_ordinal_type, global_ordinal_type, node_type> xpetra_tcrs_type;

    Teuchos::RCP<tpetra_crs_type> AWO;
    Teuchos::RCP<tpetra_crs_type> A;
    Teuchos::RCP<tpetra_crs_type> B;
    Teuchos::RCP<tpetra_crs_type> B_T;
    Teuchos::RCP<tpetra_crs_type> B_1;
    Teuchos::RCP<tpetra_crs_type> B_2;
    Teuchos::RCP<tpetra_crs_type> B_G;
    Teuchos::RCP<tpetra_crs_type> C;
    Teuchos::RCP<tpetra_crs_type> Cb;

    Teuchos::RCP<xpetra_crs_type> xA;
    Teuchos::RCP<xpetra_crs_type> xB;
    Teuchos::RCP<xpetra_crs_type> xB_T;
    Teuchos::RCP<xpetra_crs_type> xC;

    Teuchos::RCP<xpetra_wcrs_type> xwA;
    Teuchos::RCP<xpetra_wcrs_type> xwB;
    Teuchos::RCP<xpetra_wcrs_type> xwB_T;
    Teuchos::RCP<xpetra_wcrs_type> xwC;

    Teuchos::RCP<xpetra_bcrs_type> blockedMatrix;

    Matrices(){};

    Matrices(Teuchos::RCP<Graph> graph)
    {
        AWO = Teuchos::rcp(new tpetra_crs_type(graph->AWO));
        AWO->fillComplete();
        A = Teuchos::rcp(new tpetra_crs_type(graph->A));
        A->fillComplete();
    };
};
}; // namespace discrete_problem
}; // namespace katropton

#endif //WMATRICES_H
