#ifndef RESULTS_DEF_HPP_
#define RESULTS_DEF_HPP_

#include "Results_decl.hpp"
#include "Stress_computation.h"

/* @todo should be removed */
#ifdef WAVES_USE_TBOXVTK
#include "tboxVtk.h"
#include "wVtkExport_KIM2CLEAN.h"
#endif

template <class Scalar, class LocalOrdinal, class GlobalOrdinal>
void Write_Results(
    const katropton::Problem &pbl,
    Teuchos::RCP<Tpetra::Vector<Scalar, LocalOrdinal, GlobalOrdinal>> xPost,
    Teuchos::RCP<katropton::discrete_problem::Map> map,
    Teuchos::RCP<RandomField<Scalar, Kokkos::DefaultExecutionSpace>> random_field,
    std::string name,
    size_t ast,
    bool final_result,
    bool write_txt)
{
    /* @todo [AC] temporary fix for initializing Mem in katropton */
    for (auto e : pbl.msh->elems)
    {
        if (e->type() == tbox::ELTYPE::HEX8 || e->type() == tbox::ELTYPE::TETRA4) // 3D
            e->getVMem().update(true);
        //else // 2D
        //    e->getVMem().update(false);
    }
    /* end temporary fix*/

    // Extraction of types and ensemble format

    xPost->template sync<Kokkos::HostSpace>();
    auto x_2d = xPost->template getLocalView<Kokkos::HostSpace>();
    auto X = Kokkos::subview(x_2d, Kokkos::ALL(), 0);

    typedef EnsembleTraits<Scalar> ET;

    const size_t ensemble_size = ET::size;

    typedef decltype(X) X_View;
    typename Kokkos::FlatArrayType<X_View>::type X_flat = X;

    size_t numPrimalDPN = map->numPrimalDPN;

    // Construction of the maps

    std::map<LocalOrdinal, GlobalOrdinal> nodesWO_LO_to_GO;
    std::map<GlobalOrdinal, LocalOrdinal> nodesWO_GO_to_LO;
    std::map<GlobalOrdinal, LocalOrdinal> elements_GO_to_LO;
    std::vector<GlobalOrdinal> contactElementID;
    std::map<LocalOrdinal, GlobalOrdinal> lagrange_LO_to_nodes_GO;
    std::map<GlobalOrdinal, LocalOrdinal> nodes_GO_to_lagrange_LO;

    LocalOrdinal index = 0;

    for (LocalOrdinal i = 0; i < map->mapElems->getNodeNumElements(); ++i)
    {
        if (pbl.msh->elems[i]->type() == tbox::ELTYPE::HEX8 || pbl.msh->elems[i]->type() == tbox::ELTYPE::TETRA4)
        {
            GlobalOrdinal global_i = map->mapElems->getGlobalElement(i);
            elements_GO_to_LO[global_i] = index;
            ++index;
        }
    }

    for (LocalOrdinal i = 0; i < map->mapNodesWO->getNodeNumElements(); ++i)
    {
        GlobalOrdinal global_i = map->mapNodesWO->getGlobalElement(i);
        nodesWO_LO_to_GO[i] = global_i;
        nodesWO_GO_to_LO[global_i] = i;
    }

    LocalOrdinal numNodes = nodesWO_LO_to_GO.size();
    LocalOrdinal numMechanicalDualDOFs = map->lagrange_to_dof.size();
    LocalOrdinal numPrimalDOFs = numPrimalDPN * numNodes;

    std::map<LocalOrdinal, GlobalOrdinal> elems_LO_to_GO;
    std::map<GlobalOrdinal, LocalOrdinal> elems_GO_to_LO;

    for (GlobalOrdinal i = 0; i < pbl.msh->elems.size(); ++i)
    {
        if (pbl.msh->elems[i]->type() == tbox::ELTYPE::HEX8 || pbl.msh->elems[i]->type() == tbox::ELTYPE::TETRA4)
        {
            LocalOrdinal local_i = pbl.msh->elems[i]->no - 1;
            elems_LO_to_GO[local_i] = i;
            elems_GO_to_LO[i] = local_i;
        }
    }

    if (pbl.Contacts.size() >= 1)
    {
        std::vector<GlobalOrdinal> tmpcontactElementID;
        for (size_t i = 0; i < pbl.Contacts.size(); ++i)
            for (LocalOrdinal j = 0; j < pbl.Contacts[i]->slave_elems.size(); ++j)
                tmpcontactElementID.push_back(pbl.Contacts[i]->slave_elems[j]->no - 1);

        if (tmpcontactElementID.size() > 1)
        {
            std::sort(tmpcontactElementID.begin(), tmpcontactElementID.end());

            if (tmpcontactElementID.size() >= 1)
            {
                contactElementID.push_back(tmpcontactElementID[0]);
                if (tmpcontactElementID.size() > 1)
                    for (LocalOrdinal i = 1; i < tmpcontactElementID.size(); ++i)
                        if (tmpcontactElementID[i] != tmpcontactElementID[i - 1])
                            contactElementID.push_back(tmpcontactElementID[i]);
            }
        }
    }

    // Values in which we are interested

    std::vector<double> displacement_x(numNodes);
    std::vector<double> displacement_y(numNodes);
    std::vector<double> displacement_z(numNodes);
    std::vector<double> shear_modulus_per_node(numNodes);

    std::vector<double> sigma_vm_nodes(numNodes);
    std::vector<size_t> adjacent_elems(numNodes);
    std::vector<gmm::dense_matrix<double>> stress_at_nodes(numNodes);
    std::vector<gmm::dense_matrix<double>> strain_at_nodes(numNodes);

    std::vector<gmm::dense_matrix<double>> stress(pbl.msh->elems.size());
    std::vector<gmm::dense_matrix<double>> strain(pbl.msh->elems.size());

    std::vector<double> sigma_vm(pbl.msh->elems.size());

    std::vector<double> G_vector(pbl.msh->elems.size());

    std::vector<tbox::Pt> displacement(numNodes);

    std::vector<double> displacement_u(numNodes);
    std::vector<double> T(numNodes);
    std::vector<double> pressure(numMechanicalDualDOFs);

    std::vector<double> h_fluxes(numMechanicalDualDOFs);
    std::vector<double> t_x(numMechanicalDualDOFs);
    std::vector<double> t_y(numMechanicalDualDOFs);

    for (auto &k : stress)
        gmm::resize(k, 3, 3);

    for (auto &k : strain)
        gmm::resize(k, 3, 3);

    for (auto &k : stress_at_nodes)
        gmm::resize(k, 3, 3);

    for (auto &k : strain_at_nodes)
        gmm::resize(k, 3, 3);

    gmm::dense_matrix<double> zero_3x3(3, 3);
    gmm::clear(zero_3x3);

    // Fill the result vectors

    for (size_t ell = 0; ell < ensemble_size; ++ell)
    {

        tbox::Results *results = new tbox::Results();
        tbox::Results *results_contact = new tbox::Results();

        results->scalars_at_nodes["x"] = &displacement_x;
        results->scalars_at_nodes["y"] = &displacement_y;
        results->scalars_at_nodes["z"] = &displacement_z;
        results->scalars_at_nodes["u"] = &displacement_u;
        results->scalars_at_nodes["G"] = &shear_modulus_per_node;
        results->scalars_at_nodes["von Mises"] = &sigma_vm_nodes;
        results->tensors_at_nodes["sigma"] = &stress_at_nodes;
        results->tensors_at_nodes["epsilon"] = &strain_at_nodes;

        if (numPrimalDPN == 4)
            results->scalars_at_nodes["T"] = &T;

        results->tensors_at_elems["sigma"] = &stress;
        results->tensors_at_elems["epsilon"] = &strain;

        results->scalars_at_elems["von Mises"] = &sigma_vm;

        results->scalars_at_elems["G"] = &G_vector;

        results->vectors_at_nodes["displacement"] = &displacement;

        if (pbl.Contacts.size() > 0)
        {
            results_contact->scalars_at_nodes["p"] = &pressure;
            if (numPrimalDPN == 4)
                results_contact->scalars_at_nodes["heat fluxes"] = &h_fluxes;
            if (map->numDualDPN == 3 || map->numDualDPN == 4)
            {
                results_contact->scalars_at_nodes["tangential stress 1"] = &t_x;
                results_contact->scalars_at_nodes["tangential stress 2"] = &t_y;
            }
        }

        for (LocalOrdinal i = 0; i < numNodes; ++i)
        {
            double dx = X_flat[ensemble_size * numPrimalDPN * i + ell];
            double dy = X_flat[ensemble_size * (numPrimalDPN * i + 1) + ell];
            double dz = X_flat[ensemble_size * (numPrimalDPN * i + 2) + ell];
            displacement_x[i] = dx;
            displacement_y[i] = dy;
            displacement_z[i] = dz;
            displacement_u[i] = sqrt(dx * dx + dy * dy + dz * dz);
            gmm::copy(zero_3x3, strain_at_nodes[i]);
            gmm::copy(zero_3x3, stress_at_nodes[i]);
            shear_modulus_per_node[i] = ET::coeff(random_field->operator()(pbl.msh->nodes[nodesWO_LO_to_GO[i]]->pos.x[0], pbl.msh->nodes[nodesWO_LO_to_GO[i]]->pos.x[1], pbl.msh->nodes[nodesWO_LO_to_GO[i]]->pos.x[2]), ell);
            displacement[i] = tbox::Pt(dx, dy, dz);
            if (numPrimalDPN == 4)
                T[i] = X_flat[ensemble_size * (numPrimalDPN * i + 3) + ell];
        }
        if (pbl.Contacts.size() > 0)
        {
            for (LocalOrdinal i = 0; i < numMechanicalDualDOFs; ++i)
            {
                int i_node = map->lm_to_dof_global[i];
                lagrange_LO_to_nodes_GO[i] = i_node;
                nodes_GO_to_lagrange_LO[i_node] = i;
                pressure[i] = X_flat[ensemble_size * (i * map->numDualDPN + numPrimalDOFs) + ell];
            }
            if (map->numDualDPN == 2 || map->numDualDPN == 4)
                for (LocalOrdinal i = 0; i < numMechanicalDualDOFs; ++i)
                    h_fluxes[i] = X_flat[ensemble_size * (i * map->numDualDPN + map->numDualDPN - 1 + numPrimalDOFs) + ell];
            if (map->numDualDPN == 3 || map->numDualDPN == 4)
                for (LocalOrdinal i = 0; i < numMechanicalDualDOFs; ++i)
                {
                    t_x[i] = X_flat[ensemble_size * (i * map->numDualDPN + 1 + numPrimalDOFs) + ell];
                    t_y[i] = X_flat[ensemble_size * (i * map->numDualDPN + 2 + numPrimalDOFs) + ell];
                }
        }

        for (auto m : pbl.media)
        {
            tbox::Tag *current_tag = m->tag;

            double E = ET::coeff(m->E, ell);
            double nu = ET::coeff(m->nu, ell);

            double lambda = (nu * E) / ((1 + nu) * (1 - 2 * nu));
            double G = E / (2 * (1 + nu));

            for (LocalOrdinal i_elem = 0; i_elem < current_tag->elems.size(); ++i_elem)
            {
                if (random_field->isRandom)
                {
                    double x = 0., y = 0., z = 0.;
                    size_t element_size = current_tag->elems[i_elem]->nodes.size();
                    for (size_t j = 0; j < element_size; ++j)
                    {
                        x += current_tag->elems[i_elem]->nodes[j]->pos.x[0] / element_size;
                        y += current_tag->elems[i_elem]->nodes[j]->pos.x[1] / element_size;
                        z += current_tag->elems[i_elem]->nodes[j]->pos.x[2] / element_size;
                    }
                    G = ET::coeff(random_field->operator()(x, y, z), ell);
                }
                gmm::dense_matrix<double> H(9, 9);
                gmm::dense_matrix<double> D(3, 3);

                for (auto i = 0; i < 3; ++i)
                    for (auto j = 0; j < 3; ++j)
                        for (auto k = 0; k < 3; ++k)
                            for (auto l = 0; l < 3; ++l)
                            {
                                if (i == j && k == l)
                                    H(i * 3 + j, k * 3 + l) = lambda;
                                if (i == k && j == l)
                                    H(i * 3 + j, k * 3 + l) = H(i * 3 + j, k * 3 + l) + G;
                                if (i == l && j == k)
                                    H(i * 3 + j, k * 3 + l) = H(i * 3 + j, k * 3 + l) + G;
                            }
                if (numPrimalDPN == 4)
                {
                    double kappa = lambda + (2. / 3.) * G;
                    double beta = ET::coeff(m->beta, ell);

                    for (auto i = 0; i < 3; ++i)
                        D(i, i) = -3. * kappa * beta;
                }
                std::vector<double> x_tmp(8), y_tmp(8), z_tmp(8), T_tmp(8);
                gmm::dense_matrix<double> Strain(3, 3);
                gmm::dense_matrix<double> Sigma(3, 3);

                auto e = current_tag->elems[i_elem];
                std::vector<gmm::dense_matrix<double>> Strain_at_node(e->nodes.size());
                std::vector<gmm::dense_matrix<double>> Sigma_at_node(e->nodes.size());

                for (auto &k : Strain_at_node)
                    gmm::resize(k, 3, 3);
                for (auto &k : Sigma_at_node)
                    gmm::resize(k, 3, 3);

                if (e->type() == tbox::ELTYPE::HEX8 || e->type() == tbox::ELTYPE::TETRA4)
                {
                    for (size_t i = 0; i < e->nodes.size(); ++i)
                    {
                        int node_i = nodesWO_GO_to_LO[e->nodes[i]->row];
                        x_tmp[i] = displacement_x[node_i];
                        y_tmp[i] = displacement_y[node_i];
                        z_tmp[i] = displacement_z[node_i];
                        if (numPrimalDPN == 4)
                            T_tmp[i] = T[node_i];
                        else
                            T_tmp[i] = 0.;

                        gmm::dense_matrix<double> Strain_tmp(3, 3);
                        gmm::clear(Strain_tmp);
                        gmm::copy(Strain_tmp, Strain_at_node[i]);
                        gmm::copy(Strain_tmp, Sigma_at_node[i]);
                    }

                    strain_stress_x_y_z(e, Strain, Sigma, H, D, x_tmp, y_tmp, z_tmp, T_tmp, numPrimalDPN == 4);

                    gmm::copy(Strain, strain[elems_LO_to_GO[e->no - 1]]);
                    gmm::copy(Sigma, stress[elems_LO_to_GO[e->no - 1]]);

                    G_vector[elems_LO_to_GO[e->no - 1]] = G;

                    sigma_vm[elems_LO_to_GO[e->no - 1]] = sqrt((pow(Sigma(0, 0) - Sigma(1, 1), 2.) + pow(Sigma(1, 1) - Sigma(2, 2), 2.) + pow(Sigma(2, 2) - Sigma(0, 0), 2.) + 6 * (pow(Sigma(0, 1), 2.) + pow(Sigma(1, 2), 2.) + pow(Sigma(2, 0), 2.))) / 2);

                    strain_stress_x_y_z_at_node(e, Strain_at_node, Sigma_at_node, H, D, x_tmp, y_tmp, z_tmp, T_tmp, numPrimalDPN == 4);

                    for (size_t i = 0; i < e->nodes.size(); ++i)
                    {
                        LocalOrdinal node_i = nodesWO_GO_to_LO[e->nodes[i]->row];
                        if (ell == 0)
                            ++adjacent_elems[node_i]; // Increases the number of adjacent elements by one only for the first sample
                        gmm::add(Strain_at_node[i], strain_at_nodes[node_i]);
                        gmm::add(Sigma_at_node[i], stress_at_nodes[node_i]);
                    }
                }
            }
        }
        for (LocalOrdinal i = 0; i < numNodes; ++i)
        {
            gmm::scale(strain_at_nodes[i], 1. / adjacent_elems[i]);
            gmm::scale(stress_at_nodes[i], 1. / adjacent_elems[i]);

            gmm::dense_matrix<double> Sigma(3, 3);
            Sigma = stress_at_nodes[i];

            sigma_vm_nodes[i] = sqrt((pow(Sigma(0, 0) - Sigma(1, 1), 2.) + pow(Sigma(1, 1) - Sigma(2, 2), 2.) + pow(Sigma(2, 2) - Sigma(0, 0), 2.) + 6 * (pow(Sigma(0, 1), 2.) + pow(Sigma(1, 2), 2.) + pow(Sigma(2, 0), 2.))) / 2);
        }

        /* @todo temporary fix to output Kim's data */
        // All the post-processing related to katropton should be done here,
        // then tbox::MshExport should be used, either by passing tbox::GmshExport or tboxVtk::VtkExport

        tboxVtk::VtkExport_KIM2CLEAN vtkWriter(pbl.msh);
        vtkWriter.save_MPI(pbl.msh->name + "_ast" + std::to_string(ast) + "_s" + std::to_string(ell) + ".pvtu", *results, nodesWO_LO_to_GO, nodesWO_GO_to_LO, elements_GO_to_LO);
        if (pbl.Contacts.size() >= 1)
            vtkWriter.save_MPI(pbl.msh->name + "_ast" + std::to_string(ast) + "_s" + std::to_string(ell) + "_contact.vtu", *results_contact, contactElementID, lagrange_LO_to_nodes_GO, nodes_GO_to_lagrange_LO);
    }
}

#endif /* RESULTS_DEF_HPP_ */
