#ifndef WVECTORS_H
#define WVECTORS_H

#include "katropton.h"

#include "Map.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Xpetra_BlockedMultiVector.hpp>
#include <Xpetra_MultiVector.hpp>

namespace katropton
{
namespace discrete_problem
{
template <typename scalar>
class Vectors
{
public:
    //
    // Convenience typedefs
    //
    typedef Map::local_ordinal_type local_ordinal_type;
    typedef Map::global_ordinal_type global_ordinal_type;
    typedef Map::node_type node_type;
    typedef Map::tpetra_map_type tpetra_map_type;
    typedef Map::xpetra_bmap_type xpetra_bmap_type;

    typedef Tpetra::Vector<scalar, local_ordinal_type, global_ordinal_type, node_type> tpetra_vector_type;
    typedef Tpetra::MultiVector<scalar, local_ordinal_type, global_ordinal_type, node_type> tpetra_mvector_type;

    typedef Xpetra::BlockedMultiVector<scalar, local_ordinal_type, global_ordinal_type, node_type> xpetra_bmvector_type;
    typedef Xpetra::MultiVector<scalar, local_ordinal_type, global_ordinal_type, node_type> xpetra_mvector_type;
    typedef Xpetra::TpetraMultiVector<scalar, local_ordinal_type, global_ordinal_type, node_type> xpetra_tmvector_type;

    Teuchos::RCP<tpetra_vector_type> bWO;
    Teuchos::RCP<tpetra_vector_type> b;

    Teuchos::RCP<tpetra_vector_type> x;
    Teuchos::RCP<tpetra_vector_type> lagrange;

    Teuchos::RCP<tpetra_vector_type> initialGap;

    Teuchos::RCP<xpetra_bmvector_type> rhsBlockedMultiVector;
    Teuchos::RCP<xpetra_bmvector_type> solutionBlockedMultiVector;

    Teuchos::RCP<xpetra_mvector_type> rhsMultiVector;
    Teuchos::RCP<xpetra_mvector_type> solutionMultiVector;

    Vectors(){};
    Vectors(Teuchos::RCP<Map> map)
    {
        bWO = rcp(new tpetra_vector_type(map->mapDofsWO, true));
        b = rcp(new tpetra_vector_type(map->mapDofs, true));
        x = rcp(new tpetra_vector_type(map->mapDofs, true));
    };
};
}; // namespace discrete_problem
}; // namespace katropton

#endif //WVECTORS_H
