#include "mortar.h"
//#include "wElement.h"

#include "Epetra_MpiComm.h"
#include "Epetra_SerialComm.h"
#include "Teuchos_ParameterList.hpp"

// MOERTEL headers
#include "mrtr_manager.H"
#include "mrtr_segment_bilinearquad.H"
#include "mrtr_segment_bilineartri.H"

void compute_lagrange_msh(const Tpetra::Vector<>::local_ordinal_type numMyContacts,
                          Kokkos::View<int **, Kokkos::LayoutRight> mesh_Contacts,
                          Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes,
                          Kokkos::View<int **, Kokkos::LayoutRight> mesh_Elems,
                          const Tpetra::Vector<>::local_ordinal_type Contacts_slave_nodes,
                          const Tpetra::Vector<>::local_ordinal_type Contacts_slave_elems,
                          const Tpetra::Vector<>::local_ordinal_type Contacts_master_nodes,
                          const Tpetra::Vector<>::local_ordinal_type Contacts_master_elems,
                          Teuchos::RCP<const Tpetra::Map<>> mapNodesWO,
                          Epetra_Map map00,
                          const Tpetra::Vector<>::local_ordinal_type numNodes,
                          const Tpetra::Vector<>::local_ordinal_type numDPN,
                          std::vector<int> &lagrange_msh,
                          bool signorini,
                          MPI_Comm comm);

Teuchos::RCP<Tpetra::CrsMatrix<>> compute_M_D(const Tpetra::Vector<>::local_ordinal_type numMyContacts,
                                              Kokkos::View<int **, Kokkos::LayoutRight> mesh_Contacts,
                                              Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes,
                                              Kokkos::View<int **, Kokkos::LayoutRight> mesh_Elems,
                                              const Tpetra::Vector<>::local_ordinal_type Contacts_slave_nodes,
                                              const Tpetra::Vector<>::local_ordinal_type Contacts_slave_elems,
                                              const Tpetra::Vector<>::local_ordinal_type Contacts_master_nodes,
                                              const Tpetra::Vector<>::local_ordinal_type Contacts_master_elems,
                                              Teuchos::RCP<const Tpetra::Map<>> mapNodes,
                                              Teuchos::RCP<const Tpetra::Map<>> mapNodesWO,
                                              Teuchos::RCP<const Tpetra::Map<>> mapElems,
                                              Epetra_Map map00,
                                              const Tpetra::Vector<>::local_ordinal_type numNodes,
                                              const Tpetra::Vector<>::local_ordinal_type numDPN,
                                              Teuchos::RCP<Tpetra::Vector<Tpetra::Vector<>::global_ordinal_type, Tpetra::Vector<>::local_ordinal_type, Tpetra::Vector<>::global_ordinal_type>> &lagrange_msh0,
                                              bool signorini,
                                              MPI_Comm comm)
{
    Epetra_MpiComm Comm(comm);
    //Epetra_SerialComm Comm;
    Teuchos::RCP<const Teuchos::Comm<int>> Tcomm = Teuchos::rcp(new Teuchos::MpiComm<int>(comm));
    const size_t myRank = Tcomm->getRank();

    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    typedef Tpetra::CrsMatrix<> crs_matrix_type;

    const local_ordinal_type Contacts_slave_nodes_size = 0;
    const local_ordinal_type Contacts_slave_elems_size = 1;
    const local_ordinal_type Contacts_master_nodes_size = 2;
    const local_ordinal_type Contacts_master_elems_size = 3;
    const local_ordinal_type Contacts_slave_normal = 4;
    const local_ordinal_type Contacts_master_normal = 7;

    const local_ordinal_type elems_size = 1;
    const local_ordinal_type elems_nodes = 3;

    // ------------------------------------------------------------- //
    // create numMyContacts empty MOERTEL::Interface instances
    // ------------------------------------------------------------- //
    int printlevel = 0;
    std::vector<Teuchos::RCP<MOERTEL::Interface>> interfaces(numMyContacts);
    for (auto i = 0; i < numMyContacts; ++i)
        interfaces[i] = Teuchos::rcp(new MOERTEL::Interface(i, false, Comm, printlevel));

    // ------------------------------------------------------------- //
    // Add nodes on both sides of interface to interfaces
    // loop all nodes in the maps add them
    // to the interface with unique ids
    // ------------------------------------------------------------- //
    //if(myRank == 0)
    for (auto i = 0; i < numMyContacts; ++i)
    {
        for (auto j = 0; j < mesh_Contacts(i, Contacts_slave_nodes_size); ++j)
        {
            // get unique node id
            auto global_nodeid = mesh_Contacts(i, Contacts_slave_nodes + j);
            if (mapNodes->isNodeGlobalElement(global_nodeid))
            {
                // get node coordinates
                auto local_nodeid = mapNodesWO->getLocalElement(global_nodeid);
                double coord[3];
                for (auto k = 0; k < 3; ++k)
                    coord[k] = mesh_Nodes(local_nodeid, k);
                // create a moertel node
                MOERTEL::Node node(global_nodeid, coord, 1, &global_nodeid, false, printlevel);
                // add node to interface i on side j
                //std::cout<< node << std::endl;
                interfaces[i]->AddNode(node, 0);

                //std::cout<< node.LMDof() << std::endl;

                if (myRank == 1 || myRank == 2)
                    std::cout << "Proc " << myRank << ", interface " << i << ", node global id " << global_nodeid << " " << j << "/" << mesh_Contacts(i, Contacts_slave_nodes_size) << " (slave)" << std::endl;
            }
        }
        for (auto j = 0; j < mesh_Contacts(i, Contacts_master_nodes_size); ++j)
        {
            // get unique node id
            auto global_nodeid = mesh_Contacts(i, Contacts_master_nodes + j);
            if (mapNodes->isNodeGlobalElement(global_nodeid))
            {
                // get node coordinates
                auto local_nodeid = mapNodesWO->getLocalElement(global_nodeid);
                double coord[3];
                for (auto k = 0; k < 3; ++k)
                    coord[k] = mesh_Nodes(local_nodeid, k);
                // create a moertel node
                MOERTEL::Node node(global_nodeid, coord, 1, &global_nodeid, false, printlevel);
                // add node to interface i on side j
                interfaces[i]->AddNode(node, 1);

                //std::cout<< "Proc "<< myRank<< ", interface "<< i << ", node global id "<< global_nodeid << "/" << mesh_Contacts(i,Contacts_master_nodes_size) << " (master)"<< std::endl;
            }
        }
    }

    // ------------------------------------------------------------- //
    // add segments on both sides of the interface to the interface
    // ------------------------------------------------------------- //

    auto i_index = 0;
    //if(myRank == 0)
    for (auto i = 0; i < numMyContacts; ++i)
    {
        for (auto j = 0; j < mesh_Contacts(i, Contacts_slave_elems_size); ++j)
            if (mesh_Elems(mesh_Contacts(i, Contacts_slave_elems + j), elems_size) == 4)
            {
                i_index = mapElems->getGlobalElement(mesh_Contacts(i, Contacts_slave_elems + j));
                //std::cout<< "Proc "<< myRank<< ", interface "<< i << ", element " << i_index <<"/" << mesh_Contacts(i,Contacts_slave_elems_size) << ", node global id ";
                global_ordinal_type nodeids[4];
                for (auto k = 0; k < 4; ++k)
                {
                    nodeids[k] = mapNodesWO->getGlobalElement(mesh_Elems(mesh_Contacts(i, Contacts_slave_elems + j), elems_nodes + k));
                    //std::cout << nodeids[k] << " ";
                }
                //std::cout << "(slave)"<< std::endl;

                MOERTEL::Segment_BiLinearQuad segment(i_index, 4, nodeids, printlevel);
                interfaces[i]->AddSegment(segment, 0);
            }
            else if (mesh_Elems(mesh_Contacts(i, Contacts_slave_elems + j), elems_size) == 3)
            {
                global_ordinal_type nodeids[3];
                for (auto k = 0; k < 3; ++k)
                    nodeids[k] = mapNodesWO->getGlobalElement(mesh_Elems(mesh_Contacts(i, Contacts_slave_elems + j), elems_nodes + k));
                MOERTEL::Segment_BiLinearTri segment(i_index++, 3, nodeids, printlevel);
                interfaces[i]->AddSegment(segment, 0);
            }
        for (auto j = 0; j < mesh_Contacts(i, Contacts_master_elems_size); ++j)
            if (mesh_Elems(mesh_Contacts(i, Contacts_master_elems + j), elems_size) == 4)
            {
                i_index = mapElems->getGlobalElement(mesh_Contacts(i, Contacts_master_elems + j));
                //std::cout<< "Proc "<< myRank<< ", interface "<< i << ", element " << i_index <<"/" << mesh_Contacts(i,Contacts_master_elems_size)<< ", node global id ";
                global_ordinal_type nodeids[4];
                for (auto k = 0; k < 4; ++k)
                {
                    nodeids[k] = mapNodesWO->getGlobalElement(mesh_Elems(mesh_Contacts(i, Contacts_master_elems + j), elems_nodes + k));
                    //std::cout << nodeids[k] << " ";
                }

                //std::cout << "(master)"<< std::endl;
                MOERTEL::Segment_BiLinearQuad segment(i_index, 4, nodeids, printlevel);
                interfaces[i]->AddSegment(segment, 1);
            }
            else if (mesh_Elems(mesh_Contacts(i, Contacts_master_elems + j), elems_size) == 3)
            {
                global_ordinal_type nodeids[3];
                for (auto k = 0; k < 3; ++k)
                    nodeids[k] = mapNodesWO->getGlobalElement(mesh_Elems(mesh_Contacts(i, Contacts_master_elems + j), elems_nodes + k));
                MOERTEL::Segment_BiLinearTri segment(i_index++, 3, nodeids, printlevel);
                interfaces[i]->AddSegment(segment, 1);
            }
    }

    // ------------------------------------------------------------- //
    // choose the mortar side of the interface (0 or 1)
    //
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
        interfaces[i]->SetMortarSide(1);

    // ------------------------------------------------------------- //
    // As we do not know the mortar side yet (we decided to le the
    // package choose it), we can not set a dual trace function (mortar space)
    // as we don't know the side to set it to
    // so we just give orders for the function type
    // ------------------------------------------------------------- //
    //if(myRank == 0)
    std::cout << std::endl
              << "Proc " << myRank << ", number of interfaces " << numMyContacts << std::endl;
    for (auto i = 0; i < numMyContacts; ++i)
    {
        if (mesh_Contacts(i, Contacts_slave_elems_size) >= 1)
        {
            if (mesh_Elems(mesh_Contacts(i, Contacts_slave_elems), elems_size) == 4)
            {
                interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_BiLinearQuad, // primal trace space
                                                //MOERTEL::Function::func_DualBiLinearQuad);  // dual mortar space (recommended)
                                                MOERTEL::Function::func_BiLinearQuad); // mortar space (not recommended)
                std::cout << "Proc " << myRank << ", interface " << i << ", SetFunctionTypes func_BiLinearQuad" << std::endl;
            }
            else if (mesh_Elems(mesh_Contacts(i, Contacts_slave_elems), elems_size) == 3)
                interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_LinearTri, // primal trace space
                                                //MOERTEL::Function::func_DualLinearTri);  // dual mortar space (recommended)
                                                MOERTEL::Function::func_LinearTri); // mortar space (not recommended)
        }

        else if (mesh_Contacts(i, Contacts_master_elems_size) >= 1)
        {
            if (mesh_Elems(mesh_Contacts(i, Contacts_master_elems), elems_size) == 4)
            {
                interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_BiLinearQuad, // primal trace space
                                                //MOERTEL::Function::func_DualBiLinearQuad);  // dual mortar space (recommended)
                                                MOERTEL::Function::func_BiLinearQuad); // mortar space (not recommended)
                std::cout << "Proc " << myRank << ", interface " << i << ", SetFunctionTypes func_BiLinearQuad" << std::endl;
            }
            else if (mesh_Elems(mesh_Contacts(i, Contacts_master_elems), elems_size) == 3)
                interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_LinearTri, // primal trace space
                                                //MOERTEL::Function::func_DualLinearTri);  // dual mortar space (recommended)
                                                MOERTEL::Function::func_LinearTri); // mortar space (not recommended)
        }
    }

    //std::cout << "Proc " << myRank << " 341 " << map00.MyGID(341) << mapNodes->isNodeGlobalElement (341) <<std::endl;
    //std::cout << "Proc " << myRank << " 342 " << map00.MyGID(342) << mapNodes->isNodeGlobalElement (342) <<std::endl;
    //std::cout << "Proc " << myRank << " 54 " << map00.MyGID(54) << mapNodes->isNodeGlobalElement (54) <<std::endl;

    // ------------------------------------------------------------- //
    // complete the interfaces
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
        if (!interfaces[i]->Complete())
        {
            std::cout << "Interface " << i << " completion returned false\n";
            exit(EXIT_FAILURE);
        }

    /*
    for (auto i=0; i<numMyContacts; ++i)
    {
        MOERTEL::Node** nodes = interfaces[i]->GetNodeView();
        for (auto side=0; side<2; ++side)
        {
            int mynode = interfaces[i]->GlobalNnode(side);
            std::cout << i << " " << side << " "<< mynode << std::endl;

            for (auto j=0; j<mynode; ++j)
            {
                

                //if (interfaces[i]->NodePID(j) != -1)
                {
                    MOERTEL::Node* node = nodes[j];//interfaces[i]->GetNodeView(j);

                    int  nlmdof = node->Nlmdof();
                    //std::cout << nlmdof << std::endl;
                    if (nlmdof > 0)
                    {
                        const int* dof = node->Dof();
                        const int* lmdof = node->LMDof();
                        std::cout<< j << " " << dof[0] << " " << lmdof[0] << std::endl;  
                    }
                }
            }
        }
    }
 */

    // ------------------------------------------------------------- //
    // create an empty MOERTEL::Manager for 3D problems
    // It organizes everything from integration to solution
    // ------------------------------------------------------------- //
    MOERTEL::Manager manager(Comm, MOERTEL::Manager::manager_3D, printlevel);

    // ------------------------------------------------------------- //
    // Add the interfaces to the manager
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
        manager.AddInterface(*(interfaces[i]));

    // ------------------------------------------------------------- //
    // for mortar integration, the mortar manager needs to know about
    // the rowmap of the original (uncoupled) problem because it will
    // create coupling matrices D and M matching that rowmap
    // ------------------------------------------------------------- //
    manager.SetProblemMap(&map00);

    // ============================================================= //
    // choose integration parameters
    // ============================================================= //
    Teuchos::ParameterList &moertelparams = manager.Default_Parameters();
    // this does affect this 3D case only
    moertelparams.set("exact values at gauss points", true);
    // 1D interface possible values are 1,2,3,4,5,6,7,8,10 (2 recommended with linear shape functions)
    moertelparams.set("number gaussian points 1D", 2);
    // 2D interface possible values are 3,6,12,13,16,19,27 (12 recommended with linear functions)
    moertelparams.set("number gaussian points 2D", 12);

    // ============================================================= //
    // Here we are done with the construction phase of the interface
    // so we can integrate the mortar integrals
    // (Note we have not yet evaluated the PDE at all!)
    // ============================================================= //
    manager.Mortar_Integrate();

    const Epetra_CrsMatrix *D = manager.D();
    const Epetra_CrsMatrix *M = manager.M();

    std::map<int, int> lm_to_dof = manager.lm_to_dof();

    // =============
    const size_t numPrimalDOFs = numNodes * numDPN;
    const size_t numMechanicalDualDOFs = (D->NumGlobalCols() - numNodes);
    const size_t numGlDofs = (numDPN == 4) ? numPrimalDOFs + 2 * numMechanicalDualDOFs : numPrimalDOFs + numMechanicalDualDOFs;
    int indexBase = 0;

    // --------------

    /*
    if (myRank==0)
    {
        std::cout << myRank << "::" << lm_to_dof.size() << std::endl;
        for (auto it = lm_to_dof.begin(); it != lm_to_dof.end(); ++it)
            std::cout << "Lagrange node: " << it->first << " geometrical node: " << it->second << std::endl;
        
        std::cout << "numNodes = " << numNodes << std::endl;
        std::cout << "D->NumGlobalCols () = " << D->NumGlobalCols () << std::endl;
    }

    create lagrange_msh a Tpetra vector of size D->NumGlobalCols () - numNodes using the map deduce from it->first distributed.
    fill the distributed vector with the it->second
    export all on rank 0
    */

    const size_t numLagrangeNodesRank0 = (myRank == 0) ? D->NumGlobalCols() - numNodes : 0;

    Teuchos::RCP<const Tpetra::Map<>> mapLagrangeNodes;
    {
        std::vector<global_ordinal_type> myLagrangeNodes = {};

        size_t numMyRows = D->NumMyRows();
        for (auto i = 0; i < numMyRows; ++i)
        {
            size_t global_id = D->GRID(i);
            if (global_id < numNodes)
                continue;
            myLagrangeNodes.push_back(global_id - numNodes);
        }

        mapLagrangeNodes = rcp(new Tpetra::Map<>(D->NumGlobalCols() - numNodes, myLagrangeNodes, indexBase, Tcomm));
    }

    Teuchos::RCP<const Tpetra::Map<>> mapLagrangeNodes0(new Tpetra::Map<>(D->NumGlobalCols() - numNodes, numLagrangeNodesRank0, indexBase, Tcomm));

    Teuchos::RCP<Tpetra::Vector<global_ordinal_type, local_ordinal_type, global_ordinal_type>> distributed_lagrange_msh(new Tpetra::Vector<global_ordinal_type, local_ordinal_type, global_ordinal_type>(mapLagrangeNodes, false));
    lagrange_msh0 = rcp(new Tpetra::Vector<global_ordinal_type, local_ordinal_type, global_ordinal_type>(mapLagrangeNodes0, false));

    for (auto it = lm_to_dof.begin(); it != lm_to_dof.end(); ++it)
        distributed_lagrange_msh->replaceGlobalValue(it->first - numNodes, it->second);

    const Tpetra::Export<> exportLagrangeNodes(mapLagrangeNodes, mapLagrangeNodes0);

    lagrange_msh0->doExport(*distributed_lagrange_msh, exportLagrangeNodes, Tpetra::ADD);

    for (auto i = 0; i < numMyContacts; ++i)
    {
        MOERTEL::Node **nodes = interfaces[i]->GetNodeView();
        for (auto side = 0; side < 2; ++side)
        {
            int mynode = interfaces[i]->GlobalNnode(side);
            std::cout << i << " " << side << " " << mynode << std::endl;

            for (auto j = 0; j < mynode; ++j)
            {

                if (interfaces[i]->NodePID(j) != -1)
                {
                    MOERTEL::Node *node = nodes[j]; //interfaces[i]->GetNodeView(j);

                    const double *n_;
                    node->BuildAveragedNormal();
                    n_ = node->Normal();
                    std::cout << n_[0] << " " << n_[1] << " " << n_[2] << std::endl;
                }
            }
        }
    }

    // --------------

    Teuchos::RCP<const Tpetra::Map<>> mapDofs;
    {
        const size_t numMyNodes = mapNodes->getNodeNumElements();
        Teuchos::ArrayView<const global_ordinal_type> myNodes = mapNodes->getNodeElementList();

        std::vector<global_ordinal_type> myDofsTot = {};
        for (auto i = 0; i < numMyNodes; ++i)
            for (auto j = 0; j < numDPN; ++j)
                myDofsTot.push_back(numDPN * myNodes[i] + j);

        size_t numMyRows = D->NumMyRows();
        for (auto i = 0; i < numMyRows; ++i)
        {
            size_t global_id = D->GRID(i);
            if (global_id < numNodes)
                continue;
            size_t mech_id = (global_id - numNodes) + numPrimalDOFs;
            myDofsTot.push_back(mech_id);
            if (numDPN == 4)
                myDofsTot.push_back(mech_id + numMechanicalDualDOFs);
        }

        mapDofs = rcp(new Tpetra::Map<>(numGlDofs, myDofsTot, indexBase, Tcomm));
        /*
        std::cout << "moertel: Proc " << myRank << ", numMyNodes "<< numMyNodes
                  << ", numMyRows "<<  numMyRows
                  << ", num of my nonzeros D " << D->NumMyNonzeros ()
                  << ", num of my nonzeros M " << M->NumMyNonzeros ()
                  << ", number of B rows " << mapDofs->getNodeNumElements ()
                  << ", initial " << map00.NumMyPoints() << std::endl;
                  */
    }

    Teuchos::RCP<const Tpetra::Map<>> mapDofs0;
    {
        std::vector<global_ordinal_type> myDofsTot = {};
        for (auto i = 0; i < numNodes; ++i)
            for (auto j = 0; j < numDPN; ++j)
                myDofsTot.push_back(numDPN * i + j);

        size_t numMyRows = D->NumMyRows();
        for (auto i = 0; i < numMyRows; ++i)
        {
            size_t global_id = D->GRID(i);
            if (global_id < numNodes)
                continue;
            size_t mech_id = (global_id - numNodes) + numPrimalDOFs;
            myDofsTot.push_back(mech_id);
            if (numDPN == 4)
                myDofsTot.push_back(mech_id + numMechanicalDualDOFs);
        }

        mapDofs0 = rcp(new Tpetra::Map<>(numGlDofs, myDofsTot, indexBase, Tcomm));
    }
    /*
    if(myRank == 0)
        lagrange_msh.resize(D->NumMyRows()- numNodes);
    else
        lagrange_msh.resize(0);

    compute_lagrange_msh(numMyContacts,
                     mesh_Contacts,
                     mesh_Nodes,
                     mesh_Elems,
                     Contacts_slave_nodes,
                     Contacts_slave_elems,
                     Contacts_master_nodes,
                     Contacts_master_elems,
                     mapNodesWO,
                     map00,
                     numNodes,
                     numDPN,
                     lagrange_msh,
                     signorini,
                     comm
                 );
*/

    Teuchos::RCP<crs_matrix_type> A0(new crs_matrix_type(mapDofs0, 0));
    Teuchos::RCP<crs_matrix_type> A(new crs_matrix_type(mapDofs, 0));

    {
        int DNumMyRow = D->NumMyRows();
        int DMaxNumIndices = D->MaxNumEntries();
        double *values = new double[DMaxNumIndices];
        int DNumIndices;
        int *Indices_int = new int[DMaxNumIndices];

        const Epetra_Map DrowMap = D->RowMap();
        const Epetra_Map DcolMap = D->ColMap();

        //for (auto i = 0; i < 10; ++i)
        //    std::cout << "i: " << i << ", global map00: " << map00.GID(i) << ", global DrowMap: " <<  DrowMap.GID(i) << ", D->GRID(i): "<< D->GRID(i) << ", global DcolMap: " <<  DcolMap.GID(i) << std::endl;

        for (size_t i = 0; i < DNumMyRow; ++i)
        {
            size_t global_id = D->GRID(i);
            if (global_id < numNodes)
                continue;

            //D->ExtractGlobalRowCopy(global_id, DMaxNumIndices, DNumIndices, values,Indices_int);
            /*
            {
                std::cout << "GlobalRowCopy(" << i << "," << global_id <<  "):";
                for (auto k=0; k< DNumIndices; ++k)
                    std::cout << "(" <<  D->LCID(Indices_int[k]) << ","<<  Indices_int[k] << "," << values[k] << ") ";
                D->ExtractMyRowCopy(i, DMaxNumIndices, DNumIndices, values,Indices_int);
                std::cout << std::endl;
                std::cout << "LocalRowCopy(" << i << "," << global_id <<  "):";
                for (auto k=0; k< DNumIndices; ++k)
                    std::cout << "(" <<  Indices_int[k] << "," <<  D->GCID(Indices_int[k]) << "," << values[k] << ") ";
                std::cout << std::endl;
            }
*/

            D->ExtractGlobalRowCopy(global_id, DMaxNumIndices, DNumIndices, values, Indices_int);
            /*
            bool max_value = 0;
            local_ordinal_type max_index;
            for (auto k=0; k< DNumIndices; ++k)
                if(max_value < values[k])
                {
                    max_value = values[k];
                    max_index = Indices_int[k];
                }

            std::cout << "(" <<  global_id << ","<<  max_index << ") "  << std::endl;
*/
            size_t interface_j = 0;
            bool doIStop = false;
            for (auto k = 0; k < DNumIndices; ++k)
            {
                for (auto j = 0; j < numMyContacts; ++j)
                {
                    for (auto l = 0; l < mesh_Contacts(j, Contacts_slave_nodes_size); ++l)
                    {
                        if (mesh_Contacts(j, Contacts_slave_nodes + l) == Indices_int[k])
                        {
                            interface_j = j;
                            doIStop = true;
                        }
                        if (doIStop)
                            break;
                    }
                    if (doIStop)
                        break;
                }
                if (doIStop)
                    break;
            }

            for (auto k = 0; k < DNumIndices; ++k)
            {
                global_ordinal_type row = (global_id - numNodes) + numPrimalDOFs;

                global_ordinal_type col1 = Indices_int[k] * numDPN;
                global_ordinal_type col2 = Indices_int[k] * numDPN + 1;
                global_ordinal_type col3 = Indices_int[k] * numDPN + 2;

                double value1 = values[k] * mesh_Contacts(interface_j, Contacts_slave_normal);
                double value2 = values[k] * mesh_Contacts(interface_j, Contacts_slave_normal + 1);
                double value3 = values[k] * mesh_Contacts(interface_j, Contacts_slave_normal + 2);
                A0->insertGlobalValues(row,
                                       Teuchos::tuple<global_ordinal_type>(col1),
                                       Teuchos::tuple<double>(value1));
                A0->insertGlobalValues(row,
                                       Teuchos::tuple<global_ordinal_type>(col2),
                                       Teuchos::tuple<double>(value2));
                A0->insertGlobalValues(row,
                                       Teuchos::tuple<global_ordinal_type>(col3),
                                       Teuchos::tuple<double>(value3));

                A0->insertGlobalValues(col1,
                                       Teuchos::tuple<global_ordinal_type>(row),
                                       Teuchos::tuple<double>(value1));
                A0->insertGlobalValues(col2,
                                       Teuchos::tuple<global_ordinal_type>(row),
                                       Teuchos::tuple<double>(value2));
                A0->insertGlobalValues(col3,
                                       Teuchos::tuple<global_ordinal_type>(row),
                                       Teuchos::tuple<double>(value3));
                if (numDPN == 4)
                {
                    global_ordinal_type row2 = row + numMechanicalDualDOFs;
                    global_ordinal_type col4 = Indices_int[k] * numDPN + 3;
                    A0->insertGlobalValues(row2,
                                           Teuchos::tuple<global_ordinal_type>(col4),
                                           Teuchos::tuple<double>(values[k]));
                    A0->insertGlobalValues(col4,
                                           Teuchos::tuple<global_ordinal_type>(row2),
                                           Teuchos::tuple<double>(values[k]));
                }
            }
        }
    }
    if (signorini == false)
    {
        int MNumMyRow = M->NumMyRows();
        int MMaxNumIndices = M->MaxNumEntries();
        double *values = new double[MMaxNumIndices];
        int MNumIndices;
        int *Indices_int = new int[MMaxNumIndices];

        for (size_t i = 0; i < MNumMyRow; ++i)
        {
            size_t global_id = M->GRID(i);
            if (global_id < numNodes)
                continue;
            M->ExtractGlobalRowCopy(global_id, MMaxNumIndices, MNumIndices, values, Indices_int);

            size_t interface_j = 0;
            {
                int DMaxNumIndices = D->MaxNumEntries();
                double *values = new double[DMaxNumIndices];
                int DNumIndices;
                int *Indices_int = new int[DMaxNumIndices];

                D->ExtractGlobalRowCopy(global_id, DMaxNumIndices, DNumIndices, values, Indices_int);

                bool doIStop = false;
                for (auto k = 0; k < DNumIndices; ++k)
                {
                    for (auto j = 0; j < numMyContacts; ++j)
                    {
                        for (auto l = 0; l < mesh_Contacts(j, Contacts_slave_nodes_size); ++l)
                        {
                            if (mesh_Contacts(j, Contacts_slave_nodes + l) == Indices_int[k])
                            {
                                interface_j = j;
                                doIStop = true;
                            }
                            if (doIStop)
                                break;
                        }
                        if (doIStop)
                            break;
                    }
                    if (doIStop)
                        break;
                }
            }

            for (auto k = 0; k < MNumIndices; ++k)
            {

                global_ordinal_type row = (global_id - numNodes) + numPrimalDOFs;

                global_ordinal_type col1 = Indices_int[k] * numDPN;
                global_ordinal_type col2 = Indices_int[k] * numDPN + 1;
                global_ordinal_type col3 = Indices_int[k] * numDPN + 2;

                double value1 = values[k] * mesh_Contacts(interface_j, Contacts_master_normal);
                double value2 = values[k] * mesh_Contacts(interface_j, Contacts_master_normal + 1);
                double value3 = values[k] * mesh_Contacts(interface_j, Contacts_master_normal + 2);

                A0->insertGlobalValues(row,
                                       Teuchos::tuple<global_ordinal_type>(col1),
                                       Teuchos::tuple<double>(value1));
                A0->insertGlobalValues(row,
                                       Teuchos::tuple<global_ordinal_type>(col2),
                                       Teuchos::tuple<double>(value2));
                A0->insertGlobalValues(row,
                                       Teuchos::tuple<global_ordinal_type>(col3),
                                       Teuchos::tuple<double>(value3));

                A0->insertGlobalValues(col1,
                                       Teuchos::tuple<global_ordinal_type>(row),
                                       Teuchos::tuple<double>(value1));
                A0->insertGlobalValues(col2,
                                       Teuchos::tuple<global_ordinal_type>(row),
                                       Teuchos::tuple<double>(value2));
                A0->insertGlobalValues(col3,
                                       Teuchos::tuple<global_ordinal_type>(row),
                                       Teuchos::tuple<double>(value3));

                if (numDPN == 4)
                {
                    global_ordinal_type row2 = row + numMechanicalDualDOFs;
                    global_ordinal_type col4 = Indices_int[k] * numDPN + 3;
                    A0->insertGlobalValues(row2,
                                           Teuchos::tuple<global_ordinal_type>(col4),
                                           Teuchos::tuple<double>(values[k]));
                    A0->insertGlobalValues(col4,
                                           Teuchos::tuple<global_ordinal_type>(row2),
                                           Teuchos::tuple<double>(values[k]));
                }
            }
        }
    }
    A0->fillComplete();
    const Tpetra::Export<> exportA(mapDofs0, mapDofs);

    A->doExport(*A0, exportA, Tpetra::INSERT);

    A->fillComplete();
    return A;
}

void compute_lagrange_msh(const Tpetra::Vector<>::local_ordinal_type numMyContacts,
                          Kokkos::View<int **, Kokkos::LayoutRight> mesh_Contacts,
                          Kokkos::View<double **, Kokkos::LayoutRight> mesh_Nodes,
                          Kokkos::View<int **, Kokkos::LayoutRight> mesh_Elems,
                          const Tpetra::Vector<>::local_ordinal_type Contacts_slave_nodes,
                          const Tpetra::Vector<>::local_ordinal_type Contacts_slave_elems,
                          const Tpetra::Vector<>::local_ordinal_type Contacts_master_nodes,
                          const Tpetra::Vector<>::local_ordinal_type Contacts_master_elems,
                          Teuchos::RCP<const Tpetra::Map<>> mapNodesWO,
                          Epetra_Map map00,
                          const Tpetra::Vector<>::local_ordinal_type numNodes,
                          const Tpetra::Vector<>::local_ordinal_type numDPN,
                          std::vector<int> &lagrange_msh,
                          bool signorini,
                          MPI_Comm comm)
{
    Epetra_MpiComm Comm(comm);
    Teuchos::RCP<const Teuchos::Comm<int>> Tcomm = Teuchos::rcp(new Teuchos::MpiComm<int>(comm));
    const size_t myRank = Tcomm->getRank();
    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    const local_ordinal_type Contacts_slave_nodes_size = 0;
    const local_ordinal_type Contacts_slave_elems_size = 1;
    const local_ordinal_type Contacts_master_nodes_size = 2;
    const local_ordinal_type Contacts_master_elems_size = 3;

    const local_ordinal_type elems_size = 1;
    const local_ordinal_type elems_nodes = 3;

    // ------------------------------------------------------------- //
    // create numMyContacts empty MOERTEL::Interface instances
    // ------------------------------------------------------------- //
    int printlevel = 0;
    std::vector<Teuchos::RCP<MOERTEL::Interface>> interfaces(numMyContacts);
    for (auto i = 0; i < numMyContacts; ++i)
        interfaces[i] = Teuchos::rcp(new MOERTEL::Interface(i, false, Comm, printlevel));

    // ------------------------------------------------------------- //
    // Add nodes on both sides of interface to interfaces
    // loop all nodes in the maps add them
    // to the interface with unique ids
    // ------------------------------------------------------------- //
    if (myRank == 0)
        for (auto i = 0; i < numMyContacts; ++i)
        {
            for (auto j = 0; j < mesh_Contacts(i, Contacts_slave_nodes_size); ++j)
            {
                // get unique node id
                auto global_nodeid = mesh_Contacts(i, Contacts_slave_nodes + j);
                // get node coordinates
                auto local_nodeid = mapNodesWO->getLocalElement(global_nodeid);
                double coord[3];
                for (auto k = 0; k < 3; ++k)
                    coord[k] = mesh_Nodes(local_nodeid, k);
                // create a moertel node
                MOERTEL::Node node(global_nodeid, coord, 1, &global_nodeid, false, printlevel);
                // add node to interface i on side j
                //std::cout<< node << std::endl;
                interfaces[i]->AddNode(node, 0);
            }
            for (auto j = 0; j < mesh_Contacts(i, Contacts_master_nodes_size); ++j)
            {
                // get unique node id
                auto global_nodeid = mesh_Contacts(i, Contacts_master_nodes + j);
                // get node coordinates
                auto local_nodeid = mapNodesWO->getLocalElement(global_nodeid);
                double coord[3];
                for (auto k = 0; k < 3; ++k)
                    coord[k] = mesh_Nodes(local_nodeid, k);
                // create a moertel node
                MOERTEL::Node node(global_nodeid, coord, 1, &global_nodeid, false, printlevel);
                // add node to interface i on side j
                interfaces[i]->AddNode(node, 1);
            }
        }

    // ------------------------------------------------------------- //
    // add segments on both sides of the interface to the interface
    // ------------------------------------------------------------- //
    auto i_index = 0;
    if (myRank == 0)
        for (auto i = 0; i < numMyContacts; ++i)
        {
            for (auto j = 0; j < mesh_Contacts(i, Contacts_slave_elems_size); ++j)
                if (mesh_Elems(mesh_Contacts(i, Contacts_slave_elems + j), elems_size) == 4)
                {
                    global_ordinal_type nodeids[4];
                    for (auto k = 0; k < 4; ++k)
                        nodeids[k] = mapNodesWO->getGlobalElement(mesh_Elems(mesh_Contacts(i, Contacts_slave_elems + j), elems_nodes + k));
                    MOERTEL::Segment_BiLinearQuad segment(i_index++, 4, nodeids, printlevel);
                    interfaces[i]->AddSegment(segment, 0);
                }
                else if (mesh_Elems(mesh_Contacts(i, Contacts_slave_elems + j), elems_size) == 3)
                {
                    global_ordinal_type nodeids[3];
                    for (auto k = 0; k < 3; ++k)
                        nodeids[k] = mapNodesWO->getGlobalElement(mesh_Elems(mesh_Contacts(i, Contacts_slave_elems + j), elems_nodes + k));
                    MOERTEL::Segment_BiLinearTri segment(i_index++, 3, nodeids, printlevel);
                    interfaces[i]->AddSegment(segment, 0);
                }
            for (auto j = 0; j < mesh_Contacts(i, Contacts_master_elems_size); ++j)
                if (mesh_Elems(mesh_Contacts(i, Contacts_master_elems + j), elems_size) == 4)
                {
                    global_ordinal_type nodeids[4];
                    for (auto k = 0; k < 4; ++k)
                        nodeids[k] = mapNodesWO->getGlobalElement(mesh_Elems(mesh_Contacts(i, Contacts_master_elems + j), elems_nodes + k));
                    MOERTEL::Segment_BiLinearQuad segment(i_index++, 4, nodeids, printlevel);
                    interfaces[i]->AddSegment(segment, 1);
                }
                else if (mesh_Elems(mesh_Contacts(i, Contacts_master_elems + j), elems_size) == 3)
                {
                    global_ordinal_type nodeids[3];
                    for (auto k = 0; k < 3; ++k)
                        nodeids[k] = mapNodesWO->getGlobalElement(mesh_Elems(mesh_Contacts(i, Contacts_master_elems + j), elems_nodes + k));
                    MOERTEL::Segment_BiLinearTri segment(i_index++, 3, nodeids, printlevel);
                    interfaces[i]->AddSegment(segment, 1);
                }
        }

    // ------------------------------------------------------------- //
    // choose the mortar side of the interface (0 or 1)
    //
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
        interfaces[i]->SetMortarSide(1);

    // ------------------------------------------------------------- //
    // As we do not know the mortar side yet (we decided to le the
    // package choose it), we can not set a dual trace function (mortar space)
    // as we don't know the side to set it to
    // so we just give orders for the function type
    // ------------------------------------------------------------- //
    if (myRank == 0)
        for (auto i = 0; i < numMyContacts; ++i)
        {
            if (mesh_Elems(mesh_Contacts(i, Contacts_slave_elems), elems_size) == 4)
                interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_BiLinearQuad,      // primal trace space
                                                MOERTEL::Function::func_DualBiLinearQuad); // dual mortar space (recommended)
                                                                                           //MOERTEL::Function::func_BiLinearQuad);    // mortar space (not recommended)
            else if (mesh_Elems(mesh_Contacts(i, Contacts_slave_elems), elems_size) == 3)
                interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_LinearTri,      // primal trace space
                                                MOERTEL::Function::func_DualLinearTri); // dual mortar space (recommended)
                                                                                        //MOERTEL::Function::func_LinearTri);    // mortar space (not recommended)
        }

    // ------------------------------------------------------------- //
    // complete the interfaces
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
        if (!interfaces[i]->Complete())
        {
            std::cout << "Interface " << i << " completion returned false\n";
            exit(EXIT_FAILURE);
        }
    // ------------------------------------------------------------- //
    // create an empty MOERTEL::Manager for 3D problems
    // It organizes everything from integration to solution
    // ------------------------------------------------------------- //
    MOERTEL::Manager manager(Comm, MOERTEL::Manager::manager_3D, printlevel);

    // ------------------------------------------------------------- //
    // Add the interfaces to the manager
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
        manager.AddInterface(*(interfaces[i]));

    // ------------------------------------------------------------- //
    // for mortar integration, the mortar manager needs to know about
    // the rowmap of the original (uncoupled) problem because it will
    // create coupling matrices D and M matching that rowmap
    // ------------------------------------------------------------- //
    manager.SetProblemMap(&map00);

    // ============================================================= //
    // choose integration parameters
    // ============================================================= //
    Teuchos::ParameterList &moertelparams = manager.Default_Parameters();
    // this does affect this 3D case only
    moertelparams.set("exact values at gauss points", true);
    // 1D interface possible values are 1,2,3,4,5,6,7,8,10 (2 recommended with linear shape functions)
    moertelparams.set("number gaussian points 1D", 2);
    // 2D interface possible values are 3,6,12,13,16,19,27 (12 recommended with linear functions)
    moertelparams.set("number gaussian points 2D", 12);

    // ============================================================= //
    // Here we are done with the construction phase of the interface
    // so we can integrate the mortar integrals
    // (Note we have not yet evaluated the PDE at all!)
    // ============================================================= //
    manager.Mortar_Integrate();

    const Epetra_CrsMatrix *D = manager.D();

    if (myRank == 0)
        lagrange_msh.resize(D->NumMyRows() - numNodes);

    int DNumMyRow = D->NumMyRows();
    int DMaxNumIndices = D->MaxNumEntries();
    double *values = new double[DMaxNumIndices];
    int DNumIndices;
    int *Indices_int = new int[DMaxNumIndices];

    const Epetra_Map DrowMap = D->RowMap();
    const Epetra_Map DcolMap = D->ColMap();

    if (myRank == 0)
        for (size_t i = 0; i < DNumMyRow; ++i)
        {
            D->ExtractGlobalRowCopy(i, DMaxNumIndices, DNumIndices, values, Indices_int);

            for (auto k = 0; k < DNumIndices; ++k)
                lagrange_msh[DrowMap.LID((int)i) - numNodes] = Indices_int[k];
        }
}
