#include "mortar2.h"
//#include "wElement.h"

#include <MatrixMarket_Tpetra.hpp>

#include "Epetra_MpiComm.h"
#include "Epetra_SerialComm.h"
#include "Teuchos_ParameterList.hpp"

// MOERTEL headers
#include "mrtr_manager.H"
#include "mrtr_segment_bilinearquad.H"
#include "mrtr_segment_bilineartri.H"

Teuchos::RCP<Tpetra::CrsMatrix<>> compute_B(Teuchos::RCP<katropton::discrete_problem::Elements_list> elements_list,
                                            Teuchos::RCP<katropton::discrete_problem::Nodes_list> nodes_list,
                                            Teuchos::RCP<katropton::discrete_problem::Contacts_list> contacts_list,
                                            Teuchos::RCP<katropton::discrete_problem::Map> map, MPI_Comm comm, bool SignoriniCheck)
{

    //std::cout << " compute_B double : a" << std::endl;
    Epetra_MpiComm Comm(comm);
    //Epetra_SerialComm Comm;
    Teuchos::RCP<const Teuchos::Comm<int>> Tcomm = Teuchos::rcp(new Teuchos::MpiComm<int>(comm));
    //const size_t myRank = Tcomm->getRank();
    //const size_t numRank = Tcomm->getSize();

    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    typedef Tpetra::CrsMatrix<> crs_matrix_type;

    const local_ordinal_type numMyContacts = contacts_list->getContactNumber();

    //std::cout << " compute_B double : b" << std::endl;
    // ------------------------------------------------------------- //
    // create numMyContacts empty MOERTEL::Interface instances
    // ------------------------------------------------------------- //
    int printlevel = 0;
    std::vector<Teuchos::RCP<MOERTEL::Interface>> interfaces(numMyContacts);
    for (auto i = 0; i < numMyContacts; ++i)
        interfaces[i] = Teuchos::rcp(new MOERTEL::Interface(i, false, Comm, printlevel));

    //std::cout << " compute_B double : c" << std::endl;
    // ------------------------------------------------------------- //
    // Add nodes on both sides of interface to interfaces
    // loop all nodes in the maps add them
    // to the interface with unique ids
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
    {
        const local_ordinal_type numMySlaveNodes = contacts_list->getSlaveNodesSize(i);

        for (auto j = 0; j < numMySlaveNodes; ++j)
        {
            // get unique node id
            auto global_nodeid = contacts_list->getSlaveNode(i, j);

            if (map->mapNodes->isNodeGlobalElement(global_nodeid))
            {
                // get node coordinates
                auto local_nodeid = map->mapNodesWO->getLocalElement(global_nodeid);
                double coord[3];
                for (auto k = 0; k < 3; ++k)
                    coord[k] = nodes_list->nodes(local_nodeid, k);
                //std::cout<< "slave node: " << global_nodeid << " "<< coord[0]  << " " << coord[1] << " " << coord[2] << std::endl;

                //std::cout << "interface " << i <<" rank " << myRank << " slave node: " << global_nodeid << " "<< coord[0]  << " " << coord[1] << " " << coord[2] << std::endl;

                // create a moertel node
                MOERTEL::Node node(global_nodeid, coord, 1, &global_nodeid, contacts_list->isASlaveDirichletNode(i, global_nodeid), printlevel);
                // add node to interface i on side j
                interfaces[i]->AddNode(node, 0);
            }
        }

        const local_ordinal_type numMyMasterNodes = contacts_list->getMasterNodesSize(i);

        //std::cout << numMyMasterNodes << std::endl;

        for (auto j = 0; j < numMyMasterNodes; ++j)
        {
            // get unique node id
            auto global_nodeid = contacts_list->getMasterNode(i, j);

            //std::cout << global_nodeid << std::endl;
            //std::cout <<map->mapNodes->isNodeGlobalElement (global_nodeid)<< std::endl;

            if (map->mapNodes->isNodeGlobalElement(global_nodeid))
            {
                // get node coordinates
                auto local_nodeid = map->mapNodesWO->getLocalElement(global_nodeid);
                double coord[3];
                for (auto k = 0; k < 3; ++k)
                    coord[k] = nodes_list->nodes(local_nodeid, k);

                //std::cout << "interface " << i <<" rank " << myRank << " master node: " << global_nodeid << " "<< coord[0]  << " " << coord[1] << " " << coord[2] << std::endl;

                //std::cout<< "master node: " << global_nodeid << " "<< coord[0]  << " " << coord[1] << " " << coord[2] << std::endl;
                // create a moertel node
                MOERTEL::Node node(global_nodeid, coord, 1, &global_nodeid, false, printlevel);
                // add node to interface i on side j
                interfaces[i]->AddNode(node, 1);
            }
        }
    }

    //std::cout << " compute_B double : d" << std::endl;
    //std::cout<< "nodes added"<<std::endl;
    // ------------------------------------------------------------- //
    // add segments on both sides of the interface to the interface
    // ------------------------------------------------------------- //

    auto i_index = 0;

    for (auto i = 0; i < numMyContacts; ++i)
    {
        const local_ordinal_type numMySlaveElems = contacts_list->getSlaveElementsSize(i);

        for (auto j = 0; j < numMySlaveElems; ++j)
        {
            auto elem_local_index = contacts_list->getSlaveElement(i, j);

            if (elements_list->Get_Element_Size(elem_local_index) == 4)
            {
                i_index = map->mapElems->getGlobalElement(elem_local_index);

                global_ordinal_type nodeids[4];
                for (auto k = 0; k < 4; ++k)
                    nodeids[k] = map->mapNodesWO->getGlobalElement(elements_list->Get_Element_Node(elem_local_index, k));

                MOERTEL::Segment_BiLinearQuad segment(i_index, 4, nodeids, printlevel);
                interfaces[i]->AddSegment(segment, 0);

                //std::cout << "rank " << myRank << " slave elem: " << i_index << ", nodes:" << nodeids[0] << " " << nodeids[1] << " "<< nodeids[2] << " "<< nodeids[3]<<  std::endl;
            }
            else if (elements_list->Get_Element_Size(elem_local_index) == 3)
            {
                i_index = map->mapElems->getGlobalElement(elem_local_index);

                global_ordinal_type nodeids[3];
                for (auto k = 0; k < 3; ++k)
                    nodeids[k] = map->mapNodesWO->getGlobalElement(elements_list->Get_Element_Node(elem_local_index, k));
                MOERTEL::Segment_BiLinearTri segment(i_index++, 3, nodeids, printlevel);
                interfaces[i]->AddSegment(segment, 0);
            }
        }
        //std::cout<< "slave elems added"<<std::endl;
        const local_ordinal_type numMyMasterElems = contacts_list->getMasterElementsSize(i);

        for (auto j = 0; j < numMyMasterElems; ++j)
        {
            auto elem_local_index = contacts_list->getMasterElement(i, j);

            if (elements_list->Get_Element_Size(elem_local_index) == 4)
            {
                i_index = map->mapElems->getGlobalElement(elem_local_index);

                global_ordinal_type nodeids[4];
                for (auto k = 0; k < 4; ++k)
                    nodeids[k] = map->mapNodesWO->getGlobalElement(elements_list->Get_Element_Node(elem_local_index, k));

                MOERTEL::Segment_BiLinearQuad segment(i_index, 4, nodeids, printlevel);
                interfaces[i]->AddSegment(segment, 1);

                //std::cout << "rank " << myRank << " master elem: " << i_index << ", nodes:" << nodeids[0] << " " << nodeids[1] << " "<< nodeids[2] << " "<< nodeids[3]<<  std::endl;

                //std::cout<< "master elem: " << i_index << ", nodes:" << nodeids[0] << " " << elements_list->Get_Element_Node(elem_local_index,0) << " "
                //<< nodeids[1] << " " << elements_list->Get_Element_Node(elem_local_index,1) << " "
                //<< nodeids[2] << " " << elements_list->Get_Element_Node(elem_local_index,2) << " "
                //<< nodeids[3] << " " << elements_list->Get_Element_Node(elem_local_index,3) <<  std::endl;
            }
            else if (elements_list->Get_Element_Size(elem_local_index) == 3)
            {
                i_index = map->mapElems->getGlobalElement(elem_local_index);

                global_ordinal_type nodeids[3];
                for (auto k = 0; k < 3; ++k)
                    nodeids[k] = map->mapNodesWO->getGlobalElement(elements_list->Get_Element_Node(elem_local_index, k));
                MOERTEL::Segment_BiLinearTri segment(i_index++, 3, nodeids, printlevel);
                interfaces[i]->AddSegment(segment, 1);
            }
        }
        //std::cout<< "master elems added"<<std::endl;
    }

    //std::cout << " compute_B double : e" << std::endl;
    //std::cout<< "elems added"<<std::endl;

    // ------------------------------------------------------------- //
    // choose the mortar side of the interface (0 or 1)
    //
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
        interfaces[i]->SetMortarSide(1);

    // ------------------------------------------------------------- //
    // As we do not know the mortar side yet (we decided to le the
    // package choose it), we can not set a dual trace function (mortar space)
    // as we don't know the side to set it to
    // so we just give orders for the function type
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
    {
        if (contacts_list->getSlaveElementsSize(i) >= 1)
        {
            if (elements_list->Get_Element_Size(contacts_list->getSlaveElement(i, 0)) == 4)
            {
                if (contacts_list->isStandardShapeFunction(i))
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_BiLinearQuad,
                                                    MOERTEL::Function::func_BiLinearQuad);
                }
                else
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_BiLinearQuad,
                                                    MOERTEL::Function::func_DualBiLinearQuad);
                }
            }
            else if (elements_list->Get_Element_Size(contacts_list->getSlaveElement(i, 0)) == 3)
            {
                if (contacts_list->isStandardShapeFunction(i))
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_LinearTri,
                                                    MOERTEL::Function::func_LinearTri);
                }
                else
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_LinearTri,
                                                    MOERTEL::Function::func_DualLinearTri);
                }
            }
        }

        else if (contacts_list->getMasterElementsSize(i) >= 1)
        {
            if (elements_list->Get_Element_Size(contacts_list->getMasterElement(i, 0)) == 4)
            {
                if (contacts_list->isStandardShapeFunction(i))
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_BiLinearQuad,
                                                    MOERTEL::Function::func_BiLinearQuad);
                }
                else
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_BiLinearQuad,
                                                    MOERTEL::Function::func_DualBiLinearQuad);
                }
            }
            else if (elements_list->Get_Element_Size(contacts_list->getMasterElement(i, 0)) == 3)
            {
                if (contacts_list->isStandardShapeFunction(i))
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_LinearTri,
                                                    MOERTEL::Function::func_LinearTri);
                }
                else
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_LinearTri,
                                                    MOERTEL::Function::func_DualLinearTri);
                }
            }
        }
    }
    //std::cout << " compute_B double : f" << std::endl;
    //std::cout<< "functions setted"<<std::endl;
    // ------------------------------------------------------------- //
    // complete the interfaces
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
        if (!interfaces[i]->Complete())
        {
            //std::cout << "Interface " << i << " completion returned false\n";
            exit(EXIT_FAILURE);
        }
    //std::cout<< "interface completed"<<std::endl;
    //std::cout << " compute_B double : g" << std::endl;
    // ------------------------------------------------------------- //
    // create an empty MOERTEL::Manager for 3D problems
    // It organizes everything from integration to solution
    // ------------------------------------------------------------- //
    MOERTEL::Manager manager(Comm, MOERTEL::Manager::manager_3D, printlevel);

    // ------------------------------------------------------------- //
    // Add the interfaces to the manager
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
        manager.AddInterface(*(interfaces[i]));

    // ------------------------------------------------------------- //
    // for mortar integration, the mortar manager needs to know about
    // the rowmap of the original (uncoupled) problem because it will
    // create coupling matrices D and M matching that rowmap
    // ------------------------------------------------------------- //
    Teuchos::ArrayView<const global_ordinal_type> myNodes = map->mapNodes->getNodeElementList();

    int numMyNodes = map->mapNodes->getNodeNumElements();
    int numNodes = map->mapNodes->getGlobalNumElements();

    int myNodesArray[numMyNodes];

    for (auto i = 0; i < numMyNodes; ++i)
        myNodesArray[i] = (int)myNodes[i];

    Epetra_Map map00((int)numNodes, (int)numMyNodes, myNodesArray, (int)map->mapNodes->getIndexBase(), Comm);
    manager.SetProblemMap(&map00);

    //std::cout << " compute_B double : h" << std::endl;
    // ============================================================= //
    // choose integration parameters
    // ============================================================= //
    Teuchos::ParameterList &moertelparams = manager.Default_Parameters();
    // this does affect this 3D case only
    moertelparams.set("exact values at gauss points", true);
    // 1D interface possible values are 1,2,3,4,5,6,7,8,10 (2 recommended with linear shape functions)
    moertelparams.set("number gaussian points 1D", 2);
    // 2D interface possible values are 3,6,12,13,16,19,27 (12 recommended with linear functions)
    moertelparams.set("number gaussian points 2D", 12);

    // ============================================================= //
    // Here we are done with the construction phase of the interface
    // so we can integrate the mortar integrals
    // (Note we have not yet evaluated the PDE at all!)
    // ============================================================= //
    std::cout << "start the integration" << std::endl;

    bool do_integrate = true;
    int n_integration = 0;
    int n_integration_max = 10;

    while (do_integrate)
    {
        try
        {
            do_integrate = false;
            ++n_integration;
            manager.Mortar_Integrate();
        }
        catch (...)
        {
            do_integrate = true;
        }

        if (n_integration >= n_integration_max)
            break;
    }

    if (n_integration != 1)
        std::cout << "Has integrate " << n_integration << " times" << std::endl;

    //catch(...){std::cout << "WARNING! manager.Mortar_Integrate() throws an error"<<std::endl;}
    std::cout << "end the integration" << std::endl;
    const Epetra_CrsMatrix *D = manager.D();
    const Epetra_CrsMatrix *M = manager.M();

    //std::cout << " compute_B double : i" << std::endl;

    //std::cout << *D << std::endl;

    //std::cout << *M << std::endl;
    std::map<int, int> lm_to_dof = manager.lm_to_dof();

    const size_t numDPN = map->numPrimalDPN;
    const size_t numMechanicalLagrangeMultipliers = (D->NumGlobalCols() - numNodes);
    //const size_t numMechanicalDualDOFs = (D->NumGlobalCols ()-numNodes);
    //const size_t numLMPN = (numDPN==4) ? 2: 1;

    size_t numLMPN;
    if (contacts_list->hasAtLeastOneSticking() && numDPN == 4)
        numLMPN = 4;
    else if (contacts_list->hasAtLeastOneSticking())
        numLMPN = 3;
    else if (numDPN == 4)
        numLMPN = 2;
    else
        numLMPN = 1;

    map->numDualDPN = numLMPN;

    const size_t numTotalLagrangeMultipliers = numLMPN * numMechanicalLagrangeMultipliers;

    int indexBase = 0;

    //const size_t numLagrangeNodesRank0 = (myRank == 0) ? D->NumGlobalCols ()-numNodes : 0;

    // ---------------

    {
        Teuchos::ArrayView<const global_ordinal_type> myNodes = map->mapNodes->getNodeElementList();

        std::vector<global_ordinal_type> myLagrangeDofs = {};

        size_t numMyRows = D->NumMyRows();
        for (auto i = 0; i < numMyRows; ++i)
        {
            size_t global_id = D->GRID(i);
            if (global_id < numNodes)
                continue;
            size_t lagrange_id = global_id - numNodes;
            for (auto j = 0; j < numLMPN; ++j)
                myLagrangeDofs.push_back(lagrange_id * numLMPN + j);
        }

        map->mapLagrangeDofs = rcp(new Tpetra::Map<>(numTotalLagrangeMultipliers, myLagrangeDofs, indexBase, Tcomm));
    }

    //std::cout << " compute_B double : j" << std::endl;
    //std::cout << *(map->mapLagrangeDofs) << std::endl;

    // ---------------
    for (auto i = lm_to_dof.begin(); i != lm_to_dof.end(); ++i)
        map->lm_to_dof_global[i->first - numNodes] = i->second;

    for (auto i = lm_to_dof.begin(); i != lm_to_dof.end(); ++i)
        if (map->mapNodes->isNodeGlobalElement(i->second))
        {
            map->lm_to_dof[i->first - numNodes] = i->second;
            //std::cout << i->first-numNodes << " : "<<i->second <<  std::endl;
        }

    for (auto i = map->lm_to_dof.begin(); i != map->lm_to_dof.end(); ++i)
        map->lagrange_to_dof[i->first - numNodes] = (i->second) * map->numPrimalDPN;

    // As lm_to_dof is one to one, we can create dof_to_lm as follows
    for (auto i = map->lm_to_dof.begin(); i != map->lm_to_dof.end(); ++i)
        map->dof_to_lm[i->second] = i->first;

    //std::cout << "lm_to_dof.size() = " << lm_to_dof.size() << std::endl;
    //std::cout << "map->lm_to_dof.size() = " << map->lm_to_dof.size() << std::endl;
    //std::cout << " compute_B double : k" << std::endl;
    // ---------------

    // Compute the outward normal
    typedef Tpetra::MultiVector<double, local_ordinal_type, global_ordinal_type> double_multivector_type;
    //Teuchos::RCP<double_multivector_type> areaWO = Teuchos::rcp(  new double_multivector_type(map->mapNodesWO,numMyContacts,true));
    //Teuchos::RCP<double_multivector_type> area = Teuchos::rcp(  new double_multivector_type(map->mapNodes,numMyContacts,true));

    Teuchos::RCP<double_multivector_type> normalWO = Teuchos::rcp(new double_multivector_type(map->mapNodesWO, 3 * numMyContacts, true));
    Teuchos::RCP<double_multivector_type> normalSUM = Teuchos::rcp(new double_multivector_type(map->mapNodes, 3 * numMyContacts, true));

    std::vector<global_ordinal_type> myNodes2 = {};

    for (auto i = 0; i < numNodes; ++i)
        myNodes2.push_back(i);

    Teuchos::RCP<Tpetra::Map<>> fullNodeMap = rcp(new Tpetra::Map<>(numNodes, myNodes2, indexBase, Tcomm));
    Teuchos::RCP<double_multivector_type> normal = Teuchos::rcp(new double_multivector_type(fullNodeMap, 3 * numMyContacts, true));

    //std::cout << " compute_B double : l" << std::endl;

    //std::cout << "myRank " << myRank << " fullNodeMap " << fullNodeMap->getNodeNumElements() << " "  << fullNodeMap->getGlobalNumElements() << std::endl;
    {
        for (auto i = 0; i < numMyContacts; ++i)
        {
            const local_ordinal_type numMySlaveElems = contacts_list->getSlaveElementsSize(i);

            for (auto j = 0; j < numMySlaveElems; ++j)
            {
                auto elem_local_index = contacts_list->getSlaveElement(i, j);

                if (elements_list->Get_Element_Size(elem_local_index) == 4)
                {
                    i_index = map->mapElems->getGlobalElement(elem_local_index);

                    global_ordinal_type node_id[4];
                    double coordinates[4 * 3];
                    for (auto k = 0; k < 4; ++k)
                    {
                        node_id[k] = map->mapNodesWO->getGlobalElement(elements_list->Get_Element_Node(elem_local_index, k));
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        for (auto l = 0; l < 3; ++l)
                            coordinates[k * 3 + l] = nodes_list->nodes(local_nodeid, l);
                    }

                    double vec1[3];
                    double vec2[3];
                    double vec3[3];
                    double current_normal[3];
                    double current_normal2[3];
                    double norm1;
                    double norm2;
                    double current_area;

                    vec1[0] = coordinates[3] - coordinates[0];
                    vec1[1] = coordinates[4] - coordinates[1];
                    vec1[2] = coordinates[5] - coordinates[2];

                    vec2[0] = coordinates[6] - coordinates[0];
                    vec2[1] = coordinates[7] - coordinates[1];
                    vec2[2] = coordinates[8] - coordinates[2];

                    vec3[0] = coordinates[9] - coordinates[0];
                    vec3[1] = coordinates[10] - coordinates[1];
                    vec3[2] = coordinates[11] - coordinates[2];

                    current_normal[0] = vec1[1] * vec2[2] - vec1[2] * vec2[1];
                    current_normal[1] = vec1[2] * vec2[0] - vec1[0] * vec2[2];
                    current_normal[2] = vec1[0] * vec2[1] - vec1[1] * vec2[0];

                    current_normal2[0] = vec2[1] * vec3[2] - vec2[2] * vec3[1];
                    current_normal2[1] = vec2[2] * vec3[0] - vec2[0] * vec3[2];
                    current_normal2[2] = vec2[0] * vec3[1] - vec2[1] * vec3[0];

                    norm1 = sqrt(pow(current_normal[0], 2) + pow(current_normal[1], 2) + pow(current_normal[2], 2));
                    norm2 = sqrt(pow(current_normal2[0], 2) + pow(current_normal2[1], 2) + pow(current_normal2[2], 2));

                    if (norm1 > 0.)
                    {
                        current_normal[0] = current_normal[0] / norm1;
                        current_normal[1] = current_normal[1] / norm1;
                        current_normal[2] = current_normal[2] / norm1;

                        for (auto l = 0; l < 3; ++l)
                            if (abs(current_normal[l]) <= 1e-10)
                                current_normal[l] = 0.;
                    }
                    else
                    {
                        current_normal[0] = 0.;
                        current_normal[1] = 0.;
                        current_normal[2] = 0.;
                    }

                    current_area = (norm1 + norm2) / 2.;

                    //std::cout << current_area << " " << current_normal[0] << " " << current_normal[1] << " " << current_normal[2] << std::endl;
                    //std::cout << current_normal2[0] << " " << current_normal2[1] << " " << current_normal2[2] << std::endl;

                    for (auto k = 0; k < 4; ++k)
                    {
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        //areaWO->sumIntoLocalValue(local_nodeid,i,current_area);
                        for (auto l = 0; l < 3; ++l)
                            normalWO->sumIntoLocalValue(local_nodeid, i * 3 + l, current_area * current_normal[l]);
                    }
                }
                else if (elements_list->Get_Element_Size(elem_local_index) == 3)
                {
                    i_index = map->mapElems->getGlobalElement(elem_local_index);

                    global_ordinal_type node_id[3];
                    double coordinates[3 * 3];
                    for (auto k = 0; k < 3; ++k)
                    {
                        node_id[k] = map->mapNodesWO->getGlobalElement(elements_list->Get_Element_Node(elem_local_index, k));
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        for (auto l = 0; l < 3; ++l)
                            coordinates[k * 3 + l] = nodes_list->nodes(local_nodeid, l);
                    }

                    double vec1[3];
                    double vec2[3];
                    double current_normal[3];
                    double norm;
                    double current_area;

                    vec1[0] = coordinates[3] - coordinates[0];
                    vec1[1] = coordinates[4] - coordinates[1];
                    vec1[2] = coordinates[5] - coordinates[2];

                    vec2[0] = coordinates[6] - coordinates[0];
                    vec2[1] = coordinates[7] - coordinates[1];
                    vec2[2] = coordinates[8] - coordinates[2];

                    current_normal[0] = vec1[1] * vec2[2] - vec1[2] * vec2[1];
                    current_normal[1] = vec1[2] * vec2[0] - vec1[0] * vec2[2];
                    current_normal[2] = vec1[0] * vec2[1] - vec1[1] * vec2[0];

                    norm = sqrt(pow(current_normal[0], 2) + pow(current_normal[1], 2) + pow(current_normal[2], 2));

                    current_area = norm / 2.;

                    if (norm > 0.)
                    {
                        current_normal[0] = current_normal[0] / norm;
                        current_normal[1] = current_normal[1] / norm;
                        current_normal[2] = current_normal[2] / norm;

                        for (auto l = 0; l < 3; ++l)
                            if (abs(current_normal[l]) <= 1e-10)
                                current_normal[l] = 0.;
                    }
                    else
                    {
                        current_normal[0] = 0.;
                        current_normal[1] = 0.;
                        current_normal[2] = 0.;
                    }
                    //std::cout << "Slave Element: " << i_index << ": " << node_id[0] << " " << node_id[1] << " " << node_id[2] << std::endl;
                    //std::cout << "Slave " << norm << " " << current_normal[0] << " " << current_normal[1] << " " << current_normal[2] << std::endl;

                    for (auto k = 0; k < 3; ++k)
                    {
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        //areaWO->sumIntoLocalValue(local_nodeid,i,current_area);
                        for (auto l = 0; l < 3; ++l)
                            normalWO->sumIntoLocalValue(local_nodeid, i * 3 + l, current_area * current_normal[l]);
                    }
                }
            }

            const local_ordinal_type numMyMasterElems = contacts_list->getMasterElementsSize(i);

            for (auto j = 0; j < numMyMasterElems; ++j)
            {
                auto elem_local_index = contacts_list->getMasterElement(i, j);

                if (elements_list->Get_Element_Size(elem_local_index) == 4)
                {
                    i_index = map->mapElems->getGlobalElement(elem_local_index);

                    global_ordinal_type node_id[4];
                    double coordinates[4 * 3];
                    for (auto k = 0; k < 4; ++k)
                    {
                        node_id[k] = map->mapNodesWO->getGlobalElement(elements_list->Get_Element_Node(elem_local_index, k));
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        for (auto l = 0; l < 3; ++l)
                            coordinates[k * 3 + l] = nodes_list->nodes(local_nodeid, l);
                    }

                    double vec1[3];
                    double vec2[3];
                    double vec3[3];
                    double current_normal[3];
                    double current_normal2[3];
                    double norm1;
                    double norm2;
                    double current_area;

                    vec1[0] = coordinates[3] - coordinates[0];
                    vec1[1] = coordinates[4] - coordinates[1];
                    vec1[2] = coordinates[5] - coordinates[2];

                    vec2[0] = coordinates[6] - coordinates[0];
                    vec2[1] = coordinates[7] - coordinates[1];
                    vec2[2] = coordinates[8] - coordinates[2];

                    vec3[0] = coordinates[9] - coordinates[0];
                    vec3[1] = coordinates[10] - coordinates[1];
                    vec3[2] = coordinates[11] - coordinates[2];

                    current_normal[0] = vec1[1] * vec2[2] - vec1[2] * vec2[1];
                    current_normal[1] = vec1[2] * vec2[0] - vec1[0] * vec2[2];
                    current_normal[2] = vec1[0] * vec2[1] - vec1[1] * vec2[0];

                    current_normal2[0] = vec2[1] * vec3[2] - vec2[2] * vec3[1];
                    current_normal2[1] = vec2[2] * vec3[0] - vec2[0] * vec3[2];
                    current_normal2[2] = vec2[0] * vec3[1] - vec2[1] * vec3[0];

                    norm1 = sqrt(pow(current_normal[0], 2) + pow(current_normal[1], 2) + pow(current_normal[2], 2));
                    norm2 = sqrt(pow(current_normal2[0], 2) + pow(current_normal2[1], 2) + pow(current_normal2[2], 2));

                    if (norm1 > 0.)
                    {
                        current_normal[0] = current_normal[0] / norm1;
                        current_normal[1] = current_normal[1] / norm1;
                        current_normal[2] = current_normal[2] / norm1;

                        for (auto l = 0; l < 3; ++l)
                            if (abs(current_normal[l]) <= 1e-10)
                                current_normal[l] = 0.;
                    }
                    else
                    {
                        current_normal[0] = 0.;
                        current_normal[1] = 0.;
                        current_normal[2] = 0.;
                    }

                    current_area = (norm1 + norm2) / 2.;

                    //std::cout << "Element: " << i_index << ": " << node_id[0] << " " << node_id[1] << " " << node_id[2] << " " << node_id[3] << std::endl;
                    //std::cout << current_area << " " << current_normal[0] << " " << current_normal[1] << " " << current_normal[2] << std::endl;
                    //std::cout << current_normal2[0] << " " << current_normal2[1] << " " << current_normal2[2] << std::endl;

                    for (auto k = 0; k < 4; ++k)
                    {
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        //areaWO->sumIntoLocalValue(local_nodeid,i,current_area);
                        for (auto l = 0; l < 3; ++l)
                            normalWO->sumIntoLocalValue(local_nodeid, i * 3 + l, current_area * current_normal[l]);
                    }
                }
                else if (elements_list->Get_Element_Size(elem_local_index) == 3)
                {
                    i_index = map->mapElems->getGlobalElement(elem_local_index);

                    global_ordinal_type node_id[3];
                    double coordinates[3 * 3];
                    for (auto k = 0; k < 3; ++k)
                    {
                        node_id[k] = map->mapNodesWO->getGlobalElement(elements_list->Get_Element_Node(elem_local_index, k));
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        for (auto l = 0; l < 3; ++l)
                            coordinates[k * 3 + l] = nodes_list->nodes(local_nodeid, l);
                    }

                    double vec1[3];
                    double vec2[3];
                    double current_normal[3];
                    double norm;
                    double current_area;

                    vec1[0] = coordinates[3] - coordinates[0];
                    vec1[1] = coordinates[4] - coordinates[1];
                    vec1[2] = coordinates[5] - coordinates[2];

                    vec2[0] = coordinates[6] - coordinates[0];
                    vec2[1] = coordinates[7] - coordinates[1];
                    vec2[2] = coordinates[8] - coordinates[2];

                    current_normal[0] = vec1[1] * vec2[2] - vec1[2] * vec2[1];
                    current_normal[1] = vec1[2] * vec2[0] - vec1[0] * vec2[2];
                    current_normal[2] = vec1[0] * vec2[1] - vec1[1] * vec2[0];

                    norm = sqrt(pow(current_normal[0], 2) + pow(current_normal[1], 2) + pow(current_normal[2], 2));

                    current_area = norm / 2.;

                    if (norm > 0.)
                    {
                        current_normal[0] = current_normal[0] / norm;
                        current_normal[1] = current_normal[1] / norm;
                        current_normal[2] = current_normal[2] / norm;

                        for (auto l = 0; l < 3; ++l)
                            if (abs(current_normal[l]) <= 1e-10)
                                current_normal[l] = 0.;
                    }
                    else
                    {
                        current_normal[0] = 0.;
                        current_normal[1] = 0.;
                        current_normal[2] = 0.;
                    }

                    //std::cout << "Element: " << i_index << ": " << node_id[0] << " " << node_id[1] << " " << node_id[2] << std::endl;
                    //std::cout << norm << " " << current_normal[0] << " " << current_normal[1] << " " << current_normal[2] << std::endl;

                    for (auto k = 0; k < 3; ++k)
                    {
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        //areaWO->sumIntoLocalValue(local_nodeid,i,current_area);
                        for (auto l = 0; l < 3; ++l)
                            normalWO->sumIntoLocalValue(local_nodeid, i * 3 + l, current_area * current_normal[l]);
                    }
                }
            }
        }
    }

    //std::cout << " compute_B double : m" << std::endl;

    Tpetra::Export<> export1(map->mapNodesWO, map->mapNodes);
    Tpetra::Export<> export2(map->mapNodes, normal->getMap());
    Tpetra::Import<> import1(map->mapNodes, normal->getMap());

    normalSUM->doExport(*normalWO, export1, Tpetra::ADD);
    normal->doImport(*normalSUM, import1, Tpetra::INSERT);

    //Tpetra::MatrixMarket::Writer< crs_matrix_type >::writeDenseFile(std::string("normalSUM_mm.txt"),normalSUM);
    //Tpetra::MatrixMarket::Writer< crs_matrix_type >::writeDenseFile(std::string("normal_mm.txt"),normal);

    //area->doExport(*areaWO, export1, Tpetra::ADD);

    normal->template sync<Kokkos::HostSpace>();
    auto normal_2d = normal->template getLocalView<Kokkos::HostSpace>();
    /*
    Teuchos::RCP<double_multivector_type> tangential_1, tangential_2;

    if(contacts_list->hasAtLeastOneSticking())
    {
      tangential_1 = Teuchos::rcp(  new double_multivector_type(fullNodeMap,3*numMyContacts,true));
      tangential_2 = Teuchos::rcp(  new double_multivector_type(fullNodeMap,3*numMyContacts,true));

      for(size_t i=0; i<fullNodeMap.getNodeNumElements (); ++i)
      {

      }
    }
*/
    //std::cout << "-----------------" << std::endl;
    /*
    std::ofstream outputfile;
    outputfile.open ("D.text");
    outputfile << *D << std::endl;
    outputfile.close();
*/
    //Teuchos::RCP<crs_matrix_type> B (new crs_matrix_type (map->mapLagrangeDofs,map->mapDofs,0));

    //std::cout << " compute_B double : n" << std::endl;

    size_t maxNumEntPerRow = 3 * (D->MaxNumEntries() + M->MaxNumEntries());
    Teuchos::RCP<crs_matrix_type> B(new crs_matrix_type(map->mapLagrangeDofs, maxNumEntPerRow));
    {
        int DNumMyRow = D->NumMyRows();
        int DMaxNumIndices = D->MaxNumEntries();
        double *values = new double[DMaxNumIndices];
        int DNumIndices;
        int *Indices_int = new int[DMaxNumIndices];

        const Epetra_Map DrowMap = D->RowMap();
        const Epetra_Map DcolMap = D->ColMap();

        for (size_t i = 0; i < DNumMyRow; ++i)
        {
            size_t global_id = D->GRID(i);
            if (global_id < numNodes)
                continue;

            D->ExtractGlobalRowCopy(global_id, DMaxNumIndices, DNumIndices, values, Indices_int);

            // find which interface interface_j is related to the current Lagrange multiplier
            size_t interface_j = 0; // 0 by default
            if (contacts_list->getContactNumber() > 1)
            {
                bool doIStop = false;
                for (auto k = 0; k < DNumIndices; ++k)
                {
                    for (auto j = 0; j < numMyContacts; ++j)
                    {
                        const local_ordinal_type numMySlaveNodes = contacts_list->getSlaveNodesSize(j);

                        for (auto l = 0; l < numMySlaveNodes; ++l)
                        {
                            auto global_nodeid = contacts_list->getSlaveNode(j, l);

                            if (global_nodeid == Indices_int[k])
                            {
                                interface_j = j;
                                doIStop = true;
                            }
                            if (doIStop)
                                break;
                        }
                        if (doIStop)
                            break;
                    }
                    if (doIStop)
                        break;
                }
            }

            global_ordinal_type row = (global_id - numNodes) * numLMPN;

            for (auto k = 0; k < DNumIndices; ++k)
            {
                global_ordinal_type col1 = Indices_int[k] * numDPN;
                global_ordinal_type col2 = Indices_int[k] * numDPN + 1;
                global_ordinal_type col3 = Indices_int[k] * numDPN + 2;

                global_ordinal_type global_node_id = Indices_int[k];
                local_ordinal_type local_node_id = normal->getMap()->getLocalElement(global_node_id);

                double current_normal[3];
                current_normal[0] = normal_2d(local_node_id, interface_j * 3 + 0);
                current_normal[1] = normal_2d(local_node_id, interface_j * 3 + 1);
                current_normal[2] = normal_2d(local_node_id, interface_j * 3 + 2);

                double norm = sqrt(pow(current_normal[0], 2) + pow(current_normal[1], 2) + pow(current_normal[2], 2));

                //std::cout << "D: interface_j " << interface_j << " rank " << myRank << " node " << global_node_id << " norm " << norm << " " << current_normal[0] << " " << current_normal[1] << " " << current_normal[2] << " " << values[k] << std::endl;

                if (norm > 0.)
                {
                    current_normal[0] = current_normal[0] / norm;
                    current_normal[1] = current_normal[1] / norm;
                    current_normal[2] = current_normal[2] / norm;
                }
                else
                {
                    current_normal[0] = 0.;
                    current_normal[1] = 0.;
                    current_normal[2] = 0.;
                }

                double current_tangential_1[3];
                double current_tangential_2[3];

                current_tangential_1[0] = 0.;
                current_tangential_1[1] = 0.;
                current_tangential_1[2] = 1.;

                current_tangential_2[0] = 0.;
                current_tangential_2[1] = 0.;
                current_tangential_2[2] = 0.;

                if (contacts_list->hasAtLeastOneSticking())
                {
                    double dot_n_z = current_normal[0] * current_tangential_1[0] + current_normal[1] * current_tangential_1[1] + current_normal[2] * current_tangential_1[2];

                    if (acos(abs(dot_n_z)) <= 1e-6 || contacts_list->isTying(interface_j))
                    {
                        if (contacts_list->isTying(interface_j))
                        {
                            current_normal[0] = 0.;
                            current_normal[1] = 0.;
                            current_normal[2] = 1.;
                        }
                        current_tangential_1[0] = 1.;
                        current_tangential_1[1] = 0.;
                        current_tangential_1[2] = 0.;

                        current_tangential_2[0] = 0.;
                        current_tangential_2[1] = 1.;
                        current_tangential_2[2] = 0.;
                    }
                    else
                    {
                        current_tangential_1[0] = current_tangential_1[0] - dot_n_z * current_normal[0];
                        current_tangential_1[1] = current_tangential_1[1] - dot_n_z * current_normal[1];
                        current_tangential_1[2] = current_tangential_1[2] - dot_n_z * current_normal[2];

                        double norm = sqrt(pow(current_tangential_1[0], 2) + pow(current_tangential_1[1], 2) + pow(current_tangential_1[2], 2));

                        current_tangential_1[0] = current_tangential_1[0] / norm;
                        current_tangential_1[1] = current_tangential_1[1] / norm;
                        current_tangential_1[2] = current_tangential_1[2] / norm;

                        current_tangential_2[0] = -(current_normal[1] * current_tangential_1[2] - current_normal[2] * current_tangential_1[1]);
                        current_tangential_2[1] = -(current_normal[2] * current_tangential_1[0] - current_normal[0] * current_tangential_1[2]);
                        current_tangential_2[2] = -(current_normal[0] * current_tangential_1[1] - current_normal[1] * current_tangential_1[0]);
                    }
                }

                /*
                for (auto l=0; l< 3; ++l)
                    if (current_normal[l] != contacts_list->getSlaveNormal(interface_j,l))
                        //std::cout << "ATTENTION SLAVE NODE " << global_node_id << " comp " << l << " " <<current_normal[l] << " " << contacts_list->getSlaveNormal(interface_j,l) << std::endl;
                */

                double value1 = values[k] * current_normal[0] * contacts_list->getMortarScale(interface_j); //contacts_list->getSlaveNormal(interface_j,0);
                double value2 = values[k] * current_normal[1] * contacts_list->getMortarScale(interface_j); //contacts_list->getSlaveNormal(interface_j,1);
                double value3 = values[k] * current_normal[2] * contacts_list->getMortarScale(interface_j); //contacts_list->getSlaveNormal(interface_j,2);

                //double value1 = values[k]* contacts_list->getSlaveNormal(interface_j,0);
                //double value2 = values[k]* contacts_list->getSlaveNormal(interface_j,1);
                //double value3 = values[k]* contacts_list->getSlaveNormal(interface_j,2);

                B->insertGlobalValues(row,
                                      Teuchos::tuple<global_ordinal_type>(col1),
                                      Teuchos::tuple<double>(value1));
                B->insertGlobalValues(row,
                                      Teuchos::tuple<global_ordinal_type>(col2),
                                      Teuchos::tuple<double>(value2));
                B->insertGlobalValues(row,
                                      Teuchos::tuple<global_ordinal_type>(col3),
                                      Teuchos::tuple<double>(value3));

                if (contacts_list->isSticking(interface_j))
                {
                    value1 = values[k] * current_tangential_1[0] * contacts_list->getMortarScale(interface_j);
                    value2 = values[k] * current_tangential_1[1] * contacts_list->getMortarScale(interface_j);
                    value3 = values[k] * current_tangential_1[2] * contacts_list->getMortarScale(interface_j);

                    B->insertGlobalValues(row + 1,
                                          Teuchos::tuple<global_ordinal_type>(col1),
                                          Teuchos::tuple<double>(value1));
                    B->insertGlobalValues(row + 1,
                                          Teuchos::tuple<global_ordinal_type>(col2),
                                          Teuchos::tuple<double>(value2));
                    B->insertGlobalValues(row + 1,
                                          Teuchos::tuple<global_ordinal_type>(col3),
                                          Teuchos::tuple<double>(value3));

                    value1 = values[k] * current_tangential_2[0] * contacts_list->getMortarScale(interface_j);
                    value2 = values[k] * current_tangential_2[1] * contacts_list->getMortarScale(interface_j);
                    value3 = values[k] * current_tangential_2[2] * contacts_list->getMortarScale(interface_j);

                    B->insertGlobalValues(row + 2,
                                          Teuchos::tuple<global_ordinal_type>(col1),
                                          Teuchos::tuple<double>(value1));
                    B->insertGlobalValues(row + 2,
                                          Teuchos::tuple<global_ordinal_type>(col2),
                                          Teuchos::tuple<double>(value2));
                    B->insertGlobalValues(row + 2,
                                          Teuchos::tuple<global_ordinal_type>(col3),
                                          Teuchos::tuple<double>(value3));
                }
                if (numDPN == 4)
                {
                    global_ordinal_type row2 = row + numLMPN - 1;
                    global_ordinal_type col4 = Indices_int[k] * numDPN + 3;
                    B->insertGlobalValues(row2,
                                          Teuchos::tuple<global_ordinal_type>(col4),
                                          Teuchos::tuple<double>(values[k]));
                }
            }

            if (SignoriniCheck)
                if (contacts_list->isNotSignorini(interface_j) == false)
                    continue;
            {
                int MMaxNumIndices = M->MaxNumEntries();
                double *values = new double[MMaxNumIndices];
                int MNumIndices;
                int *Indices_int = new int[MMaxNumIndices];

                M->ExtractGlobalRowCopy(global_id, MMaxNumIndices, MNumIndices, values, Indices_int);

                for (auto k = 0; k < MNumIndices; ++k)
                {
                    global_ordinal_type col1 = Indices_int[k] * numDPN;
                    global_ordinal_type col2 = Indices_int[k] * numDPN + 1;
                    global_ordinal_type col3 = Indices_int[k] * numDPN + 2;

                    global_ordinal_type global_node_id = Indices_int[k];
                    local_ordinal_type local_node_id = normal->getMap()->getLocalElement(global_node_id);

                    double current_normal[3];
                    current_normal[0] = normal_2d(local_node_id, interface_j * 3 + 0);
                    current_normal[1] = normal_2d(local_node_id, interface_j * 3 + 1);
                    current_normal[2] = normal_2d(local_node_id, interface_j * 3 + 2);

                    double norm = sqrt(pow(current_normal[0], 2) + pow(current_normal[1], 2) + pow(current_normal[2], 2));

                    //std::cout << "M: interface_j " << interface_j << " rank " << myRank << " node " << global_node_id << " norm " << norm << " " << current_normal[0] << " " << current_normal[1] << " " << current_normal[2] << " " << values[k] << std::endl;

                    //std::cout << norm << " " << current_normal[0] << " " << current_normal[1] << " " << current_normal[2] << std::endl;

                    if (norm > 0.)
                    {
                        current_normal[0] = current_normal[0] / norm;
                        current_normal[1] = current_normal[1] / norm;
                        current_normal[2] = current_normal[2] / norm;
                    }
                    else
                    {
                        current_normal[0] = 0.;
                        current_normal[1] = 0.;
                        current_normal[2] = 0.;
                    }

                    double current_tangential_1[3];
                    double current_tangential_2[3];

                    current_tangential_1[0] = 0.;
                    current_tangential_1[1] = 0.;
                    current_tangential_1[2] = 1.;

                    current_tangential_2[0] = 0.;
                    current_tangential_2[1] = 0.;
                    current_tangential_2[2] = 0.;

                    if (contacts_list->hasAtLeastOneSticking())
                    {
                        current_tangential_1[0] = 0.;
                        current_tangential_1[1] = 0.;
                        current_tangential_1[2] = 1.;

                        double dot_n_z = current_normal[0] * current_tangential_1[0] + current_normal[1] * current_tangential_1[1] + current_normal[2] * current_tangential_1[2];

                        if (acos(abs(dot_n_z)) <= 1e-6 || contacts_list->isTying(interface_j))
                        {
                            if (contacts_list->isTying(interface_j))
                            {
                                current_normal[0] = 0.;
                                current_normal[1] = 0.;
                                current_normal[2] = -1.;
                            }
                            current_tangential_1[0] = 1.;
                            current_tangential_1[1] = 0.;
                            current_tangential_1[2] = 0.;

                            current_tangential_2[0] = 0.;
                            current_tangential_2[1] = 1.;
                            current_tangential_2[2] = 0.;
                        }
                        else
                        {
                            current_tangential_1[0] = current_tangential_1[0] - dot_n_z * current_normal[0];
                            current_tangential_1[1] = current_tangential_1[1] - dot_n_z * current_normal[1];
                            current_tangential_1[2] = current_tangential_1[2] - dot_n_z * current_normal[2];

                            double norm = sqrt(pow(current_tangential_1[0], 2) + pow(current_tangential_1[1], 2) + pow(current_tangential_1[2], 2));

                            current_tangential_1[0] = current_tangential_1[0] / norm;
                            current_tangential_1[1] = current_tangential_1[1] / norm;
                            current_tangential_1[2] = current_tangential_1[2] / norm;

                            current_tangential_2[0] = current_normal[1] * current_tangential_1[2] - current_normal[2] * current_tangential_1[1];
                            current_tangential_2[1] = current_normal[2] * current_tangential_1[0] - current_normal[0] * current_tangential_1[2];
                            current_tangential_2[2] = current_normal[0] * current_tangential_1[1] - current_normal[1] * current_tangential_1[0];
                        }
                    }

                    /*
                    for (auto l=0; l< 3; ++l)
                        if (current_normal[l] != contacts_list->getMasterNormal(interface_j,l))
                            //std::cout << "ATTENTION MASTER NODE " << global_node_id << " comp " << l << " " <<current_normal[l] << " " << contacts_list->getMasterNormal(interface_j,l) << std::endl;
                    */

                    double value1 = -values[k] * current_normal[0] * contacts_list->getMortarScale(interface_j); //contacts_list->getMasterNormal(interface_j,0);
                    double value2 = -values[k] * current_normal[1] * contacts_list->getMortarScale(interface_j); //contacts_list->getMasterNormal(interface_j,1);
                    double value3 = -values[k] * current_normal[2] * contacts_list->getMortarScale(interface_j); //contacts_list->getMasterNormal(interface_j,2);

                    //double value1 = -values[k]* contacts_list->getMasterNormal(interface_j,0);
                    //double value2 = -values[k]* contacts_list->getMasterNormal(interface_j,1);
                    //double value3 = -values[k]* contacts_list->getMasterNormal(interface_j,2);
                    B->insertGlobalValues(row,
                                          Teuchos::tuple<global_ordinal_type>(col1),
                                          Teuchos::tuple<double>(value1));
                    B->insertGlobalValues(row,
                                          Teuchos::tuple<global_ordinal_type>(col2),
                                          Teuchos::tuple<double>(value2));
                    B->insertGlobalValues(row,
                                          Teuchos::tuple<global_ordinal_type>(col3),
                                          Teuchos::tuple<double>(value3));

                    if (contacts_list->isSticking(interface_j))
                    {
                        value1 = values[k] * current_tangential_1[0] * contacts_list->getMortarScale(interface_j);
                        value2 = values[k] * current_tangential_1[1] * contacts_list->getMortarScale(interface_j);
                        value3 = values[k] * current_tangential_1[2] * contacts_list->getMortarScale(interface_j);

                        B->insertGlobalValues(row + 1,
                                              Teuchos::tuple<global_ordinal_type>(col1),
                                              Teuchos::tuple<double>(value1));
                        B->insertGlobalValues(row + 1,
                                              Teuchos::tuple<global_ordinal_type>(col2),
                                              Teuchos::tuple<double>(value2));
                        B->insertGlobalValues(row + 1,
                                              Teuchos::tuple<global_ordinal_type>(col3),
                                              Teuchos::tuple<double>(value3));

                        value1 = values[k] * current_tangential_2[0] * contacts_list->getMortarScale(interface_j);
                        value2 = values[k] * current_tangential_2[1] * contacts_list->getMortarScale(interface_j);
                        value3 = values[k] * current_tangential_2[2] * contacts_list->getMortarScale(interface_j);

                        B->insertGlobalValues(row + 2,
                                              Teuchos::tuple<global_ordinal_type>(col1),
                                              Teuchos::tuple<double>(value1));
                        B->insertGlobalValues(row + 2,
                                              Teuchos::tuple<global_ordinal_type>(col2),
                                              Teuchos::tuple<double>(value2));
                        B->insertGlobalValues(row + 2,
                                              Teuchos::tuple<global_ordinal_type>(col3),
                                              Teuchos::tuple<double>(value3));
                    }
                    if (numDPN == 4)
                    {
                        global_ordinal_type row2 = row + numLMPN - 1;
                        global_ordinal_type col4 = Indices_int[k] * numDPN + 3;
                        B->insertGlobalValues(row2,
                                              Teuchos::tuple<global_ordinal_type>(col4),
                                              Teuchos::tuple<double>(values[k]));
                    }
                }
            }
        }
    }

    //std::cout << " compute_B double : o" << std::endl;

    B->fillComplete(map->mapDofs, map->mapLagrangeDofs);

    //std::cout << " compute_B double : p" << std::endl;
    return B;
}
