#include "wContact.h"
#include "wProblem.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include <algorithm>

#include <fstream>
#include <sstream>
#include <string>

using namespace katropton;
using namespace tbox;

bool searchCode(char const *code, char const *line)
{
    return !(strncmp(code, line, strlen(code)));
}

bool searchCode(char const *code, std::string const line)
{
    return !(strncmp(code, line.c_str(), strlen(code)));
}

bool searchCode(std::string const code, std::string const line)
{
    return line.find(code) != std::string::npos;
    //return strncmp(code.c_str(), line.c_str(), strlen(code.c_str()));
}

void readGMSH(std::string const &file_name, std::string const &name, std::vector<int> &nodes)
{
    std::cout << file_name << ": start " << std::endl;
    std::vector<int> nodes_tmp;
    std::ifstream infile(file_name);
    int tag = 0;
    std::string line;
    bool reading_physical_names = false;
    while (std::getline(infile, line))
    {
        //std::cout << line << std::endl;
        if (reading_physical_names)
        {
            if (searchCode("$EndPhysicalNames", line))
                break;

            //std::cout << line << " " << searchCode(name,line) << std::endl;
            if (searchCode(name, line))
            {
                std::istringstream line_stream(line);
                int dim = 0, no = 0;
                line_stream >> dim >> no;
                tag = no;
                std::cout << "tag: " << tag << std::endl;
                reading_physical_names = false;
            }
        }
        else if (searchCode("$Elements", line))
        {
            std::getline(infile, line);
            std::istringstream line_stream(line);
            // read number of elements
            int nbelm = 0;

            //if(sscanf(line, "%d\n", &nbelm) !=1)
            //    throw std::runtime_error("Error reading number of elements (readMshElements)");
            line_stream >> nbelm;

            for (int i = 0; i < nbelm; ++i)
            {
                std::getline(infile, line);
                std::istringstream line_stream(line);

                // read number, type and nb of tags
                int no = 0, type = 0, nbtags = 0;
                int tmp;
                //if(sscanf(line, "%d %d %d", &no, &type, &nbtags) !=3)
                //    throw std::runtime_error("Error reading &no, &type, &nbtags (readMshElements)");
                //if(verb) std::cout << no << ( (i==nbelm-1)? ".\n" : ",");
                line_stream >> no >> type >> nbtags;
                // read/create/link tags

                bool do_I_continue = true;
                for (int j = 0; j < nbtags; ++j)
                {
                    //if(sscanf(line, "%d", &tmp) !=1)
                    //    throw std::runtime_error("Error reading tags");
                    line_stream >> tmp;
                    //std::cout << tmp << " ";
                    if (tag == tmp)
                        do_I_continue = false;
                }
                if (do_I_continue)
                    continue;

                //std::cout << tmp << " " << tag<<" " << no << std::endl;
                int n = 0;
                // read nodes
                if (type == static_cast<int>(ELTYPE::HEX8)) // 8-node hexa
                    n = 8;
                else if (type == static_cast<int>(ELTYPE::QUAD4)) // 4-node quad
                    n = 4;
                else if (type == static_cast<int>(ELTYPE::TETRA4)) // 4-node tetra
                    n = 4;
                else if (type == static_cast<int>(ELTYPE::LINE2)) // 2-node line
                    n = 2;
                else if (type == static_cast<int>(ELTYPE::POINT1)) // 1-node point
                    n = 1;
                else if (type == static_cast<int>(ELTYPE::TRI3)) // 3-node triangle
                    n = 3;
                else
                {
                    continue;
                }
                int nn;
                for (int j = 0; j < n; ++j)
                {
                    //if(sscanf(line, "%d", &nn) !=1)
                    //    throw std::runtime_error("Error reading element/node links (readElNods)");
                    line_stream >> nn;
                    nodes_tmp.push_back(nn - 1);
                }
            }
        }
        else if (searchCode("$EndElements", line))
        {
            std::cout << "nodes_tmp.size(): " << nodes_tmp.size() << std::endl;
            if (nodes_tmp.size() > 0)
            {
                std::sort(nodes_tmp.begin(), nodes_tmp.end());

                nodes.push_back(nodes_tmp[0]);

                for (auto i = 1; i < nodes_tmp.size(); ++i)
                    if (nodes_tmp[i] != nodes_tmp[i - 1])
                        nodes.push_back(nodes_tmp[i]);
            }
            std::cout << "nodes.size(): " << nodes.size() << std::endl;
            break;
        }
        else if (searchCode("$PhysicalNames", line))
        {
            //std::cout<< "$PhysicalNames" << std::endl;
            reading_physical_names = true;
        }
    }
}

Contact::Contact(katropton::Problem &pbl, int no, std::string _Contact_name, Pt _slave_normal, double _dg) : Group(pbl.msh, no), Contact_name(_Contact_name), slave_normal(_slave_normal), dg(_dg)
{
    dg_vector = {dg};
    for (auto e : tag->elems)
        slave_elems.push_back(e);
}

Contact::Contact(katropton::Problem &pbl, std::string const &name, std::string _Contact_name, Pt _slave_normal, double _dg) : Group(pbl.msh, name), Contact_name(_Contact_name), slave_normal(_slave_normal), dg(_dg)
{
    dg_vector = {dg};
    for (auto e : tag->elems)
        slave_elems.push_back(e);
}

Contact::Contact(katropton::Problem &pbl, int no, std::string _Contact_name, Pt _slave_normal, std::vector<double> _dg) : Group(pbl.msh, no), Contact_name(_Contact_name), slave_normal(_slave_normal)
{
    dg = _dg[0];
    dg_vector = _dg;

    for (auto e : tag->elems)
        slave_elems.push_back(e);
}

Contact::Contact(katropton::Problem &pbl, std::string const &name, std::string _Contact_name, Pt _slave_normal, std::vector<double> _dg) : Group(pbl.msh, name), Contact_name(_Contact_name), slave_normal(_slave_normal)
{
    dg = _dg[0];
    dg_vector = _dg;

    for (auto e : tag->elems)
        slave_elems.push_back(e);
}

void Contact::setMaster(katropton::Problem &pbl, int no, Pt _master_normal)
{
    master_normal = _master_normal;
    Group master_group(pbl.msh, no);
    Tag *master_tag = master_group.tag;
    for (auto e : master_tag->elems)
        master_elems.push_back(e);

    pbl.Contacts.push_back(this);
}

void Contact::setMaster(katropton::Problem &pbl, std::string const &name, Pt _master_normal)
{
    master_normal = _master_normal;
    Group master_group(pbl.msh, name);
    Tag *master_tag = master_group.tag;
    for (auto e : master_tag->elems)
        master_elems.push_back(e);

    pbl.Contacts.push_back(this);
}

void Contact::write(std::ostream &out) const
{
    out << "Contact interface named ''" << Contact_name << "'' on " << *tag << "\n";
}

void Contact::setSlaveBoundary(std::string const &file_name, std::string const &name)
{
    readGMSH(file_name, name, slave_boundary_nodes);

    std::cout << "slave_boundary_nodes.size(): " << slave_boundary_nodes.size() << std::endl;
}

void Contact::setMasterBoundary(std::string const &file_name, std::string const &name)
{
    readGMSH(file_name, name, master_boundary_nodes);

    std::cout << "master_boundary_nodes.size(): " << master_boundary_nodes.size() << std::endl;
}

void Contact::setSlaveDirichlet(std::string const &file_name, std::string const &name)
{
    readGMSH(file_name, name, slave_dirichlet_nodes);

    std::cout << "slave_dirichlet_nodes.size(): " << slave_dirichlet_nodes.size() << std::endl;
    for (auto i = 0; i < slave_dirichlet_nodes.size(); ++i)
        std::cout << slave_dirichlet_nodes[i] << std::endl;
}

void Contact::setMasterDirichlet(std::string const &file_name, std::string const &name)
{
    readGMSH(file_name, name, master_dirichlet_nodes);

    std::cout << "master_dirichlet_nodes.size(): " << master_dirichlet_nodes.size() << std::endl;
}
