#ifndef WCONTACT_H
#define WCONTACT_H

#include "katropton.h"
#include "wGroup.h"
#include "wPt.h"
#include <vector>
#include <string>

namespace katropton
{

/**
 * @brief manage mechanical Dirichlet BC
 */

class KATROPTON_API Contact : public tbox::Group
{
public:
    std::string Contact_name;
    std::vector<tbox::Element *> slave_elems;
    std::vector<tbox::Element *> master_elems;

    std::vector<int> slave_boundary_nodes;
    std::vector<int> master_boundary_nodes;

    std::vector<int> slave_dirichlet_nodes;
    std::vector<int> master_dirichlet_nodes;

    std::vector<int> initially_closed_nodes = {};

    bool initial_closed = true;
    bool standard_shape_functions = true;
    bool is_Signorini = false;
    bool no_Update = false;
    bool sticking = false;
    bool mesh_tying = false;

    int nb_Lagrange_multipliers = 1;

    tbox::Pt slave_normal;
    tbox::Pt master_normal;

    double dg;

    double scale = 1.;

    std::vector<double> dg_vector;    

    Contact(Problem &pbl, int no, std::string _Contact_name, tbox::Pt _slave_normal, double dg = 0.);
    Contact(Problem &pbl, std::string const &name, std::string _Contact_name, tbox::Pt _slave_normal, double dg = 0.);

    Contact(Problem &pbl, int no, std::string _Contact_name, tbox::Pt _slave_normal, std::vector<double> _dg);
    Contact(Problem &pbl, std::string const &name, std::string _Contact_name, tbox::Pt _slave_normal, std::vector<double> _dg);

    void setMaster(katropton::Problem &pbl, int no, tbox::Pt _master_normal);
    void setMaster(katropton::Problem &pbl, std::string const &name, tbox::Pt _master_normal);

    void setSlaveBoundary(std::string const &file_name, std::string const &name);
    void setMasterBoundary(std::string const &file_name, std::string const &name);

    void setSlaveDirichlet(std::string const &file_name, std::string const &name);
    void setMasterDirichlet(std::string const &file_name, std::string const &name);

    void setInitialyClosed()
    {
        initial_closed = true;
    };
    void setInitialyOpen()
    {
        initial_closed = false;
    };

    void setNoUpdate()
    {
        no_Update = true;
    };
    void setUpdate()
    {
        no_Update = false;
    };

    void setMeshTying()
    {
        mesh_tying = true;
        sticking = true;
        no_Update = true;
    };

    void setSticking()
    {
        sticking = true;
    };
    void setNoSticking()
    {
        sticking = false;
    };

    void setStandardShapeFunctions()
    {
        standard_shape_functions = true;
    };
    void setDualShapeFunctions()
    {
        standard_shape_functions = false;
    };

    void setSignoriniContact()
    {
        is_Signorini = true;
    };
    void setTwoBodyContact()
    {
        is_Signorini = false;
    };

    void setInitiallyClosedNodes(std::vector<double> _initially_closed_nodes)
    {
        for (auto i = 0; i < _initially_closed_nodes.size(); ++i)
            initially_closed_nodes.push_back(((int)_initially_closed_nodes[i]));
    };

    void setScale(double a)
    {
        scale = a;
    };

#ifndef SWIG
    virtual void write(std::ostream &out) const;
#endif
};

} // namespace katropton

#endif //WCONTACT_H
