#include <Kokkos_Core.hpp>

#include "Stokhos_Tpetra_MP_Vector.hpp"
#include "Stokhos_Tpetra_Utilities_MP_Vector.hpp"
#include "Tpetra_ConfigDefs.hpp"

#include "Tpetra_Map.hpp"
#include "Tpetra_MultiVector.hpp"
#include "Tpetra_Vector.hpp"
#include "Tpetra_CrsGraph.hpp"
#include "Tpetra_CrsMatrix.hpp"
#include "Stokhos_Tpetra_CG.hpp"
#include "katropton.h"
#include "wProblem.h"
#include "wTag.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"
#include "wMedium.h"
#include "wuDirichlet.h"
#include "wNeumann.h"
#include "wSource.h"
#include "wContact.h"
#include "wPt.h"
#include <map>
#include "wTag.h"
#include <math.h>
#include "wHex8.h"
#include "wQuad4.h"
#include "wTri3.h"
#include "wCacheHex8.h"
#include "wGaussHex8.h"
#include <mpi.h>
//using namespace tbox;

template <typename scalar>
void meshConvert(katropton::Problem pbl,
                 Kokkos::View<int **, Kokkos::LayoutRight> &mesh_Elems,
                 Kokkos::View<double **, Kokkos::LayoutRight> &mesh_Nodes,
                 Kokkos::View<scalar ***, Kokkos::LayoutRight> &values_Materials,
                 Kokkos::View<int **, Kokkos::LayoutRight> &mesh_BCs,
                 Kokkos::View<int **, Kokkos::LayoutRight> &mesh_nBCs,
                 Kokkos::View<int **, Kokkos::LayoutRight> &mesh_Sources,
                 Kokkos::View<int **, Kokkos::LayoutRight> &mesh_Contacts,
                 Kokkos::View<scalar ***, Kokkos::LayoutRight> &values_dBCs,
                 Kokkos::View<scalar ***, Kokkos::LayoutRight> &values_nBCs,
                 Kokkos::View<scalar ***, Kokkos::LayoutRight> &values_Sources,
                 Teuchos::RCP<const Tpetra::Map<>> mapNodesWO,
                 Teuchos::RCP<const Tpetra::Map<>> mapElems,
                 std::vector<Tpetra::Vector<>::global_ordinal_type> myNodesWO,
                 Tpetra::Vector<>::local_ordinal_type numDPN,
                 Tpetra::Vector<>::local_ordinal_type &numMyElems,
                 Tpetra::Vector<>::local_ordinal_type &elems_type,
                 Tpetra::Vector<>::local_ordinal_type &elems_size,
                 Tpetra::Vector<>::local_ordinal_type &elems_media,
                 Tpetra::Vector<>::local_ordinal_type &elems_nodes,
                 Tpetra::Vector<>::local_ordinal_type &numMyMedias,
                 Tpetra::Vector<>::local_ordinal_type &materials_T,
                 Tpetra::Vector<>::local_ordinal_type &materials_E,
                 Tpetra::Vector<>::local_ordinal_type &materials_nu,
                 Tpetra::Vector<>::local_ordinal_type &materials_k,
                 Tpetra::Vector<>::local_ordinal_type &materials_beta,

                 Tpetra::Vector<>::local_ordinal_type &numMyBCs,
                 Tpetra::Vector<>::local_ordinal_type &BCs_size,
                 Tpetra::Vector<>::local_ordinal_type &BCs_nodes,
                 Tpetra::Vector<>::local_ordinal_type &dBCs_time,
                 Tpetra::Vector<>::local_ordinal_type &dBCs_dofs,
                 Tpetra::Vector<>::local_ordinal_type &dBCs_values,

                 Tpetra::Vector<>::local_ordinal_type &numMynBCs,
                 Tpetra::Vector<>::local_ordinal_type &nBCs_size,
                 Tpetra::Vector<>::local_ordinal_type &nBCs_elems,
                 Tpetra::Vector<>::local_ordinal_type &nBCs_time,
                 Tpetra::Vector<>::local_ordinal_type &nBCs_dofs,
                 Tpetra::Vector<>::local_ordinal_type &nBCs_values,

                 Tpetra::Vector<>::local_ordinal_type &numMySources,
                 Tpetra::Vector<>::local_ordinal_type &sources_size,
                 Tpetra::Vector<>::local_ordinal_type &sources_elems,
                 Tpetra::Vector<>::local_ordinal_type &sources_time,
                 Tpetra::Vector<>::local_ordinal_type &sources_values,

                 Tpetra::Vector<>::local_ordinal_type &numMyContacts,
                 Tpetra::Vector<>::local_ordinal_type &Contacts_slave_nodes_size,
                 Tpetra::Vector<>::local_ordinal_type &Contacts_slave_elems_size,
                 Tpetra::Vector<>::local_ordinal_type &Contacts_master_nodes_size,
                 Tpetra::Vector<>::local_ordinal_type &Contacts_master_elems_size,
                 Tpetra::Vector<>::local_ordinal_type &Contacts_slave_normal,
                 Tpetra::Vector<>::local_ordinal_type &Contacts_master_normal,
                 Tpetra::Vector<>::local_ordinal_type &Contacts_dg,
                 Tpetra::Vector<>::local_ordinal_type &Contacts_slave_nodes,
                 Tpetra::Vector<>::local_ordinal_type &Contacts_slave_elems,
                 Tpetra::Vector<>::local_ordinal_type &Contacts_master_nodes,
                 Tpetra::Vector<>::local_ordinal_type &Contacts_master_elems)
{
    auto msh = pbl.msh;
    // ------------------------------------------
    //
    //         Trilinos initialization
    //
    // ------------------------------------------
    using std::cerr;
    using std::cout;
    using std::endl;
    using Teuchos::arcp;
    using Teuchos::Array;
    using Teuchos::ArrayRCP;
    using Teuchos::ArrayView;
    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;

    using Kokkos::ALL;
    using Kokkos::subview;

    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    typedef Kokkos::View<int **, Kokkos::LayoutRight> elems_view_type;
    typedef Kokkos::View<double **, Kokkos::LayoutRight> nodes_view_type;
    typedef Kokkos::View<scalar ***, Kokkos::LayoutRight> materials_view_type;
    typedef Kokkos::View<int **, Kokkos::LayoutRight> bcs_view_type;
    typedef Kokkos::View<int **, Kokkos::LayoutRight> contact_view_type;

    // Elements

    numMyElems = msh->elems.size();
    numMyMedias = pbl.media.size();

    elems_type = 0;
    elems_size = 1;
    elems_media = 2;
    elems_nodes = 3;

    mesh_Elems = elems_view_type("R", numMyElems, elems_nodes + 8);

    for (auto i = 0; i < numMyElems; ++i)
    {
        tbox::Element *e = msh->elems[i];
        mesh_Elems(i, elems_type) = e->type();
        mesh_Elems(i, elems_size) = e->nodes.size();
        for (auto j = 0; j < numMyMedias; ++j)
            if (e->ptag == (pbl.media[j]->tag))
            {
                mesh_Elems(i, elems_media) = j;
                break;
            }
        for (auto j = 0; j < e->nodes.size(); ++j)
            mesh_Elems(i, elems_nodes + j) = mapNodesWO->getLocalElement(e->nodes[j]->row);
    }

    // Nodes

    mesh_Nodes = nodes_view_type("R", myNodesWO.size(), 3);
    for (auto i = 0; i < myNodesWO.size(); ++i)
        for (auto j = 0; j < 3; ++j)
            mesh_Nodes(i, j) = msh->nodes[mapNodesWO->getGlobalElement(i)]->pos.x[j];

    // Material
    materials_T = 0;
    materials_E = 1;
    materials_nu = 2;
    materials_k = 3;
    materials_beta = 4;

    values_Materials = materials_view_type("R", numMyMedias, 1, 5);
    for (auto i = 0; i < numMyMedias; ++i)
    {
        values_Materials(i, 0, materials_T) = 0; // !!! needed to be changed latter on !!!
        values_Materials(i, 0, materials_E) = pbl.media[i]->E;
        values_Materials(i, 0, materials_nu) = pbl.media[i]->nu;
        values_Materials(i, 0, materials_k) = pbl.media[i]->k;
        values_Materials(i, 0, materials_beta) = pbl.media[i]->beta;
    }

    // BC

    numMyBCs = pbl.duBCs.size();
    local_ordinal_type maxNumNodesPerBC = 0;

    std::vector<std::vector<global_ordinal_type>> bcNodes = {};

    for (auto i = 0; i < numMyBCs; ++i)
    {
        std::vector<global_ordinal_type> myBcNodes = {};
        for (auto j = 0; j < pbl.duBCs[i]->tag->elems.size(); ++j)
        {
            local_ordinal_type e = mapElems->getLocalElement(pbl.duBCs[i]->tag->elems[j]->no);
            if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::QUAD4))
                ;
            else if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::TRI3))
                ;
            else if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::LINE2))
                ;
            else if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::POINT1))
                ;
            else
                continue;
            for (auto n1 = 0; n1 < mesh_Elems(e, elems_size); ++n1)
            {
                global_ordinal_type i_global = mapNodesWO->getGlobalElement(mesh_Elems(e, elems_nodes + n1));
                if (std::find(myBcNodes.begin(), myBcNodes.end(), i_global) == myBcNodes.end())
                    myBcNodes.push_back(i_global);
            }
        }
        bcNodes.push_back(myBcNodes);
        if (maxNumNodesPerBC < myBcNodes.size())
            maxNumNodesPerBC = myBcNodes.size();
    }

    BCs_size = 0;
    BCs_nodes = 1;
    mesh_BCs = bcs_view_type("R", numMyBCs, BCs_nodes + maxNumNodesPerBC);

    for (auto i = 0; i < numMyBCs; ++i)
    {
        mesh_BCs(i, BCs_size) = bcNodes[i].size();
        for (auto j = 0; j < bcNodes[i].size(); ++j)
            mesh_BCs(i, BCs_nodes + j) = bcNodes[i][j];
    }

    dBCs_time = 0;
    dBCs_dofs = 1;
    dBCs_values = 1 + numDPN;
    values_dBCs = materials_view_type("R", numMyBCs, 1, 1 + 2 * numDPN);
    for (auto i = 0; i < numMyBCs; ++i)
    {
        values_dBCs(i, 0, dBCs_time) = 0.; //to change latter
        for (auto j = 0; j < numDPN; ++j)
            if (pbl.duBCs[i]->which_dof[j])
                values_dBCs(i, 0, dBCs_dofs + j) = 1;
            else
                values_dBCs(i, 0, dBCs_dofs + j) = 0;

        for (auto j = 0; j < numDPN; ++j)
            values_dBCs(i, 0, dBCs_values + j) = pbl.duBCs[i]->values[j];
    }

    // nBC

    numMynBCs = pbl.nBCs.size();
    local_ordinal_type maxNumElemsPernBC = 0;

    std::vector<std::vector<global_ordinal_type>> nbcElems = {};

    for (auto i = 0; i < numMynBCs; ++i)
    {
        std::vector<global_ordinal_type> mynBcElems = {};
        for (auto j = 0; j < pbl.nBCs[i]->tag->elems.size(); ++j)
        {
            local_ordinal_type e = mapElems->getLocalElement(pbl.nBCs[i]->tag->elems[j]->no);

            if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::QUAD4))
                ;
            else if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::TRI3))
                ;
            else if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::LINE2))
                ;
            else if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::POINT1))
                ;
            else
                continue;

            mynBcElems.push_back(e);
        }
        nbcElems.push_back(mynBcElems);
        if (maxNumElemsPernBC < mynBcElems.size())
            maxNumElemsPernBC = mynBcElems.size();
    }

    nBCs_size = 0;
    nBCs_elems = 1;
    mesh_nBCs = bcs_view_type("R", numMynBCs, nBCs_elems + maxNumElemsPernBC);

    for (auto i = 0; i < numMynBCs; ++i)
    {
        mesh_nBCs(i, nBCs_size) = nbcElems[i].size();

        for (auto j = 0; j < nbcElems[i].size(); ++j)
            mesh_nBCs(i, nBCs_elems + j) = nbcElems[i][j];
    }

    nBCs_time = 0;
    nBCs_dofs = 1;
    nBCs_values = 1 + numDPN;
    values_nBCs = materials_view_type("R", numMynBCs, 1, 1 + 2 * numDPN);
    for (auto i = 0; i < numMynBCs; ++i)
    {
        values_nBCs(i, 0, nBCs_time) = 0.; //to change latter
        for (auto j = 0; j < numDPN; ++j)
            if (pbl.nBCs[i]->which_dof[j])
                values_nBCs(i, 0, nBCs_dofs + j) = 1;
            else
                values_nBCs(i, 0, nBCs_dofs + j) = 0;

        for (auto j = 0; j < numDPN; ++j)
            values_nBCs(i, 0, nBCs_values + j) = pbl.nBCs[i]->values[j];
    }
    // Sources

    numMySources = pbl.Sources.size();
    local_ordinal_type maxNumElemsPerSource = 0;

    std::vector<std::vector<global_ordinal_type>> SourceElems = {};

    for (auto i = 0; i < numMySources; ++i)
    {
        std::vector<global_ordinal_type> mySourceElems = {};
        for (auto j = 0; j < pbl.Sources[i]->tag->elems.size(); ++j)
        {
            local_ordinal_type e = mapElems->getLocalElement(pbl.Sources[i]->tag->elems[j]->no);

            if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::TETRA4))
                ;
            else if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::HEX8))
                ;
            else
                continue;

            mySourceElems.push_back(e);
        }
        SourceElems.push_back(mySourceElems);
        if (maxNumElemsPerSource < mySourceElems.size())
            maxNumElemsPerSource = mySourceElems.size();
    }

    sources_size = 0;
    sources_elems = 1;
    mesh_Sources = bcs_view_type("R", numMySources, sources_elems + maxNumElemsPerSource);

    for (auto i = 0; i < numMySources; ++i)
    {
        mesh_Sources(i, sources_size) = SourceElems[i].size();

        for (auto j = 0; j < SourceElems[i].size(); ++j)
            mesh_Sources(i, sources_elems + j) = SourceElems[i][j];
    }

    sources_time = 0;
    sources_values = 1;
    values_Sources = materials_view_type("R", numMynBCs, 1, 2);
    for (auto i = 0; i < numMySources; ++i)
    {
        values_Sources(i, 0, sources_time) = 0.; //to change latter

        values_Sources(i, 0, sources_values) = pbl.Sources[i]->value;
    }
    // Contacts

    numMyContacts = pbl.Contacts.size();
    local_ordinal_type maxNumSlaveNodesPerContactFace = 0;
    local_ordinal_type maxNumMasterNodesPerContactFace = 0;
    local_ordinal_type maxNumSlaveElemsPerContactFace = 0;
    local_ordinal_type maxNumMasterElemsPerContactFace = 0;

    std::vector<std::vector<global_ordinal_type>> contactSlaveNodes = {};
    std::vector<std::vector<global_ordinal_type>> contactMasterNodes = {};
    std::vector<std::vector<global_ordinal_type>> contactSlaveElems = {};
    std::vector<std::vector<global_ordinal_type>> contactMasterElems = {};

    for (auto i = 0; i < numMyContacts; ++i)
    {
        std::vector<global_ordinal_type> myContactSlaveNodes = {};
        std::vector<global_ordinal_type> myContactSlaveElems = {};
        for (auto j = 0; j < pbl.Contacts[i]->slave_elems.size(); ++j)
        {
            local_ordinal_type e = mapElems->getLocalElement(pbl.Contacts[i]->slave_elems[j]->no);
            if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::QUAD4))
                ;
            else if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::TRI3))
                ;
            else
                continue;
            myContactSlaveElems.push_back(e);
            for (auto n1 = 0; n1 < mesh_Elems(e, elems_size); ++n1)
            {
                global_ordinal_type i_global = mapNodesWO->getGlobalElement(mesh_Elems(e, elems_nodes + n1));
                if (std::find(myContactSlaveNodes.begin(), myContactSlaveNodes.end(), i_global) == myContactSlaveNodes.end())
                {
                    myContactSlaveNodes.push_back(i_global);
                    //out<< "i: " << i_global << endl;
                }
            }
        }
        contactSlaveNodes.push_back(myContactSlaveNodes);
        contactSlaveElems.push_back(myContactSlaveElems);

        std::vector<global_ordinal_type> myContactMasterNodes = {};
        std::vector<global_ordinal_type> myContactMasterElems = {};
        for (auto j = 0; j < pbl.Contacts[i]->master_elems.size(); ++j)
        {
            local_ordinal_type e = mapElems->getLocalElement(pbl.Contacts[i]->master_elems[j]->no);
            if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::QUAD4))
                ;
            else if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::TRI3))
                ;
            else if (mesh_Elems(e, elems_type) == static_cast<int>(tbox::ELTYPE::LINE2))
                ;
            else
                continue;
            myContactMasterElems.push_back(e);
            for (auto n1 = 0; n1 < mesh_Elems(e, elems_size); ++n1)
            {
                global_ordinal_type i_global = mapNodesWO->getGlobalElement(mesh_Elems(e, elems_nodes + n1));
                if (std::find(myContactMasterNodes.begin(), myContactMasterNodes.end(), i_global) == myContactMasterNodes.end())
                    myContactMasterNodes.push_back(i_global);
            }
        }

        contactMasterNodes.push_back(myContactMasterNodes);
        contactMasterElems.push_back(myContactMasterElems);

        if (maxNumSlaveNodesPerContactFace < myContactSlaveNodes.size())
            maxNumSlaveNodesPerContactFace = myContactSlaveNodes.size();
        if (maxNumSlaveElemsPerContactFace < myContactSlaveElems.size())
            maxNumSlaveElemsPerContactFace = myContactSlaveElems.size();

        if (maxNumMasterNodesPerContactFace < myContactMasterNodes.size())
            maxNumMasterNodesPerContactFace = myContactMasterNodes.size();
        if (maxNumMasterElemsPerContactFace < myContactMasterElems.size())
            maxNumMasterElemsPerContactFace = myContactMasterElems.size();
    }

    Contacts_slave_nodes_size = 0;
    Contacts_slave_elems_size = 1;
    Contacts_master_nodes_size = 2;
    Contacts_master_elems_size = 3;
    Contacts_slave_normal = 4;
    Contacts_master_normal = 7;
    Contacts_slave_nodes = 10;
    Contacts_slave_elems = Contacts_slave_nodes + maxNumSlaveNodesPerContactFace;
    Contacts_master_nodes = Contacts_slave_elems + maxNumSlaveElemsPerContactFace;
    Contacts_master_elems = Contacts_master_nodes + maxNumMasterNodesPerContactFace;

    Contacts_dg = Contacts_master_elems + maxNumMasterElemsPerContactFace;

    mesh_Contacts = contact_view_type("R", numMyContacts, Contacts_dg + 1);

    for (auto i = 0; i < numMyContacts; ++i)
    {

        mesh_Contacts(i, Contacts_slave_nodes_size) = contactSlaveNodes[i].size();
        mesh_Contacts(i, Contacts_slave_elems_size) = contactSlaveElems[i].size();
        mesh_Contacts(i, Contacts_master_nodes_size) = contactMasterNodes[i].size();
        mesh_Contacts(i, Contacts_master_elems_size) = contactMasterElems[i].size();

        for (auto j = 0; j < 3; ++j)
        {
            mesh_Contacts(i, Contacts_slave_normal + j) = pbl.Contacts[i]->slave_normal.x[j];
            mesh_Contacts(i, Contacts_master_normal + j) = pbl.Contacts[i]->master_normal.x[j];
        }

        mesh_Contacts(i, Contacts_dg) = pbl.Contacts[i]->dg;

        for (auto j = 0; j < contactSlaveNodes[i].size(); ++j)
            mesh_Contacts(i, Contacts_slave_nodes + j) = contactSlaveNodes[i][j];
        for (auto j = 0; j < contactSlaveElems[i].size(); ++j)
            mesh_Contacts(i, Contacts_slave_elems + j) = contactSlaveElems[i][j];
        for (auto j = 0; j < contactMasterNodes[i].size(); ++j)
            mesh_Contacts(i, Contacts_master_nodes + j) = contactMasterNodes[i][j];
        for (auto j = 0; j < contactMasterElems[i].size(); ++j)
            mesh_Contacts(i, Contacts_master_elems + j) = contactMasterElems[i][j];
    }
}

template <typename scalar>
void meshContactMPI(katropton::Problem pbl,
                    Kokkos::View<int **, Kokkos::LayoutRight> &mesh_Contacts_Elems,
                    Kokkos::View<double **, Kokkos::LayoutRight> &mesh_Contacts_Nodes,
                    Kokkos::View<int **, Kokkos::LayoutRight> &mesh_Contacts2,
                    Teuchos::RCP<const Tpetra::Map<>> &mapContactNodes,
                    Teuchos::RCP<const Tpetra::Map<>> mapNodesWO,
                    Teuchos::RCP<const Tpetra::Map<>> mapElems,
                    std::vector<Tpetra::Vector<>::global_ordinal_type> myNodesWO,
                    Tpetra::Vector<>::local_ordinal_type numDPN,
                    Tpetra::Vector<>::local_ordinal_type &numMyElems,
                    Tpetra::Vector<>::local_ordinal_type &elems_type,
                    Tpetra::Vector<>::local_ordinal_type &elems_size,
                    Tpetra::Vector<>::local_ordinal_type &elems_media,
                    Tpetra::Vector<>::local_ordinal_type &elems_nodes,
                    Tpetra::Vector<>::local_ordinal_type &numMyContacts,
                    Tpetra::Vector<>::local_ordinal_type &Contacts_slave_nodes_size,
                    Tpetra::Vector<>::local_ordinal_type &Contacts_slave_elems_size,
                    Tpetra::Vector<>::local_ordinal_type &Contacts_master_nodes_size,
                    Tpetra::Vector<>::local_ordinal_type &Contacts_master_elems_size,
                    Tpetra::Vector<>::local_ordinal_type &Contacts_slave_normal,
                    Tpetra::Vector<>::local_ordinal_type &Contacts_master_normal,
                    Tpetra::Vector<>::local_ordinal_type &Contacts_dg,
                    Tpetra::Vector<>::local_ordinal_type &Contacts_slave_nodes,
                    Tpetra::Vector<>::local_ordinal_type &Contacts_slave_elems,
                    Tpetra::Vector<>::local_ordinal_type &Contacts_master_nodes,
                    Tpetra::Vector<>::local_ordinal_type &Contacts_master_elems,
                    const size_t myRank,
                    const Tpetra::Vector<>::global_ordinal_type indexBase,
                    Teuchos::RCP<const Teuchos::Comm<int>> comm)
{
    auto msh = pbl.msh;
    // ------------------------------------------
    //
    //         Trilinos initialization
    //
    // ------------------------------------------
    using std::cerr;
    using std::cout;
    using std::endl;
    using Teuchos::arcp;
    using Teuchos::Array;
    using Teuchos::ArrayRCP;
    using Teuchos::ArrayView;
    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;

    using Kokkos::ALL;
    using Kokkos::subview;

    typedef Tpetra::Map<> map_type;
    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    typedef Kokkos::View<int **, Kokkos::LayoutRight> elems_view_type;
    typedef Kokkos::View<double **, Kokkos::LayoutRight> nodes_view_type;
    typedef Kokkos::View<int **, Kokkos::LayoutRight> contact_view_type;
    typedef Tpetra::Export<> export_type;

    std::vector<global_ordinal_type> myContactElems = {};
    for (auto i = 0; i < pbl.Contacts.size(); ++i)
    {
        for (auto j = 0; j < pbl.Contacts[i]->slave_elems.size(); ++j)
        {
            global_ordinal_type e = pbl.Contacts[i]->slave_elems[j]->no;
            if (pbl.Contacts[i]->slave_elems[j]->type() == tbox::ELTYPE::QUAD4)
                ;
            else if (pbl.Contacts[i]->slave_elems[j]->type() == tbox::ELTYPE::TRI3)
                ;
            else
                continue;
            myContactElems.push_back(e);
        }
        for (auto j = 0; j < pbl.Contacts[i]->master_elems.size(); ++j)
        {
            global_ordinal_type e = pbl.Contacts[i]->master_elems[j]->no;
            if (pbl.Contacts[i]->master_elems[j]->type() == tbox::ELTYPE::QUAD4)
                ;
            else if (pbl.Contacts[i]->master_elems[j]->type() == tbox::ELTYPE::TRI3)
                ;
            else
                continue;
            myContactElems.push_back(e);
        }
    }

    RCP<const map_type> mapContactElems;
    mapContactElems = Tpetra::createNonContigMap<local_ordinal_type, global_ordinal_type>(myContactElems, comm);

    std::vector<global_ordinal_type> allContactElems = {};
    {
        int numtasks;
        MPI_Comm_size(MPI_COMM_WORLD, &numtasks);
        int itsContactElemsSize, currentElem;
        if (myRank == 0)
        {
            for (size_t i = 1; i < numtasks; ++i)
            {
                MPI_Recv(&itsContactElemsSize, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                for (size_t j = 0; j < itsContactElemsSize; ++j)
                {
                    MPI_Recv(&currentElem, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                    allContactElems.push_back(currentElem);
                }
            }

            allContactElems.insert(allContactElems.end(), myContactElems.begin(), myContactElems.end());
        }
        else
        {
            itsContactElemsSize = myContactElems.size();
            MPI_Send(&itsContactElemsSize, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
            for (size_t j = 0; j < itsContactElemsSize; ++j)
            {
                currentElem = myContactElems[j];
                MPI_Send(&currentElem, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
            }
        }
    }

    const global_ordinal_type numContactElems = mapContactElems->getGlobalNumElements();
    //const global_ordinal_type numContactElems0 = (myRank == 0) ? numContactElems : 0;
    //RCP<const map_type> mapContactElems0 = rcp (new map_type (numContactElems, numContactElems0, indexBase, comm));

    RCP<const map_type> mapContactElems0;
    mapContactElems0 = Tpetra::createNonContigMap<local_ordinal_type, global_ordinal_type>(allContactElems, comm);

    //out << "mapContactElems->getGlobalNumElements ()" << mapContactElems->getGlobalNumElements () <<endl;

    const int SLAVE = 1;
    const int MASTER = 2;

    const global_ordinal_type numColContactElems = 9;
    RCP<const map_type> mapColContactElems = rcp(new map_type(numColContactElems, numColContactElems, indexBase, comm));

    RCP<Tpetra::CrsMatrix<global_ordinal_type>> contactElems = rcp(new Tpetra::CrsMatrix<global_ordinal_type>(mapContactElems, 0));

    const global_ordinal_type contactElems_type = 0;
    const global_ordinal_type contactElems_size = 1;
    const global_ordinal_type contactElems_nodes = 2;
    const global_ordinal_type contactElems_interface = 6;
    const global_ordinal_type contactElems_sOm = 7;
    const global_ordinal_type contactElems_mat = 8;

    for (auto i = 0; i < pbl.Contacts.size(); ++i)
    {
        for (auto j = 0; j < pbl.Contacts[i]->slave_elems.size(); ++j)
        {
            global_ordinal_type e = pbl.Contacts[i]->slave_elems[j]->no;
            if (pbl.Contacts[i]->slave_elems[j]->type() == tbox::ELTYPE::QUAD4)
                ;
            else if (pbl.Contacts[i]->slave_elems[j]->type() == tbox::ELTYPE::TRI3)
                ;
            else
                continue;

            contactElems->insertGlobalValues(e,
                                             tuple<global_ordinal_type>(contactElems_type),
                                             tuple<global_ordinal_type>(static_cast<int>(pbl.Contacts[i]->slave_elems[j]->type())));
            contactElems->insertGlobalValues(e,
                                             tuple<global_ordinal_type>(contactElems_size),
                                             tuple<global_ordinal_type>(pbl.Contacts[i]->slave_elems[j]->nodes.size()));
            for (auto k = 0; k < pbl.Contacts[i]->slave_elems[j]->nodes.size(); ++k)
                contactElems->insertGlobalValues(e,
                                                 tuple<global_ordinal_type>(contactElems_nodes + k),
                                                 tuple<global_ordinal_type>(pbl.Contacts[i]->slave_elems[j]->nodes[k]->row));

            contactElems->insertGlobalValues(e,
                                             tuple<global_ordinal_type>(contactElems_interface),
                                             tuple<global_ordinal_type>(i));
            contactElems->insertGlobalValues(e,
                                             tuple<global_ordinal_type>(contactElems_sOm),
                                             tuple<global_ordinal_type>(SLAVE));
            for (auto k = 0; k < pbl.media.size(); ++k)
                if (pbl.Contacts[i]->slave_elems[j]->ptag == (pbl.media[k]->tag))
                    contactElems->insertGlobalValues(e,
                                                     tuple<global_ordinal_type>(contactElems_mat),
                                                     tuple<global_ordinal_type>(k));
        }
        for (auto j = 0; j < pbl.Contacts[i]->master_elems.size(); ++j)
        {
            global_ordinal_type e = pbl.Contacts[i]->master_elems[j]->no;
            if (pbl.Contacts[i]->master_elems[j]->type() == tbox::ELTYPE::QUAD4)
                ;
            else if (pbl.Contacts[i]->master_elems[j]->type() == tbox::ELTYPE::TRI3)
                ;
            else
                continue;
            contactElems->insertGlobalValues(e,
                                             tuple<global_ordinal_type>(contactElems_type),
                                             tuple<global_ordinal_type>(static_cast<int>(pbl.Contacts[i]->master_elems[j]->type())));
            contactElems->insertGlobalValues(e,
                                             tuple<global_ordinal_type>(contactElems_size),
                                             tuple<global_ordinal_type>(pbl.Contacts[i]->master_elems[j]->nodes.size()));
            for (auto k = 0; k < pbl.Contacts[i]->master_elems[j]->nodes.size(); ++k)
                contactElems->insertGlobalValues(e,
                                                 tuple<global_ordinal_type>(contactElems_nodes + k),
                                                 tuple<global_ordinal_type>(pbl.Contacts[i]->master_elems[j]->nodes[k]->row));

            contactElems->insertGlobalValues(e,
                                             tuple<global_ordinal_type>(contactElems_interface),
                                             tuple<global_ordinal_type>(i));
            contactElems->insertGlobalValues(e,
                                             tuple<global_ordinal_type>(contactElems_sOm),
                                             tuple<global_ordinal_type>(MASTER));
            for (auto k = 0; k < pbl.media.size(); ++k)
                if (pbl.Contacts[i]->master_elems[j]->ptag == (pbl.media[k]->tag))
                    contactElems->insertGlobalValues(e,
                                                     tuple<global_ordinal_type>(contactElems_mat),
                                                     tuple<global_ordinal_type>(k));
        }
    }
    contactElems->fillComplete(mapColContactElems, mapContactElems);
    //std::cout << mapContactElems->getNodeNumElements() << std::endl;
    //std::cout << mapContactElems0->getNodeNumElements() << std::endl;
    //std::cout << mapContactElems->getGlobalNumElements() << std::endl;
    //std::cout << mapContactElems0->getGlobalNumElements() << std::endl;
    const export_type exportContactElems(mapContactElems, mapContactElems0); //!!!!

    RCP<Tpetra::CrsMatrix<global_ordinal_type>> contactElems0 = rcp(new Tpetra::CrsMatrix<global_ordinal_type>(mapContactElems0, 0)); //!!!!
    contactElems0->doExport(*contactElems, exportContactElems, Tpetra::INSERT);

    contactElems0->fillComplete(mapColContactElems, mapContactElems0);

    //std::cout << "Number of elems" << (contactElems->getGlobalNumEntries () )/8. << endl;
    //out << contactElems0->getGlobalNumEntries ()  << endl<< endl <<numContactElems0 << endl<< endl;

    numMyContacts = pbl.Contacts.size();
    local_ordinal_type maxNumSlaveNodesPerContactFace = 0;
    local_ordinal_type maxNumMasterNodesPerContactFace = 0;
    local_ordinal_type maxNumSlaveElemsPerContactFace = 0;
    local_ordinal_type maxNumMasterElemsPerContactFace = 0;

    std::vector<std::vector<global_ordinal_type>> contactSlaveNodes = {};
    std::vector<std::vector<global_ordinal_type>> contactMasterNodes = {};
    std::vector<std::vector<global_ordinal_type>> contactSlaveElems = {};
    std::vector<std::vector<global_ordinal_type>> contactMasterElems = {};

    for (auto i = 0; i < numMyContacts; ++i)
    {

        std::vector<global_ordinal_type> myContactSlaveNodes = {};
        std::vector<global_ordinal_type> myContactSlaveElems = {};
        std::vector<global_ordinal_type> myContactMasterNodes = {};
        std::vector<global_ordinal_type> myContactMasterElems = {};

        for (auto e_local = mapContactElems0->getMinLocalIndex(); e_local <= mapContactElems0->getMaxLocalIndex(); ++e_local)
        {
            Teuchos::ArrayView<const local_ordinal_type> local_indices;
            Teuchos::ArrayView<const global_ordinal_type> values;
            contactElems0->getLocalRowView(e_local, local_indices, values);

            if (values.size() > 0 && values[contactElems_interface] == i)
            {
                if (values[contactElems_sOm] == SLAVE)
                {
                    myContactSlaveElems.push_back(e_local);
                    for (auto k = 0; k < values[contactElems_size]; ++k)
                    {
                        global_ordinal_type i_global = values[contactElems_nodes + k];
                        if (std::find(myContactSlaveNodes.begin(), myContactSlaveNodes.end(), i_global) == myContactSlaveNodes.end())
                            myContactSlaveNodes.push_back(i_global);
                    }
                }
                else
                {
                    myContactMasterElems.push_back(e_local);
                    for (auto k = 0; k < values[contactElems_size]; ++k)
                    {
                        global_ordinal_type i_global = values[contactElems_nodes + k];
                        if (std::find(myContactMasterNodes.begin(), myContactMasterNodes.end(), i_global) == myContactMasterNodes.end())
                            myContactMasterNodes.push_back(i_global);
                    }
                }
            }
        }
        contactSlaveNodes.push_back(myContactSlaveNodes);
        contactSlaveElems.push_back(myContactSlaveElems);
        contactMasterNodes.push_back(myContactMasterNodes);
        contactMasterElems.push_back(myContactMasterElems);

        //std::cout << "i:" << i << "number of sn" << myContactSlaveNodes.size() << endl;
        //std::cout << "i:" << i << "number of mn" << myContactMasterNodes.size() << endl;
        //std::cout << "i:" << i << "number of se" << myContactSlaveElems.size() << endl;
        //std::cout << "i:" << i << "number of me" << myContactMasterElems.size() << endl;

        if (maxNumSlaveNodesPerContactFace < myContactSlaveNodes.size())
            maxNumSlaveNodesPerContactFace = myContactSlaveNodes.size();
        if (maxNumSlaveElemsPerContactFace < myContactSlaveElems.size())
            maxNumSlaveElemsPerContactFace = myContactSlaveElems.size();

        if (maxNumMasterNodesPerContactFace < myContactMasterNodes.size())
            maxNumMasterNodesPerContactFace = myContactMasterNodes.size();
        if (maxNumMasterElemsPerContactFace < myContactMasterElems.size())
            maxNumMasterElemsPerContactFace = myContactMasterElems.size();
    }

    std::vector<global_ordinal_type> allContactNodes = {};
    for (auto i = 0; i < numMyContacts; ++i)
    {
        allContactNodes.insert(allContactNodes.end(), contactSlaveNodes[i].begin(), contactSlaveNodes[i].end());
        allContactNodes.insert(allContactNodes.end(), contactMasterNodes[i].begin(), contactMasterNodes[i].end());
    }

    mapContactNodes = Tpetra::createNonContigMap<local_ordinal_type, global_ordinal_type>(allContactNodes, comm);

    mesh_Contacts_Nodes = nodes_view_type("R", allContactNodes.size(), 3);
    for (auto i = 0; i < allContactNodes.size(); ++i)
    {
        global_ordinal_type i_global = allContactNodes[i];
        for (auto j = 0; j < 3; ++j)
            mesh_Contacts_Nodes(i, j) = msh->nodes[i_global]->pos.x[j];
    }

    const global_ordinal_type numMyContactsElems = 1 + mapContactElems0->getMaxLocalIndex() - mapContactElems0->getMinLocalIndex();

    elems_type = 0;
    elems_size = 1;
    elems_media = 2;
    elems_nodes = 3;

    mesh_Contacts_Elems = elems_view_type("R", numMyContactsElems, elems_nodes + 8);

    for (auto e_local = mapContactElems0->getMinLocalIndex(); e_local <= mapContactElems0->getMaxLocalIndex(); ++e_local)
    {
        Teuchos::ArrayView<const local_ordinal_type> local_indices;
        Teuchos::ArrayView<const global_ordinal_type> values;
        contactElems0->getLocalRowView(e_local, local_indices, values);
        if (values.size())
        {
            mesh_Contacts_Elems(e_local, elems_type) = values[contactElems_type];
            mesh_Contacts_Elems(e_local, elems_size) = values[contactElems_size];
            mesh_Contacts_Elems(e_local, elems_media) = values[contactElems_mat];
            for (auto k = 0; k < values[contactElems_size]; ++k)
                mesh_Contacts_Elems(e_local, elems_nodes + k) = mapContactNodes->getLocalElement(values[contactElems_nodes + k]);
        }
    }
    Contacts_slave_nodes_size = 0;
    Contacts_slave_elems_size = 1;
    Contacts_master_nodes_size = 2;
    Contacts_master_elems_size = 3;
    Contacts_slave_normal = 4;
    Contacts_master_normal = 7;
    Contacts_slave_nodes = 10;
    Contacts_slave_elems = Contacts_slave_nodes + maxNumSlaveNodesPerContactFace;
    Contacts_master_nodes = Contacts_slave_elems + maxNumSlaveElemsPerContactFace;
    Contacts_master_elems = Contacts_master_nodes + maxNumMasterNodesPerContactFace;
    Contacts_dg = Contacts_master_elems + maxNumMasterElemsPerContactFace;

    mesh_Contacts2 = contact_view_type("R", numMyContacts, Contacts_dg + 1);

    for (auto i = 0; i < numMyContacts; ++i)
    {

        mesh_Contacts2(i, Contacts_slave_nodes_size) = contactSlaveNodes[i].size();
        mesh_Contacts2(i, Contacts_slave_elems_size) = contactSlaveElems[i].size();
        mesh_Contacts2(i, Contacts_master_nodes_size) = contactMasterNodes[i].size();
        mesh_Contacts2(i, Contacts_master_elems_size) = contactMasterElems[i].size();

        for (auto j = 0; j < 3; ++j)
        {
            mesh_Contacts2(i, Contacts_slave_normal + j) = pbl.Contacts[i]->slave_normal.x[j];
            mesh_Contacts2(i, Contacts_master_normal + j) = pbl.Contacts[i]->master_normal.x[j];
        }

        mesh_Contacts2(i, Contacts_dg) = pbl.Contacts[i]->dg;

        for (auto j = 0; j < contactSlaveNodes[i].size(); ++j)
            mesh_Contacts2(i, Contacts_slave_nodes + j) = contactSlaveNodes[i][j];
        for (auto j = 0; j < contactSlaveElems[i].size(); ++j)
            mesh_Contacts2(i, Contacts_slave_elems + j) = contactSlaveElems[i][j];
        for (auto j = 0; j < contactMasterNodes[i].size(); ++j)
            mesh_Contacts2(i, Contacts_master_nodes + j) = contactMasterNodes[i][j];
        for (auto j = 0; j < contactMasterElems[i].size(); ++j)
            mesh_Contacts2(i, Contacts_master_elems + j) = contactMasterElems[i][j];
    }
}
