#ifndef WPROBLEM_H
#define WPROBLEM_H

#include "katropton.h"
#include "wObject.h"
#include <iostream>
#include <vector>
#include <string>
#include <memory>

#include <mpi.h>

namespace katropton
{

/**
 * @brief a class containing the problem parameters
 */

class KATROPTON_API Problem : public fwk::wObject
{
public:
    std::shared_ptr<tbox::MshData> msh; ///< Mesh structure

    MPI_Comm comm;
#ifndef SWIG
    std::vector<Medium *> media = {};     ///< Materials
    std::vector<uDirichlet *> duBCs = {}; ///< mechanical Dirichlet boundary conditions
    std::vector<Neumann *> nBCs = {};
    std::vector<Contact *> Contacts = {};
    std::vector<Source *> Sources = {};
    std::vector<RandomVariable *> RandomVariables = {};
#endif
public:
    Problem(std::shared_ptr<tbox::MshData> _msh, MPI_Comm _comm);
    int execute();

#ifndef SWIG
    virtual void write(std::ostream &out) const;
#endif
};

} // namespace katropton

#endif //WPROBLEM_H
