#include "wuDirichlet.h"
#include "wProblem.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include <algorithm>
using namespace katropton;
using namespace tbox;

uDirichlet::uDirichlet(katropton::Problem &pbl, std::string const &name, std::string const &_BC_name, std::vector<int> _which_dof, std::vector<double> _values, size_t ensemble_size) : Group(pbl.msh, name), BC_name(_BC_name), which_dof(_which_dof), values(_values)
{
    for (auto e : tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);

    x_values = {};
    y_values = {};
    z_values = {};
    T_values = {};
    for (auto i = 0; i < ensemble_size; ++i)
    {
        x_values.push_back(values[0]);
        y_values.push_back(values[1]);
        z_values.push_back(values[2]);
        T_values.push_back(values[3]);
    }
    pbl.duBCs.push_back(this);
}

uDirichlet::uDirichlet(Problem &pbl, std::string const &name, std::string const &_BC_name, int x, double v_x, int y, double v_y, int z, double v_z, int T, double v_T, size_t ensemble_size) : Group(pbl.msh, name), BC_name(_BC_name)
{
    for (auto e : tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);

    which_dof = {};
    which_dof.push_back(x);
    which_dof.push_back(y);
    which_dof.push_back(z);
    which_dof.push_back(T);

    values = {};
    values.push_back(v_x);
    values.push_back(v_y);
    values.push_back(v_z);
    values.push_back(v_T);

    x_values = {};
    y_values = {};
    z_values = {};
    T_values = {};
    for (auto i = 0; i < ensemble_size; ++i)
    {
        x_values.push_back(v_x);
        y_values.push_back(v_y);
        z_values.push_back(v_z);
        T_values.push_back(v_T);
    }

    pbl.duBCs.push_back(this);
}

uDirichlet::uDirichlet(Problem &pbl, std::string const &name, std::string const &_BC_name, int x, std::vector<double> v_x, int y, std::vector<double> v_y, int z, std::vector<double> v_z, int T, std::vector<double> v_T) : Group(pbl.msh, name), BC_name(_BC_name)
{
    for (auto e : tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);

    std::cout << v_x[0] << std::endl;

    which_dof = {};
    which_dof.push_back(x);
    which_dof.push_back(y);
    which_dof.push_back(z);
    which_dof.push_back(T);

    values = {};
    values.push_back(v_x[0]);
    values.push_back(v_y[0]);
    values.push_back(v_z[0]);
    values.push_back(v_T[0]);

    x_values = {};
    y_values = {};
    z_values = {};
    T_values = {};
    for (auto i = 0; i < v_x.size(); ++i)
        x_values.push_back(v_x[i]);

    for (auto i = 0; i < v_y.size(); ++i)
        y_values.push_back(v_y[i]);

    for (auto i = 0; i < v_z.size(); ++i)
        z_values.push_back(v_z[i]);

    for (auto i = 0; i < v_T.size(); ++i)
        T_values.push_back(v_T[i]);

    //std::cout << x_values[0] << std::endl;

    //y_values = v_y;
    //z_values = v_z;
    //T_values = v_T;

    pbl.duBCs.push_back(this);
}

uDirichlet::uDirichlet(Teuchos::RCP<Problem> pbl, std::string const &name, std::string const &_BC_name, int x, std::vector<double> v_x, int y, std::vector<double> v_y, int z, std::vector<double> v_z, int T, std::vector<double> v_T) : Group(pbl->msh, name), BC_name(_BC_name)
{
    for (auto e : tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);

    //std::cout << v_x[0] << std::endl;

    which_dof = {};
    which_dof.push_back(x);
    which_dof.push_back(y);
    which_dof.push_back(z);
    which_dof.push_back(T);

    values = {};
    values.push_back(v_x[0]);
    values.push_back(v_y[0]);
    values.push_back(v_z[0]);
    values.push_back(v_T[0]);

    x_values = {};
    y_values = {};
    z_values = {};
    T_values = {};
    for (auto i = 0; i < v_x.size(); ++i)
        x_values.push_back(v_x[i]);

    for (auto i = 0; i < v_y.size(); ++i)
        y_values.push_back(v_y[i]);

    for (auto i = 0; i < v_z.size(); ++i)
        z_values.push_back(v_z[i]);

    for (auto i = 0; i < v_T.size(); ++i)
        T_values.push_back(v_T[i]);

    //std::cout << x_values[0] << std::endl;

    //y_values = v_y;
    //z_values = v_z;
    //T_values = v_T;

    pbl->duBCs.push_back(this);
    //std::cout << pbl->duBCs[pbl->duBCs.size()-1]->x_values[0] << std::endl;
}

/**
 *  @brief default implementation
 */

/*void
 Dirichlet::apply(std::vector<double> &rhs)
 {
 for(auto n : nodes)
 rhs[n->row] = val;
 }
 */

void uDirichlet::write(std::ostream &out) const
{
    out << "Dirichlet BC on " << *tag << "\n";
    out << "With fixed values of (";
    for (auto i = 0; i < (which_dof.size() - 1); ++i)
    {
        if (which_dof[i] == 1)
            out << values[i] << ", ";
        else
            out << "-, ";
    }

    if (which_dof[which_dof.size() - 1] == 1)
        out << values[which_dof.size() - 1] << ")\n";
    else
        out << "-)\n";

    out << "a '-' represents an unconstrained dof.\n";
    //for(auto n : nodes) { out << '\t' << *n << '\n'; }
}
