#ifndef WUDIRICHLET_H
#define WUDIRICHLET_H

#include "katropton.h"
#include "wGroup.h"
#include "wPt.h"
#include <vector>
#include <string>

#include <Teuchos_RCP.hpp>

namespace katropton
{

/**
 * @brief manage mechanical Dirichlet BC
 */

class KATROPTON_API uDirichlet : public tbox::Group
{
public:
    std::string const &BC_name;
    std::vector<tbox::Node *> nodes;
    std::vector<int> which_dof;
    std::vector<double> values;

    std::vector<double> x_values;
    std::vector<double> y_values;
    std::vector<double> z_values;
    std::vector<double> T_values;

    //uDirichlet(Problem &pbl, int no, std::string const &_BC_name, std::vector<int> const &_which_dof, std::vector<double> const &_values);
    uDirichlet(Problem &pbl, std::string const &name, std::string const &_BC_name, std::vector<int> _which_dof, std::vector<double> _values, size_t ensemble_size = 1);

    uDirichlet(Problem &pbl, std::string const &name, std::string const &_BC_name, int x, double v_x, int y, double v_y, int z, double v_z, int T = 0, double v_T = 0, size_t ensemble_size = 1);
    uDirichlet(Problem &pbl, std::string const &name, std::string const &_BC_name, int x, std::vector<double> v_x, int y, std::vector<double> v_y, int z, std::vector<double> v_z, int T = 0, std::vector<double> v_T = {});

    uDirichlet(Teuchos::RCP<Problem> pbl, std::string const &name, std::string const &_BC_name, int x, std::vector<double> v_x, int y, std::vector<double> v_y, int z, std::vector<double> v_z, int T = 0, std::vector<double> v_T = {});
#ifndef SWIG
    virtual void write(std::ostream &out) const;
#endif
};

} // namespace katropton

#endif //WUDIRICHLET_H
