#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import print_function
from builtins import str
from builtins import range
import katropton as m
import tbox
import tbox.gmsh as gmsh
import katropton.utilities as u
from fwk.testing import *


def main():
    comm, rank, size = u.import_MPI()

    from fwk.wutils import parseargs
    args = parseargs()

    from PyTrilinos import Teuchos
    import shlex
    import subprocess
    import os
    import numpy as np

    geo_name = '1_cube.geo'
    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    msh, mesh_name, partitioned_mesh_name = u.read_mesh_MPI(
        geo_name, file_dir, work_dir, comm, rank, size)

    ensemble_size = 1

    for k in range(1, 21):
        u.mkdir_MPI(str(k), comm, rank, size)
        os.chdir(str(k))
        pbl = m.Problem(msh, comm)

        E = 1. * np.ones(ensemble_size)
        nu = 0.2 * np.ones(ensemble_size)
        k = 0.114
        d = 0.

        hf = 1 * np.ones(ensemble_size)
        zero = np.zeros(ensemble_size)
        m.Medium(pbl, "Body 1", "test", E, nu, k, d)

        m.uDirichlet(pbl, "Surf 1 1", "Clamped", 1, 0.,
                     1, 0., 1, 0., 1, 10., ensemble_size)
        m.Neumann(pbl, "Surf 1 2", "heat fluxes", 0, zero,
                  0, zero, 0, zero, 1, hf, ensemble_size)

        dx = np.zeros(ensemble_size)
        dy = 2. * np.ones(ensemble_size)
        dz = np.zeros(ensemble_size)
        dT = 10. * np.ones(ensemble_size)

        solverList = Teuchos.ParameterList()
        solverList['Convergence Tolerance'] = 10**(-8)
        solverList['Maximum Iterations'] = 2
        solverList['Verbosity'] = 33
        solverList['Flexible Gmres'] = False
        solverList['Output Style'] = 1
        solverList['Output Frequency'] = 1

        solverList['Use preconditioner'] = True

        solverList['type'] = "BlockGmres"
        solverList['Maximum active set iteration'] = 1

        solverList['Print Teuchos timers'] = True

        mueluParams = Teuchos.ParameterList()
        mueluParams['number of equations'] = 4
        mueluParams['max levels'] = 1
        mueluParams['coarse: type'] = "RELAXATION"

        mueluParamsSub2 = Teuchos.ParameterList()
        mueluParamsSub2['relaxation: type'] = "MT Symmetric Gauss-Seidel"
        mueluParamsSub2['relaxation: sweeps'] = 1
        mueluParamsSub2['relaxation: damping factor'] = 0.9

        mueluParams['verbosity'] = "low"

        mueluParams['coarse: params'] = mueluParamsSub2

        solverList['mueluParams'] = mueluParams
        solverList['Write txt files'] = True

        useEP = (ensemble_size > 1)

        if k == 1:
            slv = m.IterativeSolver(
                pbl, 1, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 2:
            slv = m.IterativeSolver(
                pbl, 2, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 3:
            slv = m.IterativeSolver(
                pbl, 3, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 4:
            slv = m.IterativeSolver(
                pbl, 4, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 5:
            slv = m.IterativeSolver(
                pbl, 5, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 6:
            slv = m.IterativeSolver(
                pbl, 6, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 7:
            slv = m.IterativeSolver(
                pbl, 7, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 8:
            slv = m.IterativeSolver(
                pbl, 8, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 9:
            slv = m.IterativeSolver(
                pbl, 9, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 10:
            slv = m.IterativeSolver(
                pbl, 10, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 11:
            slv = m.IterativeSolver(
                pbl, 11, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 12:
            slv = m.IterativeSolver(
                pbl, 12, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 13:
            slv = m.IterativeSolver(
                pbl, 13, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 14:
            slv = m.IterativeSolver(
                pbl, 14, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 15:
            slv = m.IterativeSolver(
                pbl, 15, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 16:
            slv = m.IterativeSolver(
                pbl, 16, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 17:
            slv = m.IterativeSolver(
                pbl, 17, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 18:
            slv = m.IterativeSolver(
                pbl, 18, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 19:
            slv = m.IterativeSolver(
                pbl, 19, solverList, 4, useEP, ensemble_size)
            slv.start()
        if k == 20:
            slv = m.IterativeSolver(
                pbl, 20, solverList, 4, useEP, ensemble_size)
            slv.start()
        os.chdir('..')


if __name__ == "__main__":
    main()
