#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import print_function
from builtins import str
from builtins import range
import katropton as m
import tbox
import tbox.gmsh as gmsh
import katropton.utilities as u
from fwk.testing import *


def main():
    comm, rank, size = u.import_MPI()

    from fwk.wutils import parseargs
    args = parseargs()

    from PyTrilinos import Teuchos

    import shlex
    import subprocess
    import os
    import numpy as np

    geo_name = '2_cubes.geo'
    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    msh, mesh_name, partitioned_mesh_name = u.read_mesh_MPI(
        geo_name, file_dir, work_dir, comm, rank, size)

    ensemble_size = 1

    pbl = m.Problem(msh, comm)

    #E = np.linspace(0.5,1.5,ensemble_size)
    E = 1. * np.ones(ensemble_size)
    nu = 0.2 * np.ones(ensemble_size)
    k = 0.114
    d = 0.
    m.Medium(pbl, "Body 1", "test", E, nu, k, d)
    m.Medium(pbl, "Body 2", "test", E, nu, k, d)

    m.uDirichlet(pbl, "Surf 1 1", "Clamped", 1, 0.,
                 1, 0., 1, 0., 1, 10., ensemble_size)
    m.uDirichlet(pbl, "Surf 2 2", "Clamped", 1, 0.,
                 1, 0., 1, -2., 1, 20., ensemble_size)

    norm = tbox.Pt(0, 0, -1)
    cont = m.Contact(pbl, "Surf 1 2", "contact", norm)
    cont.setInitialyOpen()
    cont.setMaster(pbl, "Surf 2 1", norm)

    dx = np.zeros(ensemble_size)
    dy = 2. * np.ones(ensemble_size)
    dz = np.zeros(ensemble_size)
    dT = 10. * np.ones(ensemble_size)

    solverList = Teuchos.ParameterList()
    solverList['Convergence Tolerance'] = 10**(-8)
    solverList['Maximum Iterations'] = 1000
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    solverList["MueLu xml file name"] = file_dir + '/direct_solver.xml'
    solverList["convert MueLu xml file"] = True

    solverList['Use preconditioner'] = True

    solverList['type'] = "BlockGmres"
    solverList['Maximum active set iteration'] = 3

    solverList['Write matrix and vectors'] = True
    solverList['Write txt files'] = True

    useEP = (ensemble_size > 1)

    slv = m.IterativeSolver(pbl, args.k, solverList, 4, useEP, ensemble_size)

    slv.start()

    tests = CTests()
    measured_total_CPU_cost = slv.getTimers()["total"].read().getReal().sec()

    non_zero_CPU_cost = (measured_total_CPU_cost > 0.)
    tests.add(CTest('non zero total CPU cost', non_zero_CPU_cost, True))
    tests.run()


if __name__ == "__main__":
    main()
